/*****************************************************************************/
/* Copyright (c) 2005, ZeitControl cardsystems GmbH                          */
/* All rights reserved.                                                      */
/*                                                                           */
/* Redistribution and use in source and binary forms, with or without        */
/* modification, are permitted provided that the following conditions are    */
/* met:                                                                      */
/*    * Redistributions of source code must retain the above copyright       */
/*      notice, this list of conditions and the following disclaimer.        */
/*    * Redistributions in binary form must reproduce the above copyright    */
/*      notice, this list of conditions and the following disclaimer in the  */
/*      documentation and/or other materials provided with the distribution. */
/*    * Neither the name of the company ZeitControl nor the names of its     */
/*      contributors may be used to endorse or promote products derived from */
/*      this software without specific prior written permission.             */
/*                                                                           */
/* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS       */
/* "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED */
/* TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A           */
/* PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER  */
/* OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,  */
/* EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,       */
/* PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR        */
/* PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF    */
/* LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING      */
/* NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS        */
/* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.              */
/*****************************************************************************/
/* Created on 01.04.2005
 * Project: BasicCard .Net Library
 * File description: 
 *   Abstract base class for BasicCard card services.
 * --------------------------------------------------------
 * $Header: /opt/cvsroot/net/BasicCardLib.net/BasicCard/BasicCards/AbstractBasicCardService.cs,v 1.7 2011/03/02 17:36:30 michael Exp $
 * $Revision: 1.7 $
 * $Date: 2011/03/02 17:36:30 $
 * --------------------------------------------------------
 * $Log: AbstractBasicCardService.cs,v $
 * Revision 1.7  2011/03/02 17:36:30  michael
 * *** empty log message ***
 *
 * Revision 1.6  2010/07/21 16:30:24  michael
 * BugFix: member variable fileio was not initialized
 *
 * Revision 1.5  2010/06/09 15:30:03  michael
 * Changes for Windows Mobile
 *
 * Revision 1.4  2009/03/19 17:35:50  michael
 * V1.3
 *
 * Revision 1.3  2009/03/18 12:50:41  michael
 * Release Candidate .Net 2.0
 *
 * Revision 1.2  2007/10/23 11:35:29  michael
 * Intermediate Version
 *
 * Revision 1.1  2005/07/22 12:57:57  michael
 * Initial checkin
 *
 *---------------------------------------------------------
 */
using System;
using System.Collections;
using System.Security.Cryptography;
using BasicCard;
using BasicCard.Terminals;
using BasicCard.BasicCards;
using BasicCard.ProcessorCards;
using BasicCard.Framework;
using BasicCard.Crypto.Messaging;
using BasicCard.Logging;
using BasicCard.GenericCards;
using BasicCard.BasicCards.Components;
using BasicCard.BasicCards.FileSys;
using System.Collections.Generic;

namespace BasicCard.BasicCards
{
	/// <summary>
	/// BasicCard cryptographic algorithms for start encryption.
	/// </summary>
	[CLSCompliant(true), Serializable]
	public enum CryptAlgo
	{
		/// <summary>
		/// No encryption
		/// </summary>
		None = -1,
		/// <summary>
		/// Automatically select algorithm
		/// </summary>
		Auto = 0,
		/// <summary>
		/// SGLFSR
		/// </summary>
		Sglfsr = 0x11,
		/// <summary>
		/// SGLFSR with CRC
		/// </summary>
		SglfsrCrc = 0x12,
		/// <summary>
		/// DES
		/// </summary>
		Des = 0x21,
		/// <summary>
		/// Triple DES
		/// </summary>
		TripleDes = 0x22,
		/// <summary>
		/// DES with CRC
		/// </summary>
		DesCrc = 0x23,
		/// <summary>
		/// Triple DES-EDE2 (16 byte key) with CRC
		/// </summary>
		TripleDesEde2Crc = 0x24,
		/// <summary>
		/// Triple DES-EDE2 (24 byte key) with CRC
		/// </summary>
		TripleDesEde3Crc = 0x25,
		/// <summary>
		/// AES with 128 bit key
		/// </summary>
		Aes128 = 0x31,
		/// <summary>
		/// AES with 192 bit key
		/// </summary>
		Aes192 = 0x32,
		/// <summary>
		/// AES with 256 bit key
		/// </summary>
		Aes256 = 0x33,
		/// <summary>
		/// EAX using AES with 128 bit key
		/// </summary>
		EaxAes128 = 0x41,
		/// <summary>
		/// EAX using AES with 192 bit key
		/// </summary>
		EaxAes192 = 0x42,
		/// <summary>
		/// EAX using AES with 256 bit key
		/// </summary>
		EaxAes256 = 0x43,
		/// <summary>
		/// OMAC using AES with 128 bit key
		/// </summary>
		OmacAes128 = 0x81,
		/// <summary>
		/// OMAC using AES with 192 bit key
		/// </summary>
		OmacAes192 = 0x82,
		/// <summary>
		/// OMAC using AES with 256 bit key
		/// </summary>
		OmacAes256 = 0x83
	}

	/// <summary>
	/// Phase for load sequence.
	/// </summary>
	[CLSCompliant(true), Serializable]
	public enum LoadSequencePhase
	{
		/// <summary>
		/// start load sequence
		/// </summary>
		Start = 1,
		/// <summary>
		/// end load sequence
		/// </summary>
		End = 2,
		/// <summary>
		/// abort load sequence
		/// </summary>
		Abort = 3
	}


	/// <summary>
	/// AbstractBasicCardService gives protected access to all important predefined
	/// BasicCard commands and supports custom command calls. Extend this for use by
	/// your own service.
	/// <seealso cref="BasicCardService"/>
	/// </summary>
	/// <threadsafety static="true" instance="false"/>
	[CLSCompliant(true)]
	public class AbstractBasicCardService : AbstractCardService
	{
		/// <summary>
		/// Class implementing CardSelector. Recommended to use for checking
		/// for desired BasicCard by validating ATR, ATR-History and/or
		/// Application ID.
		/// <seealso cref="BasicCardSelector"/>
		/// </summary>
		/// <remarks>
		/// <para>
		/// For usage extend this class and override desired virtual Accept...
		/// function. Do not forget to set matching Check.... properties
		/// to true, otherwise the Accept... functions are not called.
		/// </para>
		/// <para>
		/// All (and only) marked validation functions are called. A card is accepted 
		/// only if all marked validation functions return true. Call order for
		/// validation functions is:
		/// </para>
		/// <list type="bullet">
		/// <item>
		/// <description>
		/// AcceptThisAtr
		/// </description>
		/// </item>
		/// <item>
		/// <description>
		/// AcceptThisApplicationID
		/// </description>
		/// </item>
		/// <item>
		/// <description>
		/// AcceptThisCard
		/// </description>
		/// </item>
		/// </list>
		/// If one accept call returns false, following accptance functions are not called.
		/// </remarks>
		/// <threadsafety static="true" instance="true"/>
		[CLSCompliant(true), Serializable]
		public abstract class AbstractBasicCardSelector : ProcessorCardSelector
		{
			private bool checkATR = false;
			private bool checkApplicationID = false;
			private bool checkCustom = false;
			private bool checkCardID = false;
			private bool canShare = true;
			/// <summary>
			/// AbstractBasicCardService using this object.
			/// </summary>
			protected AbstractBasicCardService service;

			/// <summary>
			/// Set this to true for activation of <see cref="AcceptThisAtr">ATR check</see>.
			/// <seealso cref="AcceptThisAtr"/>
			/// </summary>
			protected bool CheckAtr
			{
				get
				{
					return this.checkATR;
				}
				set
				{
					this.checkATR=value;
				}
			}

			/// <summary>
			/// Set this to true to validate the card within
			/// <see cref="AcceptThisApplicationID"/> calls.
			/// <seealso cref="AcceptThisApplicationID"/>
			/// </summary>
			protected bool CheckApplicationID
			{
				get
				{
					return this.checkApplicationID;
				}
				set
				{
					this.checkApplicationID=value;
				}
			}

			/// <summary>
			/// Set this to true to validate the card within
			/// <see cref="AcceptThisCard(BasicCardString)"/> calls.
			/// <seealso cref="AcceptThisCard(BasicCardString)"/>
			/// </summary>
			protected bool CheckCustom
			{
				get
				{
					return this.checkCustom;
				}
				set
				{
					this.checkCustom=value;
				}
			}

			/// <summary>
			/// Set this to true to validate the card id of a 
			/// MultiApplication BasicCard within 
			/// <see cref="AcceptThisCard(BasicCardString)"/> calls.
			/// <seealso cref="AcceptThisCard(BasicCardString)"/>
			/// </summary>
			protected bool CheckCardID
			{
				get
				{
					return this.checkCardID;
				}
				set
				{
					this.checkCardID=value;
				}
			}

			/// <summary>
			/// Constructor.
			/// </summary>
			/// <param name="service">AbstractBasicCardService using this object.</param>
			public AbstractBasicCardSelector(AbstractBasicCardService service)
				: base()
			{
				this.service=service;
			}

			/// <summary>
			/// If <see cref="CheckAtr"/> was set to true, this function
			/// is called for validating the Atr of an inserted card.
			/// <seealso cref="CheckAtr"/>
			/// </summary>
			/// <param name="atr">ATR of inserted card.</param>
			/// <returns>True if card is accepted or false to reject this card.</returns>
			protected virtual bool AcceptThisAtr(ProcessorCardAtr atr)
			{
				return false;
			}

			/// <summary>
			/// If <see cref="CheckApplicationID"/> was set to true, this function
			/// is called for validating the ApplicationID of an inserted card.
			/// <seealso cref="CheckApplicationID"/>
			/// </summary>
			/// <remarks>
			/// This may be called several times for each application
			/// within an MultiApplication BasicCard. If an
			/// application of an MultiApplication BasicCard was accepted, then this application will
			/// be automatically selected by calling <see cref="SelectApplication"/>.  
			/// </remarks>
			/// <param name="appId">Application ID of inserted card.</param>
			/// <returns>True if card is accepted or false to reject this card.</returns>
			protected virtual bool AcceptThisApplicationID(BasicCardString appId)
			{
				return false;
			}

			/// <summary>
			/// If <see cref="CheckCardID"/> was set to true, this function
			/// is called for validating an inserted BasicCard.
			/// <seealso cref="CheckCardID"/>
			/// </summary>
			/// <param name="cardID">Card ID of MultiApplication BasicCard or null if not configured.</param>
			/// <returns>True if card is accepted or false to reject this card.</returns>
			protected virtual bool AcceptThisCard(BasicCardString cardID)
			{
				return false;
			}

			/// <summary>
			/// If <see cref="CheckCustom"/> was set to true, this function
			/// is called for validating an inserted BasicCard.
			/// <seealso cref="CheckCustom"/>
			/// </summary>
			/// <param name="service">BasicCardService to access the inserted card.</param>
			/// <returns>True if card is accepted or false to reject this card.</returns>
			protected virtual bool AcceptThisCard(BasicCardService service)
			{
				return false;
			}

			/// <summary>
			/// AbstractBasicCardSelector calls this funtion to select a application within a 
			/// MultiApplication BasicCard. Default implementation saves application
			/// file name for use by AbstractBasicCardService and sets <see cref="CanShare"/> to false.
			/// </summary>
			/// <param name="service">BasicCardService</param>
			/// <param name="applicationFileName">Application file name to select.</param>
			protected void SelectApplication(
				BasicCardService service,
				BasicCardString applicationFileName)
			{
					this.service.applicationFile = applicationFileName;
				canShare = false;
			}

			/// <summary>
			/// See <see cref="ICardSelector.WantThisCard">ICardSelector.WantThisCard</see>.
			/// </summary>
			/// <param name="reader">Card reader to use.</param>
			/// <returns>True if card is accepted or false to reject this card.</returns>
			public override bool WantThisCard(ICardReader reader)
			{
				ProcessorCardAtr atr = GetProcessorCardAtr(reader);
				if (null == atr)
					return false; // not a processor card

				if (checkATR)
				{
					if (!AcceptThisAtr(atr))
						return false; // ATR not accepted
				}

				BasicCardService service = null;
				ICardConnection conn = null;
				try
				{
					service = new BasicCardService();
					service.doesNeedExclusiveConnection = false;
					// make sure card is powered
					if (checkApplicationID || checkCardID || checkCustom)
					{
						// connect shared
						conn = reader.Connect(true);
						service.Init(conn);
					}

					if (checkCardID)
					{
						//	first try to get CardID of MultiApplication BasicCard	
						BasicCardString cardID = null;
						try
						{
							cardID = service.CmdGetApplicationID(0, 2);
						}
						catch (CommsException)
						{
						}
						if (!AcceptThisCard(cardID))
							return false;
					}

					if (checkApplicationID)
					{
						// first try to get CardID of MultiApplication BasicCard	
						BasicCardString applicationID;
						bool moreApps = true;
						bool accepted = false;
						bool selected = false;
						for (int index = 0; moreApps && !accepted; index++)
						{
							try
							{
								applicationID =
									service.CmdGetApplicationID(index, 0);
								if (AcceptThisApplicationID(applicationID))
									accepted = true;
								if (accepted)
								{
									if (index == 0)
										selected = true;
									BasicCardString file =
										service.CmdGetApplicationID(index, 1);
									// if it is not a MultiApplication BasicCard there will be an
									// exception and selectApplication is not executed
									SelectApplication(service, file);
									selected = true;
								}
							}
							catch (CommsException ex)
							{
								switch (ex.SW1SW2)
								{
									case StatusWord.swApplicationNotFound :
										if (index > 0)
										{
											//	no more apps
											moreApps = false;
										}
										break;
									case StatusWord.swP1P2Error :
										// not multi app
										moreApps = false;
										break;
									default :
										BasicCardLogger.Instance.LogException(
											this.GetType().Name,
											"WantThisCard",
											ex);
										moreApps = false;
										break;
								}
							}
						}
						if (!(accepted && selected))
							return false; // none of the applications is accepted
					}

					if (checkCustom)
					{
						if (!AcceptThisCard(service))
						return false; // card not accepted
					}

					return true;
				}
				catch (CommsException e)
				{
					BasicCardLogger.Instance.Log(BasicCardLogLevel.Finer,
						"Card not accepted because of CommsException {0}", e);
					return false;
				}
				finally
				{
					if (null != conn)
					{
						try
						{
							conn.Dispose(false);
						}
						catch (CardReaderException e1)
						{
							BasicCardLogger.Instance.LogException(
								this.GetType().Name,
								"WantThisCard",
								e1);
						}
					}
				}
			}

			/// <summary>
			/// True if sharing the connection is possible.
			/// </summary>
			public override bool CanShare
			{
				get
				{
                    bool shared=canShare;
                    if (null != service)
                    {
                        // true if canShare is true and 
                        // service does not need exclusive connection
                        shared&=!service.NeedExclusiveConnection;
                    }
                    return shared;
				}
			}

		}

		/// <summary>
		/// Simple BasicCard selector, allow selection by ATR history,
		/// or ApplicationID. By default all processor cards are accepted.
		/// </summary>
		/// <threadsafety static="true" instance="true"/>
        [CLSCompliant(true), Serializable]
		public class BasicCardSelector : AbstractBasicCardSelector
		{
			private BasicCardString appId;
			private byte[] atrHistory;
			private byte[] atrHistMask;

			/// <summary>
			/// Constructor
			/// </summary>
			/// <param name="service">AbstractBasicCardService using this object.</param>
			public BasicCardSelector(AbstractBasicCardService service)
				: base(service)
			{
				CheckAtr=true;
			}

			/// <summary>
			/// Application id to check for. Set to null (default) to disable
			/// check for application id.
			/// </summary>
			public BasicCardString ApplicationID
			{
				set
				{
					lock (this)
					{
						this.appId = value;
						CheckApplicationID=(null != appId);
					}
				}
				get
				{
					return this.appId;
				}
			}

			/// <summary>
			/// Required atr history bytes to check for. Set to null (default) to disable
			/// check for atr history bytes.
			/// </summary>
			public BasicCardString AtrHistory
			{
				set
				{
					lock (this)
					{
						if (null!=value)
						{
							this.atrHistory = value.Characters;
							this.atrHistMask = new byte[value.Length];
							for (int i = 0; i < this.atrHistMask.Length; i++)
							{
								this.atrHistMask[i] = (byte) 0xFF;
							}
						}
						else
						{
							this.atrHistory=null;
						}
					}
				}
				get
				{
					return new BasicCardString(this.atrHistory);
				}
			}

			/// <summary>
			/// Set required atr history bytes (with mask) to check for. Set to null (default) to disable
			/// check for atr history bytes. Each history byte is checked by XOR
			/// specified history byte with real history byte and finally AND (mask)
			/// it with mask byte. If result for each byte is 0, then the atr history matches.
			/// </summary>
			/// <param name="history">History bytes or null.</param>
			/// <param name="mask">History mask.</param>
			public void SetAtrHistory(byte[] history, byte[] mask)
			{
				lock (this)
				{
					this.atrHistory = history;
					this.atrHistMask = mask;
				}
			}

			/// <summary>
			/// Checks if application id of card matches the specified application id.
			/// </summary>
			/// <param name="appId">Application id of inserted card.</param>
			/// <returns>True if card is accepted.</returns>
			protected override bool AcceptThisApplicationID(BasicCardString appId)
			{
				lock (this)
				{
					return this.appId.Equals(appId);
				}
			}

			/// <summary>
			/// Checks if ATR of card matches specified ATR.
			/// </summary>
			/// <param name="atr">ATR of card.</param>
			/// <returns>True if card is accepted.</returns>
			protected override bool AcceptThisAtr(ProcessorCardAtr atr)
			{
				lock (this)
				{
					// by default accept all processor cards
					if (atrHistory == null)
						return true;
					try
					{
						byte[] hist = atr.HistoricalBytes;
						if (null == hist)
							return false;
						if (hist.Length < atrHistMask.Length)
							return false;
						for (int i = 0; i < atrHistMask.Length; i++)
						{
							byte test =
								(byte) ((hist[i] ^ atrHistory[i]) & atrHistMask[i]);
							if (0 != test)
								return false;
						}
						return true;
					}
					catch (CommsException e)
					{
						BasicCardLogger.Instance.LogException(
							this.GetType().Name,
							"AcceptThisAtr",
							e);
						return false;
					}
				}
			}

		}	

		/// <summary>
		/// Class for storing BasicCard encryption keys for use BasicCard
		/// encryption. For this each key is either associated with a
		/// key number, a CID or an component path (String).
		/// </summary>
		private class BCKeyStore : IDisposable
		{
			private IDictionary store = new Hashtable();
            private byte[] ck = null;
            private object dispSync = new object();
            private bool disposed = false;

			private void IntSetKey(object id, object encryptionKey)
			{
                if (disposed) throw new ObjectDisposedException("BasicCard Keystore");
				if (encryptionKey == null)
				{
					store.Remove(id);
				}
				else
				{
					store[id]=encryptionKey;
				}
			}

			/// <overloads>
			/// Set encryption key.
			/// </overloads>
			/// <summary>
			/// Set encryption key identified by a BasicCard key number
			/// (0..255) or a CID (component id).
			/// </summary>
			/// <param name="keynumber">Key number or CID.</param>
			/// <param name="encryptionKey">Encryption key or null.</param>
			public void SetKey(int keynumber, byte[] encryptionKey)
			{
				IntSetKey(keynumber, encryptionKey);
			}

			/// <summary>
			/// Set encryption key by component path.
			/// </summary>
			/// <param name="keyComponent">Full qualified path/name of component containing this key.</param>
			/// <param name="encryptionKey">Encryption key or null.</param>
			public void SetKey(string keyComponent, byte[] encryptionKey)
			{
				string id = keyComponent.ToUpper();
				IntSetKey(id, encryptionKey);
			}

            /// <summary>
            /// Set encryption key for customer key KDP encryption.
            /// </summary>            
            /// <param name="encryptionKey">Encryption key or null.</param>
            public void SetCustomerKey(byte[] encryptionKey)
            {
                if (disposed) throw new ObjectDisposedException("BasicCard Keystore");
                ck = (byte[])encryptionKey.Clone();
            }

			/// <summary>
			/// Add a new index, id for an existing key. E.g. to
			/// make key 1 available though CID 0x7001 use IndexKey(0x7001, 1)
			/// </summary>
			/// <param name="newKeynumber">additional number or CID for oldKeynumber</param>
			/// <param name="oldKeynumber">number or CID of existing key</param>
			public void IndexKey(int newKeynumber, int oldKeynumber)
			{
				object key = GetKey(oldKeynumber);
                if (null == key)
                {
                    throw new CardServiceException(StatusWord.swKeyNotLoaded);
                }
				int id = newKeynumber;
				IntSetKey(id, key);
			}

			/// <overloads>
			/// Get encryption key.
			/// </overloads>
			/// <summary>
			/// Get encryption key identified by a BasicCard key number
			/// (0..255) or a CID (component id) .
			/// </summary>
			/// <param name="keynumber">key number or CID</param>
			/// <returns>encryption key</returns>
			public byte[] GetKey(int keynumber)
			{
                if (disposed) throw new ObjectDisposedException("BasicCard Keystore");
				return (byte[]) store[keynumber];
			}

			/// <summary>
			/// Get encryption key identified by a component path.
			/// </summary>
			/// <param name="keyComponent">full qualified path/name of component containing this key</param>
			/// <returns>encryption key</returns>
			public byte[] GetKey(string keyComponent)
			{
                if (disposed) throw new ObjectDisposedException("BasicCard Keystore");
				string id = keyComponent.ToUpper();
				return (byte[]) store[id];
			}

            /// <summary>
            /// Get encryption key identified by a component path.
            /// </summary>            
            /// <returns>encryption key</returns>
            public byte[] GetCustomerKey()
            {
                if (disposed) throw new ObjectDisposedException("BasicCard Keystore");
                return ck;
            }

            /// <summary>
            /// Dispose and cleanup keys
            /// </summary>
            public void Dispose()
            {
                lock (dispSync)
                {
                    if (!disposed)
                    {
                        disposed=true;
                        byte[] xck=ck;
                        if (xck != null)
                        {
                            for (int i = 0; i < xck.Length; i++) xck[0] = 0;
                        }
                        foreach (byte[] d in store.Values)
                        {
                            xck = d;
                            if (xck != null)
                            {
                                for (int i = 0; i < xck.Length; i++) xck[0] = 0;
                            }
                        }
                    }
                }
            }
        }        

		/// <summary>
		/// CommandAndResponseCrypt for DES or AES with or without CRC.
		/// </summary>
		private class CryptAesDes : ICommandAndResponseCrypt
		{
			private int protocol;
			private bool encrypt = true;
			private bool decrypt = true;
			private MessageCrypt mcrypt;
            private byte[] cryptBuffer = new byte[8 * 1024];
			private int lenCryptBuffer = 0;
			internal bool isProfessional;
			internal bool isDES;
			private AbstractBasicCardService service;

			/// <summary>
			/// Construct CommandAndResponseCrypt object for DES or AES
			/// (with or without CRC) encryption.
			/// </summary>
			/// <param name="service">AbstractBasicCardService using this object</param>
			/// <param name="keydata">encryption key</param>
			/// <param name="algo">algorithm</param>
			/// <param name="RA">RA from start encryption command</param>
			/// <param name="RB">RB from start encryption response</param>
			/// <param name="forProfessional">true for Professional or MultiApplication BasicCard </param>
			public CryptAesDes(
				AbstractBasicCardService service,
				byte[] keydata,
				CryptAlgo algo,
				byte[] RA,
				byte[] RB,
				bool forProfessional)
			{
				this.service=service;
				BaseAlgorithm algorithm;
				int keylen;
				switch (algo)
				{
					case CryptAlgo.TripleDes :
						algorithm = BaseAlgorithm.TripleDes;
						isProfessional = false;
						keylen = 16;
						isDES = true;
						break;
					case CryptAlgo.TripleDesEde2Crc:
						algorithm = BaseAlgorithm.TripleDes;
						isProfessional = true;
						keylen = 16;
						isDES = true;
						break;
					case CryptAlgo.TripleDesEde3Crc:
						algorithm = BaseAlgorithm.TripleDes;
						isProfessional = true;
						keylen = 24;
						isDES = true;
						break;
					case CryptAlgo.Des :
						algorithm = BaseAlgorithm.Des;
						isProfessional = false;
						keylen = 8;
						isDES = true;
						break;
					case CryptAlgo.DesCrc :
						algorithm = BaseAlgorithm.Des;
						isProfessional = true;
						keylen = 8;
						isDES = true;
						break;
					case CryptAlgo.Aes128 :
						algorithm = BaseAlgorithm.Aes;
						isProfessional = true;
						keylen = 16;
						isDES = false;
						break;
					case CryptAlgo.Aes192 :
						algorithm = BaseAlgorithm.Aes;
						isProfessional = true;
						keylen = 24;
						isDES = false;
						break;
					case CryptAlgo.Aes256 :
						algorithm = BaseAlgorithm.Aes;
						isProfessional = true;
						keylen = 32;
						isDES = false;
						break;
					default :
						throw new CardServiceException(
							StatusWord.swUnknownAlgorithm);
				}
				isProfessional = forProfessional;

				byte[] keytemp = keydata;
				if (keytemp.Length > keylen)
				{
					byte[] fitkey = new byte[keylen];
					System.Array.Copy(keytemp, 0, fitkey, 0, keylen);
					keytemp = fitkey;
				}
				else if (keytemp.Length < keylen)
				{
					throw new CardServiceException(StatusWord.swKeyTooShort);
				}
				mcrypt = new MessageCrypt(algorithm, keytemp);
				int blocklen = mcrypt.BlockLen;
				if ((RA.Length < (blocklen / 2)) || (RB.Length < (blocklen / 2)))
				{
					throw new CardServiceException(StatusWord.swLcLeError);
				}
				byte[] iv = new byte[blocklen];
				System.Array.Copy(RA, 0, iv, 0, blocklen / 4);
				System.Array.Copy(RB, 0, iv, blocklen / 4, blocklen / 2);
				System.Array.Copy(
					RA,
					blocklen / 4,
					iv,
					blocklen / 4 * 3,
					blocklen / 4);
				mcrypt.SetIV(iv, 0);
			}

			/// <summary>
			/// See <see cref="ICommandAndResponseCrypt.Reset"/>.
			/// </summary>
			public void Reset()
			{
				service.OnResetEncryption();
			}

			/// <summary>
			/// See <see cref="ICommandAndResponseCrypt.SetProtocol"/>.
			/// </summary>
			public void SetProtocol(int prot)
			{
				protocol = prot;
			}

			/// <summary>
			/// See <see cref="ICommandAndResponseCrypt.EncryptCommand"/>.
			/// </summary>
			public CommandApdu EncryptCommand(CommandApdu cmd)
			{
				if (!encrypt)
					return cmd;
                cmd = service.MakeExtended(cmd, mcrypt.BlockLen);
				if (isProfessional)
				{
					PackCryptBufferProfessional(cmd);
				}
				else
				{
					PackCryptBufferEnhanced(cmd);
				}
				try
				{
					mcrypt.Encrypt(cryptBuffer, 0, lenCryptBuffer);
					CommandApdu newcmd =
						new CommandApdu(
						cmd.CLA,
						cmd.INS,
						cmd.P1,
						cmd.P2);
                    newcmd.DoNotLog = cmd.DoNotLog;
                    if (cmd.UseExtendedLC || cmd.UseExtendedLE || 
                        (service.AllowExtendedLength && (lenCryptBuffer>255)))
                    {
                        newcmd.UseExtendedLength=true;
                    }
					newcmd.Append(cryptBuffer, 0, lenCryptBuffer);
					if (isProfessional)
					{
						int LE = cmd.LE;
						if (cmd.IsLEDisabled)
						{                            
							LE = mcrypt.BlockLen;
                            newcmd.LE=LE;
						}
						else
						{
							if (0 != LE)
							{
								LE += mcrypt.BlockLen;
							}
                            newcmd.LE=LE;
						}						
					}
					else
					{
						newcmd.LE=cmd.LE;
					}                    
					return newcmd;
				}
				catch (Exception e)
				{
					CardServiceException ce =
						new CardServiceException(StatusWord.swInternalError, e);
					throw ce;
				}
			}

			/// <summary>
			/// See <see cref="ICommandAndResponseCrypt.DecryptResponse"/>.
			/// </summary>
			public ResponseApdu DecryptResponse(ResponseApdu rsp)
			{
				if (!decrypt)
					return rsp;
				if (BasicCardResponse.IsSW1SW2Ok(rsp.SW1SW2))
				{
					byte[] data = rsp.Data;
					System.Array.Copy(data, 0, cryptBuffer, 0, data.Length);
					lenCryptBuffer = data.Length;
					try
					{
						if (0 == lenCryptBuffer)
							return rsp;
						ResponseApdu newrsp;
						mcrypt.Decrypt(cryptBuffer, 0, lenCryptBuffer);
						if (isProfessional)
						{
							newrsp = UnpackCryptBufferProfessional();
						}
						else
						{
							newrsp = UnpackCryptBufferEnhanced();
						}
                        newrsp.DoNotLog = rsp.DoNotLog;
						return newrsp;
					}
					catch (CommsException ce)
					{
						throw ce;
					}
					catch (Exception e)
					{
						StatusWord sw;
						if (isDES)
						{
							sw = StatusWord.swBadDesResponse;
						}
						else
						{
							sw = StatusWord.swBadAesResponse;
						}
						CardServiceException ce = new CardServiceException(sw, e);
						throw ce;
					}
				}
				else
				{
					return rsp;
				}
			}

			private void PackCryptBufferEnhanced(CommandApdu cmd)
			{
				lenCryptBuffer = 5 + (cmd.LC & 0xFF);
				cryptBuffer[0] = cmd.CLA;
				cryptBuffer[1] = cmd.INS;
				cryptBuffer[2] = cmd.P1;
				cryptBuffer[3] = cmd.P2;
				cryptBuffer[4] = cmd.LC;
				cmd.CopyIData(cryptBuffer, 5);
				int LE = (cmd.IsLEDisabled ? 0 : cmd.LE);
				cryptBuffer[lenCryptBuffer++] = (byte) LE;
				cryptBuffer[lenCryptBuffer++] = (byte) 0;
				cryptBuffer[lenCryptBuffer++] = (byte) 0;
			}            

			private void PackCryptBufferProfessional(CommandApdu cmd)
			{                
                // CRC = CRC32(CLA || INS || P1 || P2 || [LC] || [IDATA] || LeFlag || LE')
				Crc32 crcobj = new Crc32();                
				crcobj.Update(cmd.CLA); // CLA
				crcobj.Update(cmd.INS); // INS
				crcobj.Update(cmd.P1);  // P1
				crcobj.Update(cmd.P2);  // P2
                crcobj.Update(cmd.LcField); // [LC]
                // [IDATA]                
                cmd.CopyIData(cryptBuffer, 0); // copy to crypt buffer for later use
                lenCryptBuffer = cmd.ExtendedLC;
                crcobj.Update(cryptBuffer, 0, lenCryptBuffer);
                // LeFlag 1 if LE present, 0 otherwise
                bool lePresent = !cmd.IsLEDisabled;
                int LeFlag = (lePresent ? 1 : 0);
                crcobj.Update((byte)LeFlag);
                // LE' if not LE present set to 0
                // LE' is either 1 byte (standard length)
                // or 2 bytes (extended length)
                byte[] LEi = MakeLeField(cmd); // LE'
                crcobj.Update(LEi);

                // C = Messageencryption(P)
                // P = IData || PTail
                // Ptail = LeFlag || LE' || 00 .. 00 || CRC
                // Length of Ptail = BlockLen of encryption used
                // LE' = 1 byte for standard length
                //       2 byte for extended length
				// Number of zeros varies to match length of PTail
                // CRC is a 4 byte value calculated as described above
                // add real LC field to CRC                		
				
                // [IDATA] is already in cryptBuffer
                // LeFlag
				cryptBuffer[lenCryptBuffer++] = (byte) LeFlag;
                // LE'
                foreach (byte b in LEi) cryptBuffer[lenCryptBuffer++] = b;
                // 00 .. 00
                int zeros = mcrypt.BlockLen - 1 /* LeFlag */ - LEi.Length - 4 /* crc */;
				for (int i = 0; i < zeros; i++)
				{
					cryptBuffer[lenCryptBuffer++] = (byte) 0;
				}
                // CRC
                int crc = crcobj.CRC;
				cryptBuffer[lenCryptBuffer++] = (byte) ((crc >> 24) & 0xFF);
				cryptBuffer[lenCryptBuffer++] = (byte) ((crc >> 16) & 0xFF);
				cryptBuffer[lenCryptBuffer++] = (byte) ((crc >> 8) & 0xFF);
				cryptBuffer[lenCryptBuffer++] = (byte) (crc & 0xFF);
			}

			private ResponseApdu UnpackCryptBufferEnhanced()
			{
				if ((lenCryptBuffer < 8)
					|| (0 != cryptBuffer[lenCryptBuffer - 1])
					|| (0 != cryptBuffer[lenCryptBuffer - 2])
					|| (0 != cryptBuffer[lenCryptBuffer - 3])
					|| (0 != cryptBuffer[lenCryptBuffer - 4])
					|| (0 != cryptBuffer[lenCryptBuffer - 5])
					|| (0 != cryptBuffer[lenCryptBuffer - 6]))
				{
					throw new CardServiceException(StatusWord.swBadDesResponse);
				}
				ResponseApdu rsp =
					new ResponseApdu(cryptBuffer, lenCryptBuffer - 6);
				return rsp;
			}

			private ResponseApdu UnpackCryptBufferProfessional()
			{
				int zeros = (mcrypt.BlockLen == 8 ? 2 : 10);
				int lendata = lenCryptBuffer - zeros - 4;
				int crcpos = lenCryptBuffer - 4;
				int crc, crcverify;
				StatusWord sw;
				CryptAlgo algo=(CryptAlgo) service.algo;
				if ((CryptAlgo.Aes128 == algo) ||
					(CryptAlgo.Aes192 == algo) ||
					(CryptAlgo.Aes256 == algo))
				{
					sw = StatusWord.swBadAesResponse;
				}
				else
				{
					sw = StatusWord.swBadDesResponse;
				}
				if (lendata < 0)
				{
					throw new CardServiceException(sw);
				}
				crcverify = Crc32.Calculate(cryptBuffer, 0, lendata);
				crc =
					((cryptBuffer[crcpos] & 0xFF) << 24)
					+ ((cryptBuffer[crcpos + 1] & 0xFF) << 16)
					+ ((cryptBuffer[crcpos + 2] & 0xFF) << 8)
					+ ((cryptBuffer[crcpos + 3] & 0xFF));
				if (crc != crcverify)
				{
					throw new CardServiceException(sw);
				}
				for (int i = lendata; i < lendata + zeros; i++)
				{
					if (0 != cryptBuffer[i])
					{
						throw new CardServiceException(sw);
					}
				}
				ResponseApdu rsp = new ResponseApdu(cryptBuffer, lendata);
				return rsp;
			}

			/// <summary>
			/// See <see cref="ICommandAndResponseCrypt.SetCryptOptions"/>.
			/// </summary>
			public void SetCryptOptions(
				bool encryptCommand,
				bool decryptResponse)
			{
				encrypt = encryptCommand;
				decrypt = decryptResponse;
			}

		}

		/// <summary>
		/// CommandAndResponseCrypt for EAX.
		/// </summary>
		private class CryptEax : ICommandAndResponseCrypt
		{
			private int protocol;
			private bool encrypt = true;
			private bool decrypt = true;
			private Eax eax;
			private byte[] cryptBuffer = new byte[8*1024];
			private AbstractBasicCardService service;

			/// <summary>
			/// Construct CommandAndResponseCrypt object for EAX
			/// </summary>
			/// <param name="service">AbstractBasicCardService using this object</param>
			/// <param name="keydata">encryption key</param>
			/// <param name="algo">algorithm</param>
			/// <param name="RA">RA from start encryption command</param>
			/// <param name="RB">RB from start encryption response</param>
			public CryptEax(AbstractBasicCardService service, byte[] keydata, CryptAlgo algo, byte[] RA, byte[] RB)
			{
				this.service=service;
				BaseAlgorithm algorithm;
				int keylen;
				switch (algo)
				{
					case CryptAlgo.EaxAes128 :
						algorithm = BaseAlgorithm.Aes;
						keylen = 16;
						break;
					case CryptAlgo.EaxAes192 :
						algorithm = BaseAlgorithm.Aes;
						keylen = 24;
						break;
					case CryptAlgo.EaxAes256 :
						algorithm = BaseAlgorithm.Aes;
						keylen = 32;
						break;
					default :
						throw new CardServiceException(
							StatusWord.swUnknownAlgorithm);
				}
				if ((RA.Length < 8) || (RB.Length < 8))
				{
					throw new CardServiceException(StatusWord.swLcLeError);
				}
				byte[] keytemp = keydata;
				if (keytemp.Length > keylen)
				{
					byte[] fitkey = new byte[keylen];
					System.Array.Copy(keytemp, 0, fitkey, 0, keylen);
					keytemp = fitkey;
				}
				else if (keytemp.Length < keylen)
				{
					throw new CardServiceException(StatusWord.swKeyTooShort);
				}
				eax = new Eax(algorithm, keytemp);
				byte[] Nonce = new byte[16];
				System.Array.Copy(RA, 0, Nonce, 0, 4);
				System.Array.Copy(RB, 0, Nonce, 4, 8);
				System.Array.Copy(RA, 4, Nonce, 12, 4);
				eax.ProvideNonce(Nonce);
			}

			/// <summary>
			/// See <see cref="ICommandAndResponseCrypt.Reset"/>.
			/// </summary>
			public void Reset()
			{
				service.OnResetEncryption();
			}

			/// <summary>
			/// See <see cref="ICommandAndResponseCrypt.SetProtocol"/>.
			/// </summary>
			public void SetProtocol(int prot)
			{
				protocol = prot;
			}

			/// <summary>
			/// See <see cref="ICommandAndResponseCrypt.EncryptCommand"/>.
			/// </summary>
			public CommandApdu EncryptCommand(CommandApdu cmd)
			{
				if (!encrypt)
					return cmd;
                cmd = service.MakeExtended(cmd, 16);
                CommandApdu tempcmd = (CommandApdu)cmd.Clone();
                CommandApdu newcmd = null;
				if (tempcmd.IsLEDisabled)
				{
					tempcmd.LE=16;                                       
				}
				else if (tempcmd.LE == 0)
				{
					tempcmd.LE=0;
				}
				else
				{
					tempcmd.LE=cmd.LE + 16;
				}
				try
				{
                    int LC=cmd.ExtendedLC;
					eax.ProvideHeader(tempcmd.CLA);
					eax.ProvideHeader(tempcmd.INS);
					eax.ProvideHeader(tempcmd.P1);
					eax.ProvideHeader(tempcmd.P2);
                    // LC += 16;
                    tempcmd.Append(new byte[16]);                                        					
                    byte[] lcField = tempcmd.LcField;                    
                    eax.ProvideHeader(lcField);                    
                    byte[] leField = tempcmd.LeField;
                    if (leField.Length > 2)
                    {
                        byte[] temp = new byte[2];
                        Array.Copy(leField, leField.Length - 2, temp, 0, 2);
                        leField = temp;
                    }
                    eax.ProvideHeader(leField);
                    // encrypt IData
					cmd.CopyIData(cryptBuffer, 0);
					eax.ComputeCiphertext(cryptBuffer, 0, LC);                    
					// compute and add tag
					byte[] tag = eax.ComputeTag();
					eax.ProvideNonce(tag);
					if (protocol == 0)
					{
                        int tagPos = 15;
                        for (int i = leField.Length - 1; i >= 0; i--)
                        {
                            tag[tagPos--] = leField[i];
                        }
					}
                    // build new command
                    newcmd = new CommandApdu(
                        cmd.CLA,
                        cmd.INS,
                        cmd.P1,
                        cmd.P2);
                    newcmd.UseExtendedLength = tempcmd.UseExtendedLC || tempcmd.UseExtendedLE;                    
                    newcmd.Append(cryptBuffer, 0, LC);
					newcmd.Append(tag);
                    newcmd.LE = tempcmd.LE;
                    newcmd.DoNotLog = cmd.DoNotLog;
				}
				catch (Exception e)
				{
					CardServiceException ce =
						new CardServiceException(StatusWord.swInternalError, e);
					throw ce;
				}
				return newcmd;
			}

			/// <summary>
			/// See <see cref="ICommandAndResponseCrypt.DecryptResponse"/>.
			/// </summary>
			public ResponseApdu DecryptResponse(ResponseApdu rsp)
			{
				if (!decrypt)
					return rsp;
				if (BasicCardResponse.IsSW1SW2Ok(rsp.SW1SW2))
				{
					ResponseApdu newrsp;
					int lenDecrypt;
					byte[] data = rsp.Data;
					System.Array.Copy(data, 0, cryptBuffer, 0, data.Length);
					lenDecrypt = data.Length;
					if (lenDecrypt < 16)
					{
						throw new CardServiceException(
							StatusWord.swResponseTooShort);
					}
					try
					{
						byte SW1 = (byte) rsp.SW1;
						byte SW2 = (byte) rsp.SW2;
						eax.ProvideHeader(SW1);
						eax.ProvideHeader(SW2);
						lenDecrypt -= 16;
						eax.ComputePlaintext(cryptBuffer, 0, lenDecrypt);
						byte[] tag = eax.ComputeTag();
						if (!Eax.CompareTag(tag, cryptBuffer, lenDecrypt, 16))
						{
							throw new CardServiceException(
								StatusWord.swBadAesResponse);
						}
						eax.ProvideNonce(tag);
						cryptBuffer[lenDecrypt++] = SW1;
						cryptBuffer[lenDecrypt++] = SW2;
						newrsp = new ResponseApdu(cryptBuffer, lenDecrypt);
                        newrsp.DoNotLog = rsp.DoNotLog;
					}
					catch (CommsException e)
					{
						throw e;
					}
					catch (Exception e)
					{
						CardServiceException ce =
							new CardServiceException(
							StatusWord.swInternalError, e);
						throw ce;
					}
					return newrsp;
				}
				else
				{
					return rsp;
				}
			}

			/// <summary>
			/// See <see cref="ICommandAndResponseCrypt.SetCryptOptions"/>.
			/// </summary>
			public void SetCryptOptions(
				bool encryptCommand,
				bool decryptResponse)
			{
				encrypt = encryptCommand;
				decrypt = decryptResponse;
			}
		}


		/// <summary>
		/// CommandAndResponseCrypt for OMAC.
		/// </summary>
		private class CryptOmac : ICommandAndResponseCrypt
		{
			private int protocol;
			private bool encrypt = true;
			private bool decrypt = true;
			private OmacFactory.Omac1 omac;
			private byte[] cryptBuffer = new byte[8*1024];
			private AbstractBasicCardService service;

			/// <summary>
			/// Construct CommandAndResponseCrypt object for OMAC
			/// </summary>
			/// <param name="service">AbstractBasicCardService using this object</param>
			/// <param name="keydata">encryption key</param>
			/// <param name="algo">algorithm</param>
			public CryptOmac(AbstractBasicCardService service, byte[] keydata, CryptAlgo algo)
			{
				this.service=service;
				BaseAlgorithm algorithm;
				int keylen;
				switch (algo)
				{
					case CryptAlgo.OmacAes128:
						algorithm = BaseAlgorithm.Aes;
						keylen = 16;
						break;
					case CryptAlgo.OmacAes192:
						algorithm = BaseAlgorithm.Aes;
						keylen = 24;
						break;
					case CryptAlgo.OmacAes256:
						algorithm = BaseAlgorithm.Aes;
						keylen = 32;
						break;
					default :
						throw new CardServiceException(
							StatusWord.swUnknownAlgorithm);
				}
				byte[] keytemp = keydata;
				if (keytemp.Length > keylen)
				{
					byte[] fitkey = new byte[keylen];
					System.Array.Copy(keytemp, 0, fitkey, 0, keylen);
					keytemp = fitkey;
				}
				else if (keytemp.Length < keylen)
				{
					throw new CardServiceException(StatusWord.swKeyTooShort);
				}
				OmacFactory omacfactory = new OmacFactory(algorithm, keytemp);
				omac = omacfactory.CreateOmac1();
			}

			/// <summary>
			/// See <see cref="ICommandAndResponseCrypt.Reset"/>.
			/// </summary>
			public void Reset()
			{
				service.OnResetEncryption();
			}

			/// <summary>
			/// See <see cref="ICommandAndResponseCrypt.SetProtocol"/>.
			/// </summary>
			public void SetProtocol(int prot)
			{
				protocol = prot;
			}

			/// <summary>
			/// See <see cref="ICommandAndResponseCrypt.EncryptCommand"/>.
			/// </summary>
			public CommandApdu EncryptCommand(CommandApdu cmd)
			{
				if (!encrypt)
					return cmd;
				CommandApdu newcmd;
                cmd = service.MakeExtended(cmd, 16);
				newcmd = (CommandApdu) cmd.Clone();
				if (cmd.IsLEDisabled)
				{
					newcmd.LE=16;                    
				}
				else if (cmd.LE == 0)
				{
					newcmd.LE=0;
				}
				else
				{
					newcmd.LE=cmd.LE + 16;
				}
				try
				{
                    CommandApdu tempcmd = (CommandApdu)newcmd.Clone();
                    // LC + 16
                    tempcmd.Append(new byte[16]);					
                    byte[] lcField = tempcmd.LcField;					
                    byte[] leField = tempcmd.LeField;
                    if (leField.Length > 2)
                    {
                        byte[] temp = new byte[2];
                        Array.Copy(leField, leField.Length - 2, temp, 0, 2);
                        leField = temp;
                    }
					omac.Start();
					omac.Append(newcmd.CLA);
					omac.Append(newcmd.INS);
					omac.Append(newcmd.P1);
					omac.Append(newcmd.P2);
					omac.Append(lcField);
                    cmd.CopyIData(cryptBuffer, 0);
                    omac.Append(cryptBuffer, 0, cmd.ExtendedLC);
					omac.Append(leField);
					byte[] tag = omac.End();
					if (protocol == 0)
					{
                        int tagPos = 15;
                        for (int i = leField.Length - 1; i >= 0; i--)
                        {
                            tag[tagPos--] = leField[i];
                        }
					}
					newcmd.Append(tag);                    
				}
				catch (Exception e)
				{
					CardServiceException ce =
						new CardServiceException(StatusWord.swInternalError, e);
					throw ce;
				}
				return newcmd;
			}

			/// <summary>
			/// See <see cref="ICommandAndResponseCrypt.DecryptResponse"/>.
			/// </summary>
			public ResponseApdu DecryptResponse(ResponseApdu rsp)
			{
				if (!decrypt)
					return rsp;
				if (BasicCardResponse.IsSW1SW2Ok(rsp.SW1SW2))
				{
					int dataLen;
					byte[] data = rsp.Data;
					dataLen = data.Length;
					if (dataLen < 16)
					{
						throw new CardServiceException(
							StatusWord.swResponseTooShort);
					}
					try
					{
						dataLen -= 16;
						byte SW1 = (byte) rsp.SW1;
						byte SW2 = (byte) rsp.SW2;
						omac.Start();
						omac.Append(data, 0, dataLen);
						omac.Append(SW1);
						omac.Append(SW2);
						byte[] tag = omac.End();
						if (!OmacFactory.CompareOmac(tag, data, dataLen, 16))
						{
							throw new CardServiceException(
								StatusWord.swBadAesResponse);
						}                        
						System.Array.Copy(data, 0, cryptBuffer, 0, dataLen);
						cryptBuffer[dataLen++] = SW1;
						cryptBuffer[dataLen++] = SW2;
						ResponseApdu newrsp =
							new ResponseApdu(cryptBuffer, dataLen);
                        newrsp.DoNotLog = rsp.DoNotLog;
						return newrsp;
					}
					catch (CommsException e)
					{
						throw e;
					}
					catch (Exception e)
					{
						CardServiceException ce =
							new CardServiceException(
							StatusWord.swInternalError, e);
						throw ce;
					}
				}
				else
				{
					return rsp;
				}
			}

			/// <summary>
			/// See <see cref="ICommandAndResponseCrypt.SetCryptOptions"/>.
			/// </summary>
			public void SetCryptOptions(
				bool encryptCommand,
				bool decryptResponse)
			{
				encrypt = encryptCommand;
				decrypt = decryptResponse;
			}

		}

		/// <summary>
		/// BasicCard CLA
		/// </summary>
		protected const byte ClaBasicCard = (byte) 0xC0;
		/// <summary>
		/// GET STATE INS
		/// </summary>
		protected const byte InsGetState = (byte) 0x00;
		/// <summary>
		/// EEPROM SIZE INS
		/// </summary>
		protected const byte InsEepromSize = (byte) 0x02;
		/// <summary>
		/// CLEAR EEPROM INS
		/// </summary>
		protected const byte InsClearEeprom = (byte) 0x04;
		/// <summary>
		/// WRITE EEPROM INS
		/// </summary>
		protected const byte InsWriteEeprom = (byte) 0x06;
		/// <summary>
		/// READ EEPROM INS
		/// </summary>
		protected const byte InsReadEeprom = (byte) 0x08;
		/// <summary>
		/// EEPROM CRC INS
		/// </summary>
		protected const byte InsEepromCrc = (byte) 0x0A;
		/// <summary>
		/// SET STATE INS
		/// </summary>
		protected const byte InsSetState = (byte) 0x0C;
		/// <summary>
		/// GET APPLICATION ID INS
		/// </summary>
		protected const byte InsGetApplicationID = (byte) 0x0E;
		/// <summary>
		/// START ENCRYPTION INS
		/// </summary>
		protected const byte InsStartEncryption = (byte) 0x10;
		/// <summary>
		/// END ENCRYPTION INS
		/// </summary>
		protected const byte InsEndEncryption = (byte) 0x12;
		/// <summary>
		/// ECHO INS
		/// </summary>
		protected const byte InsEcho = (byte) 0x14;
		/// <summary>
		/// ASSIGN NAD INS
		/// </summary>
		protected const byte InsAssignNad = (byte) 0x16;
		/// <summary>
		/// FILE IO INS
		/// </summary>
		protected const byte InsFileIo = (byte) 0x18;
		/// <summary>
		/// GET CHALLENGE INS
		/// </summary>
		protected const byte InsGetChallenge = (byte) 0x40;
		/// <summary>
		/// EXTERNAL AUTHENTICATE INS
		/// </summary>
		protected const byte InsExternalAuthenticate = (byte) 0x42;
		/// <summary>
		/// INTERNAL AUTHENTICATE INS
		/// </summary>
		protected const byte InsInternalAuthenticate = (byte) 0x44;
		/// <summary>
		/// VERIFY INS
		/// </summary>
		protected const byte InsVerify = (byte) 0x46;
		/// <summary>
		/// GET FREE MEMORY INS
		/// </summary>
		protected const byte InsGetFreeMemory = (byte) 0x48;
		/// <summary>
		/// SELLECT APPLICATION INS
		/// </summary>
		protected const byte InsSelectApplication = (byte) 0xA0;
		/// <summary>
		/// CREATE COMPONENT INS
		/// </summary>
		protected const byte InsCreateComponent = (byte) 0xA2;
		/// <summary>
		/// DELETE COMPONENT INS
		/// </summary>
		protected const byte InsDeleteComponent = (byte) 0xA4;
		/// <summary>
		/// WRITE COMPONENT ATTR INS
		/// </summary>
		protected const byte InsWriteComponentAttr = (byte) 0xA6;
		/// <summary>
		/// READ COMPONENT ATTR INS
		/// </summary>
		protected const byte InsReadComponentAttr = (byte) 0xA8;
		/// <summary>
		/// WRITE COMPONENT DATA INS
		/// </summary>
		protected const byte InsWriteComponentData = (byte) 0xAA;
		/// <summary>
		/// READ COMPONENT DATA INS
		/// </summary>
		protected const byte InsReadComponentData = (byte) 0xAC;
		/// <summary>
		/// FIND COMPONENT INS
		/// </summary>
		protected const byte InsFindComponent = (byte) 0xAE;
		/// <summary>
		/// COMPONENT NAME INS
		/// </summary>
		protected const byte InsComponentName = (byte) 0xB0;
		/// <summary>
		/// GRANT PRIVILEGE
		/// </summary>
		protected const byte InsGrantPrivilege = (byte) 0xB2;
		/// <summary>
		/// AUTHENTICATE FILE INS
		/// </summary>
		protected const byte InsAuthenticateFile = (byte) 0xB4;
		/// <summary>
		/// READ RIGHTS LIST INS
		/// </summary>
		protected const byte InsReadRightsList = (byte) 0xB6;
		/// <summary>
		/// LOAD SEQUENCE INS
		/// </summary>
		protected const byte InsLoadSequence = (byte) 0xB8;
		/// <summary>
		/// SECURE TRANSPORT INS
		/// </summary>
		protected const byte InsSecureTransport = (byte) 0xBA;

		/// <summary>
		/// SW1 value for LEWarning (61h)
		/// </summary>
		protected const byte sw1LeWarning = (byte) (((int) StatusWord.swLeWarning)>>8);

		/// <summary>
		/// Connection to processor card
		/// </summary>
		protected IProcessorCardConnection pconn;
		/// <summary>
		/// secure random number generator
		/// </summary>
		protected RandomNumberGenerator	random = RandomNumberGenerator.Create();
		private ICommandAndResponseCrypt crypt;
		private BCKeyStore keys = new BCKeyStore();
		private CryptAlgo algo = CryptAlgo.None;
		private int key = -1;
        private bool kdpCrypto = false;
		private BasicCardFileIO fileio=null;
		private BasicCardString applicationFile = null;
        private bool allowExtendedLength = false;

        /// <summary>
        /// Allow using extended length APDU?
        /// </summary>
        public bool AllowExtendedLength
        {
            get { return allowExtendedLength; }
            set { allowExtendedLength = value; }
        }

		/// <overloads>
		/// Constructor.
		/// </overloads>
		/// <summary>
		/// Default constructor, using a BasicCardSelector which may be configured 
        /// within child constructor. By default this card selector accepts all processor smart cards.
		/// </summary>        
		public AbstractBasicCardService()
		{
			this.SetCardSelector(new BasicCardSelector(this));
		}

		/// <summary>
		/// Constructs a AbstractBasicCardService object which uses the specified card 
		/// selector for card acceptance check.
		/// </summary>
		/// <param name="selector">Card selector to use
		/// a exclusive connection is required.</param>
		public AbstractBasicCardService(ICardSelector selector)
		{
			this.SetCardSelector(selector);
		}

		/// <summary>
		/// Constructs a AbstractBasicCardService object which uses a BasicCardSelector 
		/// to select cards with specified application id.
		/// </summary>
		/// <param name="applicationID">Application id.</param>
		public AbstractBasicCardService(BasicCardString applicationID)
			: this()
		{
			((BasicCardSelector) selector).ApplicationID=applicationID;
		}

		/// <summary>
		/// Constructs a AbstractBasicCardService object which uses a BasicCardSelector 
		/// to select cards with specified application id.
		/// </summary>
		/// <param name="applicationID">Application id.</param>
		public AbstractBasicCardService(String applicationID)
			: this()
		{
			((BasicCardSelector) selector).ApplicationID=
				new BasicCardString(
					applicationID,
					BasicCardString.CharsetIdAscii);
		}

        /// <summary>
        /// Constructs a AbstractBasicCardService object which uses a BasicCardSelector 
        /// to select cards with specified application id and specified ATR history.
        /// </summary>
        /// <param name="applicationID">Application id or null if no specific application id is required.</param>
        /// <param name="atrHistory">ATR history id or null if no specific history is required.</param>
        public AbstractBasicCardService(String applicationID, String atrHistory)
            : this()
        {
            BasicCardSelector bcs = (BasicCardSelector)selector;
            if (null != applicationID)
            {
                bcs.ApplicationID = new BasicCardString(
                    applicationID,
                    BasicCardString.CharsetIdAscii);
            }
            if (null != atrHistory)
            {
                bcs.AtrHistory = new BasicCardString(
                    atrHistory,
                    BasicCardString.CharsetIdAscii);
            }
        }

        /// <summary>
        /// Constructs a AbstractBasicCardService object which uses a BasicCardSelector 
        /// to select cards with specified application id and specified ATR history.
        /// </summary>
        /// <param name="applicationID">Application id or null if no specific application id is required.</param>
        /// <param name="atrHistory">ATR history id or null if no specific history is required.</param>
        public AbstractBasicCardService(BasicCardString applicationID, BasicCardString atrHistory)
            : this()
        {
            BasicCardSelector bcs = (BasicCardSelector)selector;
            if (null != applicationID)
            {
                bcs.ApplicationID = applicationID;
            }
            if (null != atrHistory)
            {
                bcs.AtrHistory = atrHistory;
            }
        }

		/// <summary>
		/// Init service by assigning specified connection.
		/// </summary>
		/// <param name="conn">Connection to use.</param>
		public override void Init(ICardConnection conn)
		{
			base.Init (conn);
			if (typeof(IProcessorCardConnection).IsInstanceOfType(conn))
			{
				pconn = (IProcessorCardConnection) conn;
			}
			else
			{
				throw new CommsException(StatusWord.swInvalidConnection);
			}
			if (null != applicationFile)
			{
				if (!IsExclusive)
				{
					throw new CommsException(StatusWord.swInvalidConnection);
				}
				CmdSelectApplication(applicationFile);
			}
		}

		/// <summary>
		/// Reset the inserted card.
		/// </summary>
		/// <returns>ATR</returns>
		protected internal virtual ProcessorCardAtr ResetCard()
		{
			pconn.Reset();
			return (ProcessorCardAtr) pconn.Atr;
		}

		/// <summary>
		/// Call a BasicCard command.
		/// </summary>
		/// <remarks>
		/// This function does not check SW1SW2. Use <see cref="BasicCardResponse.CheckSW1SW2">BasicCardResponse.CheckSW1SW2</see> to check SW1SW2 of response.
		/// </remarks>
		/// <param name="cmd">BasicCard command.</param>
		/// <returns>BasicCard response.</returns>
		protected internal virtual BasicCardResponse DoCommandAndResponse(BasicCardCommand cmd)
		{
			return new BasicCardResponse(pconn.DoCommandAndResponse(cmd));
		}

		/// <summary>
		/// Call BasicCard GET STATE command. See BasicCard manual for details.
		/// </summary>
		/// <returns>GET STATE output. State byte followed by version bytes.</returns>
		protected internal virtual byte[] CmdGetState()
		{
			BasicCardCommand cmd =
				new BasicCardCommand(ClaBasicCard, InsGetState);
			cmd.LE=0;
			BasicCardResponse rsp = DoCommandAndResponse(cmd);
			rsp.CheckSW1SW2();
			if (rsp.Length < 1)
			{
				throw new CardServiceException(StatusWord.swResponseTooShort);
			}
			return rsp.Data;
		}

		/// <summary>
		/// Call BasicCard EEPROM SIZE command.
		/// </summary>
		/// <returns>Array with 2 int values. int[0]=eeprom start address; int[1]=eeprom length.</returns>
		protected internal virtual int[] CmdEepromSize()
		{
			BasicCardCommand cmd =
				new BasicCardCommand(ClaBasicCard, InsEepromSize);
			cmd.LE=4;
			BasicCardResponse rsp = DoCommandAndResponse(cmd);
			rsp.CheckSW1SW2();
			if (rsp.Length < 4)
			{
				throw new CardServiceException(StatusWord.swResponseTooShort);
			}
			int[] ret = new int[2];
			ret[0] = rsp.FetchBasicInteger() & 0xFFFF;
			ret[1] = rsp.FetchBasicInteger() & 0xFFFF;
			return ret;
		}

        /// <summary>
        /// Call BasicCard EEPROM SIZE (24 bit variant) command.
        /// </summary>
        /// <returns>Array with 2 int values. int[0]=eeprom start address; int[1]=eeprom length.</returns>
        protected internal virtual int[] CmdEeprom24Size()
        {
            BasicCardCommand cmd =
                new BasicCardCommand(ClaBasicCard, InsEepromSize);
            cmd.LE = 6;
            BasicCardResponse rsp = DoCommandAndResponse(cmd);
            rsp.CheckSW1SW2();
            if (rsp.Length < 4)
            {
                throw new CardServiceException(StatusWord.swResponseTooShort);
            }
            int[] ret = new int[2];
            ret[0] = rsp.Fetch24BitInteger() & 0xFFFFFF;
            ret[1] = rsp.Fetch24BitInteger() & 0xFFFFFF;
            return ret;
        }

        /// <summary>
        /// Call BasicCard EEPROM SIZE (32 bit variant) command.
        /// </summary>
        /// <returns>Array with 2 int values. int[0]=eeprom start address; int[1]=eeprom length.</returns>
        protected internal virtual int[] CmdEeprom32Size()
        {
            BasicCardCommand cmd =
                new BasicCardCommand(ClaBasicCard, InsEepromSize);
            cmd.LE = 8;
            BasicCardResponse rsp = DoCommandAndResponse(cmd);
            rsp.CheckSW1SW2();
            if (rsp.Length < 4)
            {
                throw new CardServiceException(StatusWord.swResponseTooShort);
            }
            int[] ret = new int[2];
            ret[0] = rsp.FetchBasicLong();
            ret[1] = rsp.FetchBasicLong();
            return ret;
        }

		/// <summary>
		/// Call BasicCard CLEAR EEPROM command.
		/// </summary>
		/// <remarks>
		/// This command is for loading application images into a BasicCard. It 
		/// is not recommended to use this command on application level. Using this
		/// command may destroy the BasicCard.
		/// </remarks>
		/// <param name="startAddress">Start address for clear operation.</param>
		/// <param name="length">Number of bytes to erase.</param>
		protected internal virtual void CmdClearEeprom(int startAddress, int length)
		{
			if ((0 != (startAddress & 0xFFFF0000)) || (0 != (length & 0xFFFF0000)))
			{
				throw new CardServiceException(StatusWord.swOutsideEeprom);
			}
			BasicCardCommand cmd =
				new BasicCardCommand(ClaBasicCard, InsClearEeprom);
			cmd.P1P2=((short) startAddress);
			cmd.AppendBasicInteger((short) length);
			cmd.DisableLE();
			BasicCardResponse rsp = DoCommandAndResponse(cmd);
			rsp.CheckSW1SW2();
		}

        /// <summary>
        /// Call BasicCard CLEAR EEPROM (24 bit address variant) command.
        /// </summary>
        /// <remarks>
        /// This command is for loading application images into a BasicCard. It 
        /// is not recommended to use this command on application level. Using this
        /// command may destroy the BasicCard.
        /// </remarks>
        /// <param name="startAddress">Start address for clear operation.</param>
        /// <param name="length">Number of bytes to erase.</param>
        protected internal virtual void CmdClearEeprom24(int startAddress, int length)
        {
            if ((0 != (startAddress & 0xFF000000)) || (0 != (length & 0xFF000000)))
            {
                throw new CardServiceException(StatusWord.swOutsideEeprom);
            }
            BasicCardCommand cmd =
                new BasicCardCommand(ClaBasicCard, InsClearEeprom);
            cmd.Append24BitInteger(startAddress);
            cmd.Append24BitInteger(length);
            cmd.DisableLE();
            BasicCardResponse rsp = DoCommandAndResponse(cmd);
            rsp.CheckSW1SW2();
        }

        /// <summary>
        /// Call BasicCard CLEAR EEPROM (32 bit address variant) command.
        /// </summary>
        /// <remarks>
        /// This command is for loading application images into a BasicCard. It 
        /// is not recommended to use this command on application level. Using this
        /// command may destroy the BasicCard.
        /// </remarks>
        /// <param name="startAddress">Start address for clear operation.</param>
        /// <param name="length">Number of bytes to erase.</param>
        protected internal virtual void CmdClearEeprom32(int startAddress, int length)
        {
            if ((0 != (startAddress & 0xFF000000)) || (0 != (length & 0xFF000000)))
            {
                throw new CardServiceException(StatusWord.swOutsideEeprom);
            }
            BasicCardCommand cmd =
                new BasicCardCommand(ClaBasicCard, InsClearEeprom);
            cmd.AppendBasicLong(startAddress);
            cmd.AppendBasicLong(length);
            cmd.DisableLE();
            BasicCardResponse rsp = DoCommandAndResponse(cmd);
            rsp.CheckSW1SW2();
        }

		/// <summary>
		/// Call BasicCard WRITE EEPROM command..
		/// </summary>
		/// <remarks>
		/// This command is for loading application images into a BasicCard. It 
		/// is not recommended to use this command on application level. Using this
		/// command may destroy the BasicCard.
		/// </remarks>
		/// <param name="startAddress">Start address for write operation.</param>
		/// <param name="data">Data bytes to write.</param>
		protected internal virtual void CmdWriteEeprom(int startAddress, byte[] data)
		{
			if (   (0 != (startAddress & 0xFFFF0000))
				|| (0 != (data.Length & 0xFFFF0000)))
			{
				throw new CardServiceException(StatusWord.swOutsideEeprom);
			}
			BasicCardCommand cmd =
				new BasicCardCommand(ClaBasicCard, InsWriteEeprom);
            if (allowExtendedLength && (data.Length > 255)) cmd.UseExtendedLength = true;
			cmd.P1P2=((short) startAddress);
			cmd.AppendBasicString(data);
			cmd.DisableLE();
			BasicCardResponse rsp = DoCommandAndResponse(cmd);
			rsp.CheckSW1SW2();
		}

		/// <summary>
		/// Call BasicCard READ EEPROM command
		/// </summary>
		/// <param name="startAddress">Start address for write operation.</param>
		/// <param name="length">Number of bytes to read.</param>
		/// <returns>Bytes read from card.</returns>
		protected internal virtual byte[] CmdReadEeprom(int startAddress, int length)
		{
			if ((0 != (startAddress & 0xFFFF0000)) || (0 != (length & 0xFFFFFF00)))
			{
				throw new CardServiceException(StatusWord.swOutsideEeprom);
			}
			BasicCardCommand cmd =
				new BasicCardCommand(ClaBasicCard, InsReadEeprom);
            if (allowExtendedLength && (length > 255)) cmd.UseExtendedLength = true;
			cmd.P1P2=((short) startAddress);
			cmd.LE=length;
			BasicCardResponse rsp = DoCommandAndResponse(cmd);
			rsp.CheckSW1SW2();
			return rsp.Data;
		}

        /// <summary>
        /// Call BasicCard READ EEPROM (24 bit address variant) command
        /// </summary>
        /// <param name="startAddress">Start address for write operation.</param>
        /// <param name="length">Number of bytes to read.</param>
        /// <returns>Bytes read from card.</returns>
        protected internal virtual byte[] CmdReadEeprom24(int startAddress, int length)
        {
            if ((0 != (startAddress & 0xFFFF0000)) || (0 != (length & 0xFFFFFF00)))
            {
                throw new CardServiceException(StatusWord.swOutsideEeprom);
            }
            BasicCardCommand cmd =
                new BasicCardCommand(ClaBasicCard, InsReadEeprom);
            if (allowExtendedLength && ((length+3) > 255)) cmd.UseExtendedLength = true;
            cmd.Append24BitInteger(startAddress);
            cmd.LE = length+3;
            BasicCardResponse rsp = DoCommandAndResponse(cmd);
            rsp.CheckSW1SW2();
            rsp.Fetch24BitInteger();
            return rsp.FetchBasicString().Characters;
        }

		/// <summary>
		/// Call BasicCard EEPROM CRC command.
		/// </summary>
		/// <param name="startAddress">Start address for CRC operation.</param>
		/// <param name="length">Number of bytes to check.</param>
		/// <returns>CRC checksum.</returns>
		protected internal virtual int CmdEepromCrc(int startAddress, int length)
		{
			if ((0 != (startAddress & 0xFFFF0000)) || (0 != (length & 0xFFFF0000)))
			{
				throw new CardServiceException(StatusWord.swOutsideEeprom);
			}
			BasicCardCommand cmd =
				new BasicCardCommand(ClaBasicCard, InsEepromCrc);
			cmd.P1P2=((short) startAddress);
			cmd.AppendBasicInteger((short) length);
			cmd.SetDefaultLE();
			BasicCardResponse rsp = DoCommandAndResponse(cmd);
			rsp.CheckSW1SW2();
			return rsp.FetchBasicInteger() & 0x0FFFF;
		}

        /// <summary>
        /// Call BasicCard EEPROM CRC (24 bit address variant) command.
        /// </summary>
        /// <param name="startAddress">Start address for CRC operation.</param>
        /// <param name="length">Number of bytes to check.</param>
        /// <returns>CRC checksum.</returns>
        protected internal virtual int CmdEeprom24Crc(int startAddress, int length)
        {
            if ((0 != (startAddress & 0xFF000000)) || (0 != (length & 0xFF000000)))
            {
                throw new CardServiceException(StatusWord.swOutsideEeprom);
            }
            BasicCardCommand cmd =
                new BasicCardCommand(ClaBasicCard, InsEepromCrc);
            cmd.Append24BitInteger(startAddress);
            cmd.Append24BitInteger(length);
            cmd.LE=8;
            BasicCardResponse rsp = DoCommandAndResponse(cmd);
            rsp.CheckSW1SW2();
            rsp.Fetch24BitInteger();
            rsp.Fetch24BitInteger();
            return rsp.FetchBasicInteger() & 0x0FFFF;
        }

        /// <summary>
        /// Call BasicCard EEPROM CRC (32 bit address variant) command.
        /// </summary>
        /// <param name="startAddress">Start address for CRC operation.</param>
        /// <param name="length">Number of bytes to check.</param>
        /// <returns>CRC checksum.</returns>
        protected internal virtual int CmdEeprom32Crc(int startAddress, int length)
        {
            if ((0 != (startAddress & 0xFF000000)) || (0 != (length & 0xFF000000)))
            {
                throw new CardServiceException(StatusWord.swOutsideEeprom);
            }
            BasicCardCommand cmd =
                new BasicCardCommand(ClaBasicCard, InsEepromCrc);
            cmd.AppendBasicLong(startAddress);
            cmd.AppendBasicLong(length);
            cmd.LE = 10;
            BasicCardResponse rsp = DoCommandAndResponse(cmd);
            rsp.CheckSW1SW2();
            rsp.FetchBasicLong();
            rsp.FetchBasicLong();
            return rsp.FetchBasicInteger() & 0x0FFFF;
        }

        /// <summary>
        /// Call BasicCard EEPROM CRC (23 bit CRC, 24 bit address variant) command.
        /// </summary>
        /// <param name="startAddress">Start address for CRC operation.</param>
        /// <param name="length">Number of bytes to check.</param>
        /// <returns>CRC checksum.</returns>
        protected internal virtual int CmdEeprom24Crc32(int startAddress, int length)
        {
            if ((0 != (startAddress & 0xFFFF0000)) || (0 != (length & 0xFFFF0000)))
            {
                throw new CardServiceException(StatusWord.swOutsideEeprom);
            }
            BasicCardCommand cmd =
                new BasicCardCommand(ClaBasicCard, InsEepromCrc);
            cmd.Append24BitInteger(startAddress);
            cmd.Append24BitInteger(length);
            cmd.LE = 10;
            BasicCardResponse rsp = DoCommandAndResponse(cmd);
            rsp.CheckSW1SW2();
            rsp.Fetch24BitInteger();
            rsp.Fetch24BitInteger();
            return rsp.FetchBasicLong();
        }

		/// <summary>
		/// Call BasicCard SET STATE command.
		/// </summary>
		/// <remarks>
		/// This command is for loading application images into a BasicCard. It 
		/// is not recommended to use this command on application level. Using this
		/// command may destroy the BasicCard.
		/// </remarks>
		/// <param name="state">New card state.</param>
		protected internal virtual void CmdSetState(byte state)
		{
			BasicCardCommand cmd =
				new BasicCardCommand(ClaBasicCard, InsSetState);
			cmd.P1=state;
			cmd.DisableLE();
			BasicCardResponse rsp = DoCommandAndResponse(cmd);
			rsp.CheckSW1SW2();
		}

		/// <overloads>
		/// Call BasicCard GET APPLICATION ID.
		/// </overloads>
		/// <param name="index">Index or frequested application id.</param>
		/// <param name="type">Requested data type.</param>
		/// <returns>Application ID or application file name.</returns>
		protected internal virtual BasicCardString CmdGetApplicationID(int index, int type)
		{
			BasicCardCommand cmd =
				new BasicCardCommand(
				ClaBasicCard,
				InsGetApplicationID,
				(byte) index,
				(byte) type);
			cmd.LE=0;
			BasicCardResponse rsp = DoCommandAndResponse(cmd);
			rsp.CheckSW1SW2();
			return rsp.FetchBasicString();
		}

		/// <returns>Application ID.</returns>
		protected internal virtual BasicCardString CmdGetApplicationID()
		{
			return CmdGetApplicationID(0, 0);
		}

        [Flags]
        private enum StartEncMode
        {
            Default = 0,
            ProfessionalCard = 0x01,
            MultiAppCard = 0x02,
            KdpMode = 0x04,
            WithCardKdp = 0x08,
            WithTermKdp = 0x10,
        };

		/// <summary>
		/// Call BasicCard START ENCRYPTION command and setup encryption between card and BasicCardService
		/// </summary>
		/// <param name="algorithm">algorithm as described in BasicCard manual</param>
		/// <param name="key">key number or CID of key component</param>
		/// <param name="lenR">lenR</param>
		/// <param name="mode">Start encryption mode</param>
        /// <param name="termKdp">Terminal KDP in case of KDP based encryption. null if no KDP encryption.</param>
		private void IntStartEncryption(
			CryptAlgo algorithm,
			int key,
			int lenR,
			StartEncMode mode=StartEncMode.Default,
            byte[] termKdp=null)
		{
			if (((int)algorithm) > 255)
			{
				throw new CardServiceException(StatusWord.swUnknownAlgorithm);
			}
            byte[] keydata = (0 == (mode & StartEncMode.KdpMode) ? keys.GetKey(key) : keys.GetCustomerKey());
			if (null == keydata)
			{
                throw new CardServiceException(StatusWord.swKeyNotLoaded);
			}
			byte[] RA = new byte[lenR];
			random.GetBytes(RA);
			BasicCardCommand cmd =
				new BasicCardCommand(ClaBasicCard, InsStartEncryption);
            if (0 != (mode & StartEncMode.KdpMode))
            {
                cmd.P1 = ((byte)algorithm);
                cmd.P1 |= 8;
                cmd.P2 = ((byte)0xFC);
                if (0 != (mode & StartEncMode.WithCardKdp)) cmd.P2 |= 1;
                if (0 != (mode & StartEncMode.WithTermKdp)) cmd.P2 |= 2;
            }
			else if (0!=(mode & StartEncMode.MultiAppCard))
			{
				cmd.P1P2=((short) key);
				cmd.AppendBasicByte((byte) algorithm);
			}
			else
			{
				cmd.P1=((byte) algorithm);
				cmd.P2=((byte) key);
			}
            if (lenR>0) cmd.AppendBasicString(RA);
            if (0 != (mode & StartEncMode.WithTermKdp))
            {
                if (null == termKdp) throw new ArgumentNullException("termKdp");
                cmd.AppendBasicString(termKdp);
            }
            if (0 != (mode & (StartEncMode.MultiAppCard | StartEncMode.ProfessionalCard | StartEncMode.KdpMode)))
			{
				cmd.LE=0;
			}
			else
			{
				cmd.LE=lenR;
			}
			BasicCardResponse rsp = DoCommandAndResponse(cmd);
			rsp.CheckSW1SW2();
            if ((0 == (mode & (StartEncMode.MultiAppCard | StartEncMode.ProfessionalCard)))
				&& (sw1LeWarning == rsp.SW1)
				&& ((lenR + 1) == rsp.SW2))
			{
                mode |= StartEncMode.ProfessionalCard;
			}
			if (0 == algorithm)
			{
				if (0 != (mode & (StartEncMode.MultiAppCard | StartEncMode.ProfessionalCard)))
				{
					algorithm = (CryptAlgo) rsp.FetchBasicByte();
				}
				else
				{
					int keylen = GetKeyLength(key);
					if (keylen >= 16)
					{
						algorithm = CryptAlgo.TripleDes;
					}
					else
					{
						algorithm = CryptAlgo.Des;
					}
				}
			}
			else
			{
                if (0 != (mode & (StartEncMode.MultiAppCard | StartEncMode.ProfessionalCard | StartEncMode.KdpMode)))
				{
					int usedAlgo = rsp.FetchBasicByte() & 0xFF;
					if (usedAlgo != (int)algorithm)
					{
						throw new CardServiceException(
							StatusWord.swUnexpectedResponse);
					}
				}
			}
			byte[] RB = rsp.FetchBasicString().Characters;
			if (algorithm >= CryptAlgo.OmacAes128)
			{
				crypt = new CryptOmac(this, keydata, algorithm);
			}
			else if (algorithm >= CryptAlgo.EaxAes128)
			{
				crypt = new CryptEax(this, keydata, algorithm, RA, RB);
			}
			else
			{
				crypt =
					new CryptAesDes(
						this,
						keydata,
						algorithm,
						RA,
						RB,
                        (0 != (mode & (StartEncMode.MultiAppCard | StartEncMode.ProfessionalCard | StartEncMode.KdpMode))));
			}
			crypt.SetProtocol(pconn.Protocol);
			pconn.CommandAndResponseCrypt=crypt;
			this.algo = algorithm;
			this.key = key;
            this.kdpCrypto = (0 != (mode & StartEncMode.KdpMode));
		}

		/// <summary>
		/// Called by CommandAndResponseCrypt reset function and by
		/// CmdEndEncryption.
		/// </summary>
		internal void OnResetEncryption()
		{
			algo = CryptAlgo.None;
			key = -1;
            kdpCrypto = false;
			pconn.CommandAndResponseCrypt=null;
		}        

        /// <summary>
        /// Call BasicCard START ENCRYPTION command with key derivation parameter (KDP)
        /// </summary>
        /// <param name="algorithm">encryption algorithm</param>
        /// <param name="termKdp">Terminal known part of KDP</param>
        /// <param name="withCardKdp">True if card program includes set another part of KDP</param>
        protected internal virtual void CmdStartKdpEncryption(CryptAlgo algorithm, byte[] termKdp, bool withCardKdp)
        {
            int lenR;
            switch (algorithm)
            {
                case CryptAlgo.Aes128:
                case CryptAlgo.Aes192:
                case CryptAlgo.Aes256:
                case CryptAlgo.EaxAes128:
                case CryptAlgo.EaxAes192:
                case CryptAlgo.EaxAes256:
                    lenR = 8;
                    break;
                case CryptAlgo.Des:
                case CryptAlgo.DesCrc:
                case CryptAlgo.TripleDes:
                case CryptAlgo.TripleDesEde2Crc:
                case CryptAlgo.TripleDesEde3Crc:                    
                    lenR = 4;
                    break;
                case CryptAlgo.OmacAes128:
                case CryptAlgo.OmacAes192:
                case CryptAlgo.OmacAes256:
                    lenR = 0;
                    break;
                default:
                    throw new ArgumentOutOfRangeException("algorithm");
            }
            IntStartEncryption(algorithm, 0, lenR,
                               StartEncMode.KdpMode |
                               (null != termKdp ? StartEncMode.WithTermKdp : 0) |
                               (withCardKdp ? StartEncMode.WithCardKdp : 0),
                               termKdp);
        }

        /// <summary>
        /// Call BasicCard START ENCRYPTION command for Compact or Enhanced BasicCard.
        /// <seealso cref="AbstractBasicCardService.CmdStartEncryption"/>
        /// <seealso cref="AbstractBasicCardService.CmdProEncryption"/>
        /// <seealso cref="AbstractBasicCardService.CmdSMEncryption"/>
        /// <seealso cref="AbstractBasicCardService.CmdSMAuthentication"/>
        /// <seealso cref="AbstractBasicCardService.AutoEncryption(int,string)"/>
        /// <seealso cref="AbstractBasicCardService.CmdEndEncryption"/>
        /// </summary>
        /// <param name="key">key number</param>
        /// <param name="algorithm">encryption algorithm</param>
        protected internal virtual void CmdStartEncryption(int key, CryptAlgo algorithm)
		{
			if (connection.Shared && !connection.Locked)
			{
				throw new CommsException("Encryption requires exclusive access to smart card. Use lock()");
			}
			IntStartEncryption(algorithm, key, 4);
		}        

		/// <summary>
		/// Call BasicCard START ENCRYPTION command for Professional BasicCard
		/// (lenR=8). Use <see cref="CmdStartEncryption"/> if lenR=4 should be used.
		/// <seealso cref="AbstractBasicCardService.CmdStartEncryption"/>
		/// <seealso cref="AbstractBasicCardService.CmdProEncryption"/>
		/// <seealso cref="AbstractBasicCardService.CmdSMEncryption"/>
		/// <seealso cref="AbstractBasicCardService.CmdSMAuthentication"/>
		/// <seealso cref="AbstractBasicCardService.AutoEncryption(int, string)"/>
		/// <seealso cref="AbstractBasicCardService.CmdEndEncryption"/>
		/// </summary>
		/// <param name="key">key number</param>
		/// <param name="algorithm">encryption algorithm</param>
		protected internal virtual void CmdProEncryption(int key, CryptAlgo algorithm)
		{
			if (connection.Shared && !connection.Locked)
			{
				throw new CommsException("Encryption requires exclusive access to smart card. Use Lock()");
			}
            int lenR = 8;
            switch (algorithm)
            {
                case CryptAlgo.OmacAes128:
                case CryptAlgo.OmacAes192:
                case CryptAlgo.OmacAes256:
                    lenR = 0;
                    break;
            }
			IntStartEncryption(algorithm, key, lenR, StartEncMode.ProfessionalCard);
		}

		/// <summary>
		/// Call BasicCard START ENCRYPTION command for MultiApplication BasicCard.
		/// <seealso cref="AbstractBasicCardService.CmdStartEncryption"/>
		/// <seealso cref="AbstractBasicCardService.CmdProEncryption"/>
		/// <seealso cref="AbstractBasicCardService.CmdSMEncryption"/>
		/// <seealso cref="AbstractBasicCardService.CmdSMAuthentication"/>
		/// <seealso cref="AbstractBasicCardService.AutoEncryption(int,string)"/>
		/// <seealso cref="AbstractBasicCardService.CmdEndEncryption"/>
		/// </summary>
		/// <param name="keyCID">CID of key component</param>
		/// <param name="algorithm">encryption algorithm</param>
		/// <param name="lenR">length of random RA send to card</param>
		protected internal virtual void CmdSMEncryption(int keyCID, CryptAlgo algorithm, int lenR)
		{
			if (connection.Shared && !connection.Locked)
			{
				throw new CommsException("Encryption requires exclusive access to smart card. Use Lock()");
			}
			IntStartEncryption(algorithm, keyCID, lenR, StartEncMode.MultiAppCard);
		}

		/// <summary>
		/// Call BasicCard START ENCRYPTION command for MultiApplication BasicCard MAC.
		/// <seealso cref="AbstractBasicCardService.CmdStartEncryption"/>
		/// <seealso cref="AbstractBasicCardService.CmdProEncryption"/>
		/// <seealso cref="AbstractBasicCardService.CmdSMEncryption"/>
		/// <seealso cref="AbstractBasicCardService.CmdSMAuthentication"/>
		/// <seealso cref="AbstractBasicCardService.AutoEncryption(int,string)"/>
		/// <seealso cref="AbstractBasicCardService.CmdEndEncryption"/>
		/// </summary>
		/// <param name="keyCID">CID of key component</param>
		/// <param name="algorithm">encryption algorithm</param>
		protected internal virtual void CmdSMAuthentication(int keyCID, CryptAlgo algorithm)
		{
			CmdSMEncryption(keyCID, algorithm, 0);
		}

		/// <overloads>
		/// AutoEncryption handles cmdStartEncryption, cmdProEncryption, cmdSMEncryption
		/// for the different card types. For using this function the encryption key
		/// must have been set by use of setKey and used key number (keyNum).
		/// <seealso cref="AbstractBasicCardService.CmdStartEncryption"/>
		/// <seealso cref="AbstractBasicCardService.CmdProEncryption"/>
		/// <seealso cref="AbstractBasicCardService.CmdSMEncryption"/>
		/// <seealso cref="AbstractBasicCardService.CmdSMAuthentication"/>
		/// <seealso cref="AbstractBasicCardService.AutoEncryption(int,string)"/>
		/// <seealso cref="AbstractBasicCardService.CmdEndEncryption"/>
		/// </overloads>>
		/// <param name="keyNum">keyNum is the key number, for all card types.</param>
		/// <param name="keyName">Encrypting for the MultiApplication BasicCard
		/// also requires the key's path name, in keyName. For none MultiApplication
		/// BasicCard this parameter is not required and may be null.</param>
		protected internal virtual void AutoEncryption(int keyNum, BasicCardString keyName)
		{
			bool tryAES = (GetKeyLength(keyNum) >= 16);
			try
			{
				if (tryAES)
				{
					try
					{
						CmdProEncryption(keyNum, CryptAlgo.Auto);
					}
					catch (CommsException e)
					{
						if (e.SW1SW2 == StatusWord.swLcLeError)
						{
							CmdStartEncryption(keyNum, CryptAlgo.Auto);
						}
						else
						{
							throw e;
						}
					}
				}
				else
				{
					CmdStartEncryption(keyNum, CryptAlgo.Auto);
				}
			}
			catch (CommsException e)
			{
				switch (e.SW1SW2)
				{
					case StatusWord.swUnknownAlgorithm :
						// Compact BasicCard doesn't support P1=0
						CmdStartEncryption(keyNum, CryptAlgo.SglfsrCrc);
						break;

					case StatusWord.swBadComponentType :
						// MultiApplication BasicCard
						int CID =
							CmdFindComponent(ComponentType.Key, keyName);
							keys.IndexKey(CID, keyNum);
						if (tryAES)
						{
							CmdSMEncryption(CID, CryptAlgo.Auto, 8);
						}
						else
						{
							CmdSMEncryption(CID, CryptAlgo.Auto, 4);
						}
						break;
				}
			}
		}

		/// <param name="keyNum">keyNum is the key number, for all card types.</param>
		/// <param name="keyName">Encrypting for the MultiApplication BasicCard 
		/// also requires the key's path name, in keyName. For none MultiApplication
		/// BasicCard this parameter is not required and may be null.</param>
		protected internal virtual void AutoEncryption(int keyNum, string keyName)
		{
			AutoEncryption(
					keyNum,
					new BasicCardString(
						keyName,
						BasicCardString.CharsetIdLatin1
                        ));
		}

		/// <summary>
		/// Start encryption for MultiApplication BasicCard with known key CID and
		/// key value.
		/// </summary>
		/// <param name="keyCID">CID of key component</param>
		/// <param name="keyVal">key value (bytes)</param>
		/// <param name="algorithm">encryption algorithm</param>
		protected internal virtual void SMEncryptionByCid(int keyCID, byte[] keyVal, CryptAlgo algorithm)
		{
			// Tell the Terminal program interpreter the value of the key
			SetKey(keyCID, keyVal);

			if (algorithm < CryptAlgo.OmacAes128)
			{
				//  Encryption algorithm - initialisation data required

				if (algorithm < CryptAlgo.EaxAes128)
				{
					// Four-byte initialisation data
					CmdSMEncryption(keyCID, algorithm, 4);
				}
				else
				{
					// Eight-byte initialisation data
					CmdSMEncryption(keyCID, algorithm, 8);
				}
			}
			else
			{
				//  Authentication algorithm - no initialisation data required
				CmdSMEncryption(keyCID, algorithm, 0);
			}
		}

		/// <overloads>
		/// Start encryption for MultiApplication BasicCard with known key name and
		/// key value. 
		/// </overloads>
		/// <param name="keyComponent">full qualified path of key component</param>
		/// <param name="keyVal">key value (bytes)</param>
		/// <param name="algorithm">encryption algorithm</param>
		protected internal virtual void SMEncryptionByName(
			BasicCardString keyComponent,
			byte[] keyVal,
			CryptAlgo algorithm)
		{
			int CID = CmdFindComponent(ComponentType.Key, keyComponent);
			SMEncryptionByCid(CID, keyVal, algorithm);
		}

		/// <param name="keyComponent">full qualified path of key component</param>
		/// <param name="keyVal">key value (bytes)</param>
		/// <param name="algorithm">algorithm as described in BasicCard manual</param>
		protected internal virtual void SMEncryptionByName(
			string keyComponent,
			byte[] keyVal,
			CryptAlgo algorithm)
		{
			SMEncryptionByName(
				new BasicCardString(
					keyComponent,
					BasicCardString.CharsetIdLatin1),
				keyVal,
			algorithm);
		}

		/// <summary>
		/// Call BasicCard END ENCRYPTION command and stop encryptiong between service and card.
		/// <seealso cref="AbstractBasicCardService.CmdStartEncryption"/>
		/// <seealso cref="AbstractBasicCardService.CmdProEncryption"/>
		/// <seealso cref="AbstractBasicCardService.CmdSMEncryption"/>
		/// <seealso cref="AbstractBasicCardService.CmdSMAuthentication"/>
		/// <seealso cref="AbstractBasicCardService.AutoEncryption(int,string)"/>
		/// <seealso cref="AbstractBasicCardService.CmdEndEncryption"/>
		/// </summary>
		protected internal virtual void CmdEndEncryption()
		{
			if (crypt != null)
				crypt.SetCryptOptions(true, false);
			try
			{
				BasicCardCommand cmd =
					new BasicCardCommand(ClaBasicCard, InsEndEncryption);
				cmd.DisableLE();
				BasicCardResponse rsp = DoCommandAndResponse(cmd);
				rsp.CheckSW1SW2();
				OnResetEncryption();
			}
			finally
			{
				if (crypt != null)
					crypt.SetCryptOptions(true, true);
			}
		}

		/// <summary>
		/// Call BasicCard ECHO command.
		/// </summary>
		/// <param name="increment">value to add to each string byte (0..255)</param>
		/// <param name="str">input string</param>
		/// <returns>response string</returns>
		protected internal virtual BasicCardString CmdEcho(byte increment, BasicCardString str)
        {
			BasicCardCommand cmd = new BasicCardCommand(ClaBasicCard, InsEcho);
            if (allowExtendedLength && (str.Length > 255)) cmd.UseExtendedLength = true;
			cmd.P1=increment;
			cmd.AppendBasicString(str);
			cmd.LE=str.Length;
			BasicCardResponse rsp = DoCommandAndResponse(cmd);
			rsp.CheckSW1SW2();
			BasicCardString ret = rsp.FetchBasicString();
			ret.CharsetID=str.CharsetID;
			return ret;
		}

		/// <summary>
		/// Call BasicCard ASSIGN NAD command.
		/// </summary>
		/// <param name="nad">new NAD</param>
		protected internal virtual void CmdAssignNad(byte nad)
		{
			BasicCardCommand cmd =
				new BasicCardCommand(ClaBasicCard, InsAssignNad);
			cmd.P1=nad;
			cmd.DisableLE();
			BasicCardResponse rsp = DoCommandAndResponse(cmd);
			rsp.CheckSW1SW2();
		}

		/// <summary>
		/// Call BasicCard FILE IO command.
		/// </summary>
		/// <param name="syscode">SysCode</param>
		/// <param name="filenum">filenum</param>
		/// <param name="commanddata">CommandData</param>
		/// <param name="responselen">ResponseLen</param>
		/// <returns>status (1 byte) + ResponseData</returns>
		protected internal virtual byte[] CmdFileIO(
			byte syscode,
			byte filenum,
			byte[] commanddata,
			byte responselen)
        {
			BasicCardCommand cmd = new BasicCardCommand(ClaBasicCard, InsFileIo);
            if (allowExtendedLength && ((commanddata.Length > 255) || (responselen>255))) cmd.UseExtendedLength = true;
			cmd.P1=syscode;
			cmd.P2=filenum;
			cmd.AppendBasicString(commanddata);
			if ((responselen != 0) && (pconn.Protocol == 0))
			{
				responselen++; // for T=0 add one byte 
			}
			cmd.LE=responselen;
			BasicCardResponse rsp = DoCommandAndResponse(cmd);
			rsp.CheckSW1SW2();
			return rsp.Data;
		}

    	/// <summary>
		/// Call BasicCard GET CHALLENGE command.
		/// </summary>
		/// <param name="challengeLen">length of requested challenge</param>
		/// <returns>challenge bytes</returns>
		protected internal virtual byte[] CmdGetChallenge(int challengeLen)
		{
			BasicCardCommand cmd =
				new BasicCardCommand(ClaBasicCard, InsGetChallenge);
			cmd.LE=challengeLen;
			BasicCardResponse rsp = DoCommandAndResponse(cmd);
			rsp.CheckSW1SW2();
			return rsp.Data;
		}

		/// <summary>
		/// Call BasicCard EXTERNAL AUTHENTICATE command.
		/// <seealso cref="AbstractBasicCardService.CmdGetChallenge"/>
		/// </summary>
		/// <param name="keyCID">CID of key component</param>
		/// <param name="algorithm">algorithm id</param>
		/// <param name="response">response to challenge</param>
		protected internal virtual void CmdExternalAuthenticate(
			int keyCID,
			CryptAlgo algorithm,
			byte[] response)
        {
			BasicCardCommand cmd =
				new BasicCardCommand(ClaBasicCard, InsExternalAuthenticate);
			cmd.P1P2=((short) (keyCID & 0xFFFF));
			cmd.AppendBasicByte((byte) (((int)algorithm) & 0xFF));
			cmd.AppendBasicString(response);
			cmd.DisableLE();
			BasicCardResponse rsp = DoCommandAndResponse(cmd);
			rsp.CheckSW1SW2();
		}

		/// <summary>
		/// Call BasicCard INTERNAL AUTHENTICATE command.
		/// </summary>
		/// <param name="keyCID">CID of key component</param>
		/// <param name="algorithm">algorithm id</param>
		/// <param name="challenge">challenge</param>
		/// <returns>response to challenge</returns>
		protected internal virtual byte[] CmdInternalAuthenticate(
			int keyCID,
			CryptAlgo algorithm,
			byte[] challenge)
        {
			BasicCardCommand cmd =
				new BasicCardCommand(ClaBasicCard, InsInternalAuthenticate);
			cmd.P1P2=((short) (keyCID & 0xFFFF));
			cmd.AppendBasicByte((byte) (((int)algorithm) & 0xFF));
			cmd.AppendBasicString(challenge);
			cmd.LE=challenge.Length;
			BasicCardResponse rsp = DoCommandAndResponse(cmd);
			rsp.CheckSW1SW2();
			return rsp.Data;
		}

		/// <summary>
		/// Call BasicCard VERIFY command.
		/// </summary>
		/// <param name="keyCID">CID of key/password component</param>
		/// <param name="password">password</param>
		protected internal virtual void CmdVerify(int keyCID, BasicCardString password)
        {
			BasicCardCommand cmd = new BasicCardCommand(ClaBasicCard, InsVerify);
			cmd.P1P2=((short) (keyCID & 0xFFFF));
			cmd.AppendBasicString(password);
			cmd.DisableLE();
			BasicCardResponse rsp = DoCommandAndResponse(cmd);
			rsp.CheckSW1SW2();
		}

		/// <summary>
		/// Call BasicCard GET FREE MEMORY command.
		/// </summary>
		/// <returns>int[2] with int[0]=TotalFreeMemory and int[1]=LargestFreeBlock</returns>
		protected internal virtual int[] CmdGetFreeMemory()
		{
			BasicCardCommand cmd =
				new BasicCardCommand(ClaBasicCard, InsGetFreeMemory);
			cmd.LE=4;
			BasicCardResponse rsp = DoCommandAndResponse(cmd);
			rsp.CheckSW1SW2();
			int[] ret = new int[2];
			ret[0] = ((int) rsp.FetchBasicInteger()) & 0xFFFF;
			ret[1] = ((int) rsp.FetchBasicInteger()) & 0xFFFF;
			return ret;
		}

		/// <summary>
		/// Call BasicCard SELECT APPLICATION command.
		/// </summary>
		/// <param name="filename">name of file to select as</param>
		protected internal virtual void CmdSelectApplication(BasicCardString filename)
        {
			BasicCardCommand cmd =
				new BasicCardCommand(ClaBasicCard, InsSelectApplication);
			cmd.AppendBasicString(filename);
			cmd.DisableLE();
			BasicCardResponse rsp = DoCommandAndResponse(cmd);
			rsp.CheckSW1SW2();
		}

		/// <summary>
		/// Call BasicCard CREATE COMPONENT command.
		/// </summary>
		/// <param name="componentType">component type id</param>
		/// <param name="pathname">path of component</param>
		/// <param name="attributes">attributes (may be null)</param>
		/// <param name="data">data (may be null)</param>
		/// <returns>CID of new component</returns>
		protected internal virtual int CmdCreateComponent(
			ComponentType componentType,
			BasicCardString pathname,
			byte[] attributes,
			byte[] data)
        {
			BasicCardCommand cmd =
				new BasicCardCommand(ClaBasicCard, InsCreateComponent);
			cmd.P1=(byte) componentType;
			cmd.AppendBasicByte((byte) (pathname.Length & 0xFF));
			if (null != attributes)
			{
				cmd.AppendBasicByte((byte) (attributes.Length & 0xFF));
				cmd.AppendBasicString(attributes);
			}
			else
			{
				cmd.AppendBasicByte((byte) 0);
			}
			cmd.AppendBasicString(data);
			BasicCardResponse rsp = DoCommandAndResponse(cmd);
			rsp.CheckSW1SW2();
			return ((int) rsp.FetchBasicInteger()) & 0xFFFF;
		}

		/// <summary>
		/// Call DELETE COMPONENT command.
		/// </summary>
		/// <param name="CID">CID of component</param>
		protected internal virtual void CmdDeleteComponent(int CID)
		{
			BasicCardCommand cmd =
				new BasicCardCommand(ClaBasicCard, InsDeleteComponent);
			cmd.P1P2=((short) (CID & 0xFFFF));
			cmd.DisableLE();
			BasicCardResponse rsp = DoCommandAndResponse(cmd);
			rsp.CheckSW1SW2();
		}

		/// <summary>
		/// Call WRITE COMPONENT ATTR command.
		/// </summary>
		/// <param name="CID">CID of component</param>
		/// <param name="attributes">attributes</param>
		protected internal virtual void CmdWriteComponentAttr(int CID, byte[] attributes)
		{
			BasicCardCommand cmd =
				new BasicCardCommand(ClaBasicCard, InsWriteComponentAttr);
			cmd.P1P2=((short) (CID & 0xFFFF));
			cmd.AppendBasicString(attributes);
			cmd.DisableLE();
			BasicCardResponse rsp = DoCommandAndResponse(cmd);
			rsp.CheckSW1SW2();
		}

		/// <summary>
		/// Call BasicCard READ COMPONENT ATTR command.
		/// </summary>
		/// <param name="CID">CID of component</param>
		/// <returns>attributes</returns>
		protected internal virtual byte[] CmdReadComponentAttr(int CID)
		{
			BasicCardCommand cmd =
				new BasicCardCommand(ClaBasicCard, InsReadComponentAttr);
			cmd.P1P2=((short) (CID & 0xFFFF));
			cmd.LE=0;
			BasicCardResponse rsp = DoCommandAndResponse(cmd);
			rsp.CheckSW1SW2();
			return rsp.Data;
		}

		/// <summary>
		/// Call BasicCard WRITE COMPONENT DATA command.
		/// </summary>
		/// <param name="CID">CID of component</param>
		/// <param name="data">data</param>
		protected internal virtual void CmdWriteComponentData(int CID, byte[] data)
		{
			BasicCardCommand cmd =
				new BasicCardCommand(ClaBasicCard, InsWriteComponentData);
			cmd.P1P2=((short) (CID & 0xFFFF));
			cmd.AppendBasicString(data);
			cmd.DisableLE();
			BasicCardResponse rsp = DoCommandAndResponse(cmd);
			rsp.CheckSW1SW2();
		}

		/// <summary>
		/// Call BasicCard READ COMPONENT DATA command.
		/// </summary>
		/// <param name="CID">CID of component</param>
		/// <returns>data</returns>
		protected internal virtual byte[] CmdReadComponentData(int CID)
		{
			BasicCardCommand cmd =
				new BasicCardCommand(ClaBasicCard, InsReadComponentData);
			cmd.P1P2=((short) (CID & 0xFFFF));
			cmd.LE=0;
			BasicCardResponse rsp = DoCommandAndResponse(cmd);
			rsp.CheckSW1SW2();
			return rsp.Data;
		}

		/// <summary>
		/// Call BasicCard FIND COMPONENT command.
		/// </summary>
		/// <param name="componentType">component type</param>
		/// <param name="pathname">path name of component</param>
		/// <returns>CID of component</returns>
		protected internal virtual int CmdFindComponent(
			ComponentType componentType,
			BasicCardString pathname)
		{
			BasicCardCommand cmd =
				new BasicCardCommand(ClaBasicCard, InsFindComponent);
			cmd.P1=(byte) componentType;
			cmd.AppendBasicString(pathname);
			cmd.LE=2;
			BasicCardResponse rsp = DoCommandAndResponse(cmd);
			rsp.CheckSW1SW2();
			int CID = (int) rsp.FetchBasicInteger() & 0xFFFF;
			return CID;
		}

  		/// <summary>
		/// Call BasicCard COMPONENT NAME command.
		/// </summary>
		/// <param name="CID">CID of component</param>
		/// <returns>component name</returns>
		protected internal virtual BasicCardString CmdComponentName(int CID)
		{
			BasicCardCommand cmd =
				new BasicCardCommand(ClaBasicCard, InsComponentName);
			cmd.P1P2=((short) (CID & 0xFFFF));
			cmd.LE=0;
			BasicCardResponse rsp = DoCommandAndResponse(cmd);
			rsp.CheckSW1SW2();
			return rsp.FetchBasicString();
		}

		/// <summary>
		/// Call BasicCard GRANT PRIVILEGE command.
		/// </summary>
		/// <param name="privilegeCID">CID of privilege</param>
		/// <param name="filename">name of the file to which the privilege should be granted</param>
		protected internal virtual void CmdGrantPrivilege(
			int privilegeCID,
			BasicCardString filename)
        {
			BasicCardCommand cmd =
				new BasicCardCommand(ClaBasicCard, InsGrantPrivilege);
			cmd.P1P2=((short) (privilegeCID & 0xFFFF));
			cmd.AppendBasicString(filename);
			cmd.DisableLE();
			BasicCardResponse rsp = DoCommandAndResponse(cmd);
			rsp.CheckSW1SW2();
		}

		/// <summary>
		/// Call BasicCard AUTHENTICATE FILE command.
		/// </summary>
		/// <param name="keyCID">CID of key used</param>
		/// <param name="signature">signature or MAC</param>
		protected internal virtual void CmdAuthenticateFile(int keyCID, byte[] signature)
        {
			BasicCardCommand cmd =
				new BasicCardCommand(ClaBasicCard, InsAuthenticateFile);
			cmd.P1P2=((short) (keyCID & 0xFFFF));
			cmd.AppendBasicString(signature);
			cmd.DisableLE();
			BasicCardResponse rsp = DoCommandAndResponse(cmd);
			rsp.CheckSW1SW2();
		}

		/// <summary>
		/// Call BasicCard READ RIGHTS LIST command.
		/// </summary>
		/// <param name="start">number of first right to receive</param>
		/// <param name="nmax">maximum number of rights</param>
		/// <param name="filename">name of file to process</param>
		/// <returns>rights list as array of integer, containing an CID within each array entry</returns>
		protected internal virtual int[] CmdReadRightsList(
			int start,
			int nmax,
			BasicCardString filename)
        {
			BasicCardCommand cmd =
				new BasicCardCommand(ClaBasicCard, InsReadRightsList);
			cmd.P1=((byte) (start & 0xFF));
			cmd.AppendBasicString(filename);
			cmd.LE=(nmax * 2);
			BasicCardResponse rsp = DoCommandAndResponse(cmd);
			rsp.CheckSW1SW2();
			int count = rsp.Length / 2;
			int[] ret = new int[count];
			for (int i = 0; i < count; i++)
			{	
				ret[i] = rsp.FetchBasicInteger() & 0xFFFF;
			}
			return ret;
		}

 		/// <summary>
		/// Call BasicCard LOAD SEQUENCE command.
		/// </summary>
		/// <param name="phase">load sequence phase</param>
		protected internal virtual void CmdLoadSequence(LoadSequencePhase phase)
		{
			BasicCardCommand cmd =
				new BasicCardCommand(ClaBasicCard, InsLoadSequence);
			cmd.P1=(byte)phase;
			cmd.DisableLE();
			BasicCardResponse rsp = DoCommandAndResponse(cmd);
			rsp.CheckSW1SW2();
		}

		/// <summary>
		/// Call BasicCard SECURE TRANSPORT (start session) command.
		/// </summary>
		/// <param name="keyCID">CID of key</param>
		/// <param name="algorithm">id of algorithm</param>
		/// <param name="nonce">nonce</param>
		protected internal virtual void CmdSecureTransportStartSession(
        int keyCID,
        CryptAlgo algorithm,
        byte[] nonce)
        {
			BasicCardCommand cmd =
				new BasicCardCommand(ClaBasicCard, InsSecureTransport);
			cmd.P1P2=((short) (keyCID & 0xFFFF));
			cmd.AppendBasicByte((byte) (((int)algorithm) & 0xFF));
			cmd.AppendBasicString(nonce);
			cmd.DisableLE();
			BasicCardResponse rsp = DoCommandAndResponse(cmd);
			rsp.CheckSW1SW2();
		}

		/// <summary>
		/// Call BasicCard SECURE TRANSPORT (end session) command.
		/// </summary>
		protected internal virtual void CmdSecureTransportEndSession()
		{
			BasicCardCommand cmd =
				new BasicCardCommand(ClaBasicCard, InsSecureTransport);
			cmd.DisableLE();
			BasicCardResponse rsp = DoCommandAndResponse(cmd);
			rsp.CheckSW1SW2();
		}

		/// <overloads>
		/// Set encryption key.
		/// </overloads>
		/// <param name="keyNumber">number (0..255) or CID of key </param>
		/// <param name="key">encryption key</param>
		protected void SetKey(int keyNumber, byte[] key)
		{
			keys.SetKey(keyNumber, key);
		}

		/// <param name="keyComponent">full qualified path of key component</param>
		/// <param name="key">encryption key</param>
		protected void SetKey(string keyComponent, byte[] key)
		{
			keys.SetKey(keyComponent, key);
		}

		/// <param name="keyComponent">full qualified path of key component</param>
		/// <param name="key">encryption key</param>
		protected void SetKey(BasicCardString keyComponent, byte[] key)
		{
			keys.SetKey(keyComponent.ToString(), key);
		}

        /// <summary>
        /// Compute test customer key 
        /// </summary>
        /// <remarks>
        /// Computed key is only valid in virtual/emulated cards
        /// </remarks>
        /// <param name="termKdp">Terminal part of key derivation parameter KDP</param>
        /// <param name="cardKdp">Card part of key derivation parameter KDP (may be null)</param>
        /// <returns>Customer (test) key derived from KDP.</returns>
        protected byte[] ComputeTestCustomerKeyFromKdp(IList<byte> termKdp, IList<byte> cardKdp)
        {
            if (null == termKdp) throw new ArgumentNullException("termKdp");
            byte[] kdp = new byte[32];
            for (int i = 0; i < kdp.Length; i++)
            {
                byte a, b;
                if (termKdp.Count > i) a = termKdp[i]; else a = 0;
                if ((cardKdp != null) && (cardKdp.Count > i)) b = termKdp[i]; else b = 0;
                kdp[i] = (byte)(a ^ b);
            }
            SHA256 hash = SHA256Managed.Create();
            byte[] kinput = hash.ComputeHash(kdp);
            Rijndael aes = RijndaelManaged.Create();
            aes.Mode = CipherMode.ECB;
            aes.KeySize = 192;
            aes.Key = new byte[]{0x00, 0x11, 0x22, 0x33, 0x44, 0x55, 0x66, 0x77, 
                                 0x88, 0x99, 0xAA, 0xBB, 0xCC, 0xDD, 0xEE, 0xFF,
                                 0x01, 0x23, 0x45, 0x67, 0x89, 0xAB, 0xCD, 0xEF};
            return aes.CreateEncryptor().TransformFinalBlock(kinput, 0, kinput.Length);
        }

        /// <summary>
        /// Set encryption key for KDP, customer key encryption
        /// </summary>
        /// <param name="key">encryption key</param>
        /// <seealso cref="ComputeTestCustomerKeyFromKdp"/>
        /// <seealso cref="CmdStartKdpEncryption"/>
        protected void SetCustomerKey(byte[] key)
        {
            keys.SetCustomerKey(key);
        }

		/// <summary>
		/// Get length of specified key.
		/// </summary>
		/// <param name="keyNumber">number or CID of key</param>
		/// <returns>key length</returns>
		protected int GetKeyLength(int keyNumber)
		{
			byte[] key = keys.GetKey(keyNumber);
			if (null == key)
			{
				return 0;
			}
			else
			{
				return key.Length;
			}
		}

		/// <param name="keyComponent">full qualified path/name of component containing this key</param>
		/// <returns>key length</returns>
		private int GetKeyLength(String keyComponent)
		{
			byte[] key = keys.GetKey(keyComponent);
			if (null == key)
			{
				return 0;
			}
			else
			{
				return key.Length;
			}
		}

        
        /// <summary>
        /// Get key length of customer key
        /// </summary>        
        /// <returns>key length</returns>
        protected int GetCustomerKeyLength()
        {
            byte[] key = keys.GetCustomerKey();
            if (null == key)
            {
                return 0;
            }
            else
            {
                return key.Length;
            }
        }

        /// <summary>
        /// True if encryption with customer key is active
        /// </summary>
        protected bool CustomerKeyEncryption
        {
            get
            {
                return kdpCrypto;
            }
        }

		/// <summary>
		/// Number or CID of current active encryption key. 
		/// -1 if no encryption active.
		/// </summary>
		protected int KeyNumber
		{
			get
			{
				return key;
			}
		}

		/// <summary>
		/// Id of current active encryption algorithm.
		/// </summary>
		protected CryptAlgo Algorithm
		{
			get
			{
				return algo;
			}
		}

		/// <summary>
		/// Fileio object to access BasicCard file system.
		/// </summary>
		protected internal virtual BasicCardFileIO FileIO
		{
			get
			{
				lock(this)
				{
					if (null == fileio)
					{
						fileio = new BasicCardFileIO(this);
					}
					return fileio;
				}
			}
		}

		/// <summary>
		/// Create matching BlockCrypt object.
		/// </summary>
		/// <param name="algorithm">algorithm id</param>
		/// <param name="key">encryption key</param>
		/// <returns>BlockCrypt object</returns>
		private BlockCrypt CreateCrypt(CryptAlgo algorithm, byte[] key)
        {
			int keylen;
			BaseAlgorithm algo;
			switch (algorithm)
			{
				case CryptAlgo.TripleDes :
					keylen = 16;
					algo = BaseAlgorithm.TripleDes;
					break;
				case CryptAlgo.Des :
					keylen = 8;
					algo = BaseAlgorithm.Des;
					break;
				case CryptAlgo.OmacAes128 :
					keylen = 16;
					algo = BaseAlgorithm.Aes;
					break;
				case CryptAlgo.OmacAes192 :
					keylen = 24;
					algo = BaseAlgorithm.Aes;
					break;
				case CryptAlgo.OmacAes256 :
					keylen = 32;
					algo = BaseAlgorithm.Aes;
					break;
				default :
					throw new CardServiceException(
						StatusWord.swUnknownAlgorithm);
			}
			if (null == key)
				throw new CardServiceException(StatusWord.swKeyNotFound);
			if (key.Length < keylen)
				throw new CardServiceException(StatusWord.swKeyTooShort);
			byte[] finalkey = key;
			if (key.Length > keylen)
			{
				finalkey = new byte[keylen];
				System.Array.Copy(key, 0, finalkey, 0, keylen);
			}
			return new BlockCrypt(algo, finalkey);
		}

		/// <overloads>
		/// Validate the inserted card by use of INTERNAL AUTHENTICATE. The required
		/// key must have been set before by use of setKey and key path (keyComponent).
		/// If the validation fails because of card does have a different key, then
		/// a CardServiceException with SW1SW2=swBadAuthenticate is thrown.
		/// </overloads>
		/// <param name="keyComponent">full qualified path of key component</param>
		/// <param name="algorithm">algorithm id</param>
		protected internal virtual void InternalAuthenticate(
			BasicCardString keyComponent,
			CryptAlgo algorithm)
        {
			int CID = CmdFindComponent(ComponentType.Key, keyComponent);
			byte[] key = keys.GetKey(keyComponent.ToString());
			if (null == key)
			{
				key = keys.GetKey(CID);
			}
            if (null == key)
            {
                throw new CardServiceException(StatusWord.swKeyNotLoaded);
            }
			BlockCrypt crypt = CreateCrypt(algorithm, key);
			int blocklen = crypt.BlockLen;
			byte[] challenge = new byte[blocklen];
			random.GetBytes(challenge);
			byte[] response = CmdInternalAuthenticate(CID, algorithm, challenge);
			crypt.Encrypt(challenge, 0);
			if ((null != response) && (response.Length == challenge.Length))
			{
				for (int i = 0; i < response.Length; i++)
				{
					if (response[i] != challenge[i])
					{
						throw new CardServiceException(
							StatusWord.swBadAuthenticate);
					}
				}
			}
			else
			{
				throw new CardServiceException(StatusWord.swBadAuthenticate);
			}
		}

		/// <param name="keyComponent">full qualified path of key component</param>
		/// <param name="algorithm">algorithm id</param>
		protected internal virtual void InternalAuthenticate(String keyComponent, CryptAlgo algorithm)
		{
			BasicCardString bcStr =
                new BasicCardString(
                    keyComponent,
                    BasicCardString.CharsetIdLatin1);
            InternalAuthenticate(bcStr, algorithm);
        }

		/// <overloads>
		/// Authenticate to the inserted card by use of EXTERNAL AUTHENTICATE. The required
		/// key must have been set before by use of setKey and key path (keyComponent).
		/// </overloads>
		/// <param name="keyComponent">full qualified path of key component</param>
		/// <param name="algorithm">algorithm id</param>
		protected internal virtual void ExternalAuthenticate(
			BasicCardString keyComponent,
			CryptAlgo algorithm)
        {
			int CID = CmdFindComponent(ComponentType.Key, keyComponent);
			byte[] key = keys.GetKey(keyComponent.ToString());
			if (null == key)
			{
				key = keys.GetKey(CID);
			}
            if (null == key)
            {
                throw new CardServiceException(StatusWord.swKeyNotLoaded);
            }
			BlockCrypt crypt = CreateCrypt(algorithm, key);
			int blocklen = crypt.BlockLen;
			byte[] challenge = CmdGetChallenge(blocklen);
			crypt.Encrypt(challenge, 0);
			CmdExternalAuthenticate(CID, algorithm, challenge);
		}

		/// <param name="keyComponent">full qualified path of key component</param>
		/// <param name="algorithm">algorithm id</param>
		protected internal virtual void ExternalAuthenticate(string keyComponent, CryptAlgo algorithm)
		{
			BasicCardString bcStr =
                new BasicCardString(
                    keyComponent,
                    BasicCardString.CharsetIdLatin1);
            ExternalAuthenticate(bcStr, algorithm);
        }

        private CommandApdu MakeExtended(CommandApdu cmd, int blocklen)
        {
            if ((cmd.LE >= (256 - blocklen)) ||
                (cmd.ExtendedLC >= (256 - blocklen)))
            {
                // Make extended apdu if allowed
                if (AllowExtendedLength && (!cmd.UseExtendedLC || !cmd.UseExtendedLE))
                {
                    CommandApdu newcmd =
                    new CommandApdu(
                    cmd.CLA,
                    cmd.INS,
                    cmd.P1,
                    cmd.P2);
                    newcmd.UseExtendedLength = true;
                    newcmd.Append(cmd.IData);
                    newcmd.LE = cmd.LE;
                    newcmd.DoNotLog = cmd.DoNotLog;
                    return newcmd;
                }
            }
            return cmd;
        }

        private static void ApduFieldAdd(byte[] field, byte add)
        {
            int test = field[field.Length - 1] + add;
            if (test > 255)
            {
                field[field.Length - 2]++;
                field[field.Length - 1] = (byte)(test & 0xFF);
            }
            else field[field.Length -1]=(byte) test;
        }

        private static byte[] MakeField(int value, int len)
        {
            byte[] ret = new byte[len];
            for (int i = (ret.Length - 1); i >= 0; i--)
            {
                ret[i] = (byte)(value & 0xFF);
                value >>= 8;
            }
            return ret;
        }

        /// <summary>
        /// Make 3 or 1 byte LC field
        /// </summary>
        /// <param name="cmd"></param>
        /// <returns></returns>
        private static byte[] MakeLcField(CommandApdu cmd)
        {
            return MakeField(cmd.ExtendedLC, (cmd.UseExtendedLC ? 3 : 1));
        }

        /// <summary>
        /// Make 2 or 1 byte LE field
        /// </summary>
        /// <param name="cmd"></param>
        /// <returns></returns>
        private static byte[] MakeLeField(CommandApdu cmd)
        {
            return MakeField((cmd.IsLEDisabled ? 0 : cmd.LE), (cmd.UseExtendedLE ? 2 : 1));
        }
    }
}
