/*****************************************************************************/
/* Copyright (c) 2005, ZeitControl cardsystems GmbH                          */
/* All rights reserved.                                                      */
/*                                                                           */
/* Redistribution and use in source and binary forms, with or without        */
/* modification, are permitted provided that the following conditions are    */
/* met:                                                                      */
/*    * Redistributions of source code must retain the above copyright       */
/*      notice, this list of conditions and the following disclaimer.        */
/*    * Redistributions in binary form must reproduce the above copyright    */
/*      notice, this list of conditions and the following disclaimer in the  */
/*      documentation and/or other materials provided with the distribution. */
/*    * Neither the name of the company ZeitControl nor the names of its     */
/*      contributors may be used to endorse or promote products derived from */
/*      this software without specific prior written permission.             */
/*                                                                           */
/* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS       */
/* "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED */
/* TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A           */
/* PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER  */
/* OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,  */
/* EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,       */
/* PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR        */
/* PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF    */
/* LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING      */
/* NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS        */
/* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.              */
/*****************************************************************************/
/* Created on 01.04.2005
 * Project: BasicCard .Net Library
 * File description: 
 *   BasicCard fileio support.
 * --------------------------------------------------------
 * $Header: /opt/cvsroot/net/BasicCardLib.net/BasicCard/BasicCards/FileSys/BasicCardFileIO.cs,v 1.1 2005/07/22 12:57:57 michael Exp $
 * $Revision: 1.1 $
 * $Date: 2005/07/22 12:57:57 $
 * --------------------------------------------------------
 * $Log: BasicCardFileIO.cs,v $
 * Revision 1.1  2005/07/22 12:57:57  michael
 * Initial checkin
 *
 *---------------------------------------------------------
 */
using System;

namespace BasicCard.BasicCards.FileSys
{
	/// <summary>
	/// Lock info for BasicCard files.
	/// </summary>
	[CLSCompliant(true), Serializable]
    public class BasicCardFileLockInfo
    {
        private bool readLock;
        private bool writeLock;
        private bool customLockSet;
        private bool customLockState;
        private int[] readKeys;
        private int[] writeKeys;

		/// <summary>
		/// Constructs a LockInfo object from bytes returned by GetLockInfoBytes.
		/// </summary>
		/// <param name="lockInfoBytes">lock info bytes</param>
        internal BasicCardFileLockInfo(byte[] lockInfoBytes)
        {
            if (lockInfoBytes.Length < 7)
            {
                throw new BasicCardFileIOException(
                    BasicCardFileError.feUnexpectedError);
            }
            // read lock
            switch (lockInfoBytes[0])
            {
                case 0 :
                    readLock = false;
                    readKeys = null;
                    break;
                case 1 :
                    readLock = true;
                    readKeys = new int[1];
                    readKeys[0] = lockInfoBytes[3] & 0xFF;
                    break;
                case 2 :
                    readLock = true;
                    readKeys = new int[2];
                    readKeys[0] = lockInfoBytes[3] & 0xFF;
                    readKeys[1] = lockInfoBytes[4] & 0xFF;
                    break;
                case 3 :
                    readLock = true;
                    readKeys = null;
                    break;
                default :
                    throw new BasicCardFileIOException(
                        BasicCardFileError.feUnexpectedError);
            }
            // write lock
            switch (lockInfoBytes[1])
            {
                case 0 :
                    writeLock = false;
                    writeKeys = null;
                    break;
                case 1 :
                    writeLock = true;
                    writeKeys = new int[1];
                    writeKeys[0] = lockInfoBytes[5] & 0xFF;
                    break;
                case 2 :
                    writeLock = true;
                    writeKeys = new int[2];
                    writeKeys[0] = lockInfoBytes[5] & 0xFF;
                    writeKeys[1] = lockInfoBytes[6] & 0xFF;
                    break;
                case 3 :
                    writeLock = true;
                    writeKeys = null;
                    break;
                default :
                    throw new BasicCardFileIOException(
                        BasicCardFileError.feUnexpectedError);
            }
            // custom lock
            switch (lockInfoBytes[2])
            {
                case 0 :
                    customLockSet = false;
                    customLockState = false;
                    break;
                case 1 :
                    customLockSet = true;
                    customLockState = false;
                    break;
                case 2 :
                    customLockSet = true;
                    customLockState = true;
                    break;
                default :
                    throw new BasicCardFileIOException(
                        BasicCardFileError.feUnexpectedError);
            }
        }

		/// <summary>
		/// Get read access lock state.
		/// </summary>
		/// <remarks>
		/// Even if read access is locked, read is still possible when encryption with
		/// one of the keys specified by ReadKeys is active.
		/// </remarks>
		/// <returns>true if read access id locked (prohibited), false if read access is allowed</returns>
        public bool IsReadLocked
        {
			get
			{
				return readLock;
			}
        }

		/// <summary>
		/// Keys allowing read access for read locked files. This property is null if
		/// no read keys are set.
		/// </summary>
        public int[] ReadKeys
        {
			get
			{
				return readKeys;
			}
        }

		/// <summary>
		/// Get write access lock state.
		/// </summary>
		/// <remarks>
		/// Even if write access is locked, write is still possible when encryption with
		/// one of the keys specified by WriteKeys is active.
		/// </remarks>
        public bool IsWriteLocked
        {
			get
			{
				return writeLock;
			}
        }

		/// <summary>
		/// Keys allowing write access for write locked files. This property is null if
		/// no write keys are set.
		/// </summary>
        public int[] WriteKeys
        {
			get
			{
				return writeKeys;
			}
        }

		/// <summary>
		/// True if custom lock is set or false if custom lock is not set.
		/// Use CustomLockState to get information if an existing
		/// custom lock is locked or unlocked.
		/// </summary>
        public bool IsCustomLocked
        {
			get
			{
				return customLockSet;
			}
        }

		/// <summary>
		/// If a custom lock for a file is set (IsCustomLocked). This state shows if file
		/// is locked (access prohibited) or unlocked (access allowed). This property is
		/// true if locked or false if unlocked.
		/// </summary>
        public bool CustomLockState
        {
			get
			{
				return customLockState;
			}
        }
    }

	/// <summary>
	/// Mode enumeration (Flags) for open a BasicCard file.
	/// </summary>
	[CLSCompliant(true), Flags, Serializable]
	public enum BasicCardFileOpenMode
	{
		/// <summary>
		/// Open file shared.
		/// </summary>
		ShareShared = 0x00,
		/// <summary>
		/// Open file with exclusive read access
		/// </summary>
		ShareReadLocked = 0x01,
		/// <summary>
		/// Open file with exclusive write access
		/// </summary>
		ShareWiteLocked = 0x02,
		/// <summary>
		/// Open file exclusive (none shared)
		/// </summary>
		ShareLocked = ShareReadLocked | ShareWiteLocked,
	

		/// <summary>
		/// Open file in append mode.
		/// </summary>
		ModeAppend = 0x60,
		/// <summary>
		/// Open file in binary mode.
		/// </summary>
		ModeBinary = 0x20,
		/// <summary>
		///  Open file in input mode.
		/// </summary>
		ModeInput = 0x30,
		/// <summary>
		/// Open file in output mode.
		/// </summary>
		ModeOutput = 0x40,
		/// <summary>
		/// Open file in random mode.
		/// </summary>
		ModeRandom = 0x50,

		/// <summary>
		/// Open file for read access.
		/// </summary>
		AccessRead = 0x04,
		/// <summary>
		/// Open file for write access.
		/// </summary>
		AccessWrite = 0x08,
		/// <summary>
		/// Open file for read and write access.
		/// </summary>
		AccessReadWrite = AccessRead | AccessWrite
	}

	/// <summary>
	/// BasicCard file attributes.
	/// </summary>
	[CLSCompliant(true), Flags, Serializable]
	public enum BasicCardFileAttrib
	{
		/// <summary>
		/// BasicCard file.
		/// </summary>
		/// <remarks>
		/// File flag is also set for BasicCard directories.
		/// </remarks>
		File = 0x40,
		/// <summary>
		/// BasicCarc directory.
		/// </summary>
		Dir = 0x10 | File
	}

	/// <summary>
	/// Class to access the BasicCard file system.
	/// </summary>
	[CLSCompliant(true)]
	public class BasicCardFileIO
	{
		private string namespaceCharset;
		private AbstractBasicCardService service;

		/// <summary>
		/// Construct a BasicCardFileIo object.
		/// </summary>
		/// <param name="service">BasicCard service to use</param>
		public BasicCardFileIO(AbstractBasicCardService service)
		{
			this.namespaceCharset = BasicCardString.CharsetIdLatin1;
			this.service = service;
		}

		/// <summary>
		/// Create a directory on a BasicCard.
		/// </summary>
		/// <param name="directory">directory name</param>
		public void MakeDirectory(string directory)
        {
			byte[] ret =
				service.CmdFileIO(
                (byte) 0x80,
                (byte) 0,
                ToBString(directory),
                (byte) 0);
				CheckStatus(ret, directory);
		}

		/// <summary>
		/// Remove a directory on a BasicCard.
		/// </summary>
		/// <param name="directory"></param>
		public void RemoveDirectory(String directory)
        {
			byte[] ret =
				service.CmdFileIO(
                (byte) 0x81,
                (byte) 0,
                ToBString(directory),
                (byte) 0);
			CheckStatus(ret, directory.ToString());
		}

		/// <summary>
		/// Change current directory on a BasicCard.
		/// </summary>
		/// <param name="directory">directory name</param>
		public void ChangeCurrentDirectory(String directory)
		{
			byte[] ret =
				service.CmdFileIO(
					(byte) 0x82,
					(byte) 0,
					ToBString(directory),
					(byte) 0);
			CheckStatus(ret, directory.ToString());
		}

		/// <summary>
		/// Get current directory of BasicCard.
		/// </summary>
		/// <returns>current directory</returns>
		public String GetCurrentDirectory()
        {
			byte[] ret = service.CmdFileIO((byte) 0x83, (byte) 0, null, (byte) 0);
			CheckStatus(ret, null);
			byte[] stringdata = new byte[ret.Length - 1];
			System.Array.Copy(ret, 1, stringdata, 0, stringdata.Length);
			return ToNString(stringdata);
		}

		/// <summary>
		/// Count all files/directories matching a search pattern.
		/// </summary>
		/// <param name="search">search pattern</param>
		/// <returns>file count</returns>
		public int FindFileCount(String search)
		{
			byte[] ret =
				service.CmdFileIO(
                (byte) 0x84,
                (byte) 0,
                ToBString(search),
                (byte) 0);
			CheckStatus(ret, search.ToString());
			ByteBuffer result = new ByteBuffer(ret);
			return result.GetShort(1) & 0xFFFF;
		}

		/// <summary>
		/// Get a certain file/directory name for search pattern.
		/// </summary>
		/// <param name="search">search pattern (same as with previous call to findFileCount)</param>
		/// <param name="number">number of file to return (0..count-1)</param>
		/// <returns>file/directory name</returns>
		public string FindFileGet(string search, int number)
        {
			ByteBuffer buffer = new ByteBuffer();
			buffer.Length=2;
			buffer.SetShort(0, (short) (number + 1));
			buffer.Append(ToBString(search));
			byte[] ret =
				service.CmdFileIO(
					(byte) 0x85,
					(byte) 0,
					buffer.GetData(),
					(byte) 0);
			CheckStatus(ret, search.ToString());
			byte[] stringdata = new byte[ret.Length - 1];
			System.Array.Copy(ret, 1, stringdata, 0, stringdata.Length);
			return ToNString(stringdata);
		}

		/// <summary>
		/// Find files/directories matching a search pattern. This combines FindFileCount
		/// and FindFileGet.
		/// </summary>
		/// <param name="search">search pattern</param>
		/// <returns>matching file names</returns>
		public String[] FindFiles(string search)
		{
			int count = FindFileCount(search);
			string[] ret = new string[count];
			for (int i = 0; i < count; i++)
			{
				ret[i] = FindFileGet(search, i);
			}
			return ret;
		}

		/// <summary>
		/// Delete a file.
		/// </summary>
		/// <param name="file">file name</param>
		public void DeleteFile(string file)
        {
			byte[] ret =
				service.CmdFileIO((byte) 0x86, (byte) 0, ToBString(file), (byte) 0);
			CheckStatus(ret, file.ToString());
		}

		/// <summary>
		/// Move or rename a file or directory.
		/// </summary>
		/// <param name="oldFile">old file name</param>
		/// <param name="newFile">new file name</param>
		public void RenameFile(string oldFile, string newFile)
        {
			ByteBuffer data = new ByteBuffer();
			byte[] oldbytes = ToBString(oldFile);
			byte[] newbytes = ToBString(newFile);
			data.Append((byte) oldbytes.Length);
			data.Append(oldbytes);
			data.Append(newbytes);
			byte[] ret =
				service.CmdFileIO((byte) 0x87, (byte) 0, data.GetData(), (byte) 0);
			CheckStatus(ret, oldFile + " - " + newFile);
		}

		/// <summary>
		/// Open/create a file.
		/// </summary>
		/// <param name="filename">file name</param>
		/// <param name="mode">logical or of Access.., Mode.. and Share.. flags</param>
		/// <param name="recordLength">record length</param>
		/// <returns>file number</returns>
		public int OpenFile(string filename, BasicCardFileOpenMode mode, int recordLength)
        {
			ByteBuffer data = new ByteBuffer();
			data.Length=2;
			data.SetShort(0, (short) recordLength);
			data.Append(ToBString(filename));
			byte[] ret =
				service.CmdFileIO(
					(byte) 0x89,
					(byte) mode,
					data.GetData(),
					(byte) 0);
			CheckStatus(ret, filename);
			ByteBuffer result = new ByteBuffer(ret);
			return result.GetShort(1) & 0xFFFF;
		}

		/// <summary>
		/// Close a file.
		/// </summary>
		/// <param name="fileNumber">file number</param>
		public void CloseFile(int fileNumber)
        {
			byte[] ret =
				service.CmdFileIO((byte) 0x8A, (byte) fileNumber, null, (byte) 0);
			CheckStatus(ret, null);
		}

		/// <summary>
		/// Close all open files.
		/// </summary>
		public void CloseAllFiles()
		{
			byte[] ret = service.CmdFileIO((byte) 0x8B, (byte) 0, null, (byte) 0);
			CheckStatus(ret, null);
		}

		/// <summary>
		/// Get file length.
		/// </summary>
		/// <param name="fileNumber">file number</param>
		/// <returns>length of file</returns>
		public int GetFileLength(int fileNumber)
		{
			byte[] ret =
				service.CmdFileIO((byte) 0x8D, (byte) fileNumber, null, (byte) 0);
			CheckStatus(ret, null);
			ByteBuffer result = new ByteBuffer(ret);
			return result.GetInt(1);
		}

		/// <summary>
		/// Get file position.
		/// </summary>
		/// <param name="fileNumber">file number</param>
		/// <returns>read/write position within file</returns>
		public int GetFilePosition(int fileNumber)
        {
			byte[] ret =
				service.CmdFileIO((byte) 0x8E, (byte) fileNumber, null, (byte) 0);
			CheckStatus(ret, null);
			ByteBuffer result = new ByteBuffer(ret);
			return result.GetInt(1);
		}

		/// <summary>
		/// Set file position.
		/// </summary>
		/// <param name="fileNumber">file number</param>
		/// <param name="position">read/write position within file</param>
		public void SetFilePosition(int fileNumber, int position)
        {
			ByteBuffer data = new ByteBuffer();
			data.Length=4;
			data.SetInt(0, position);
			byte[] ret =
				service.CmdFileIO(
					(byte) 0x8f,
					(byte) fileNumber,
					data.GetData(),
					(byte) 0);
			CheckStatus(ret, null);
		}

		/// <summary>
		/// Check for end of file.
		/// </summary>
		/// <param name="fileNumber">file number</param>
		/// <returns>true if eof</returns>
		public bool Eof(int fileNumber)
		{
			byte[] ret =
				service.CmdFileIO((byte) 0x90, (byte) fileNumber, null, (byte) 0);
			CheckStatus(ret, null);
			ByteBuffer result = new ByteBuffer(ret);
			return (result.GetShort(1) == -1);
		}

		/// <summary>
		/// Get bytes from file at current position. This requires the file to be opened
		/// with ModeBinary or ModeRandom.
		/// </summary>
		/// <param name="fileNumber"></param>
		/// <param name="length"></param>
		/// <returns></returns>
		public byte[] FileGet(int fileNumber, int length)
        {
			if ((length > 255) || (length < 0))
			{
				throw new BasicCardFileIOException(
					BasicCardFileError.feBadParameter);
			}
			byte[] ret =
				service.CmdFileIO(
                (byte) 0x91,
                (byte) fileNumber,
                null,
                (byte) length);
			CheckStatus(ret, null);
			ByteBuffer result = new ByteBuffer(ret);
			return result.GetBytes(1, result.Length - 1);
		}

		/// <summary>
		/// Get bytes from file at given position. This requires the file to be opened
		/// with ModeBinary or ModeRandom.
		/// </summary>
		/// <param name="fileNumber">file number</param>
		/// <param name="position">position for read operation</param>
		/// <param name="length">number of bytes requested</param>
		/// <returns>bytes read from file</returns>
		public byte[] FileGetPos(int fileNumber, int position, int length)
        {
			if ((length > 255) || (length < 0))
			{
				throw new BasicCardFileIOException(
					BasicCardFileError.feBadParameter);
			}
			ByteBuffer data = new ByteBuffer();
			data.Length=4;
			data.SetInt(0, position);
			byte[] ret =
				service.CmdFileIO(
					(byte) 0x92,
					(byte) fileNumber,
					data.GetData(),
					(byte) length);
			CheckStatus(ret, null);
			ByteBuffer result = new ByteBuffer(ret);
			return result.GetBytes(1, result.Length - 1);
		}

		/// <summary>
		/// Write (put) to file at current position. This requires the file to be opened
		/// with ModeBinary or ModeRandom.
		/// </summary>
		/// <param name="fileNumber">file number</param>
		/// <param name="data">bytes to write</param>
		public void FilePut(int fileNumber, byte[] data)
        {
			byte[] ret =
				service.CmdFileIO((byte) 0x93, (byte) fileNumber, data, (byte) 0);
			CheckStatus(ret, null);
		}

		/// <summary>
		/// Write (put) to file at given position. This requires the file to be opened
		/// with ModeBinary or ModeRandom.
		/// </summary>
		/// <param name="fileNumber">file number</param>
		/// <param name="position">position for write operation</param>
		/// <param name="data">data bytes to write</param>
		public void FilePutPos(int fileNumber, int position, byte[] data)
		{
			ByteBuffer idata = new ByteBuffer();
			idata.Length=4;
			idata.SetInt(0, position);
			idata.Append(data);
			byte[] ret =
				service.CmdFileIO(
					(byte) 0x94,
					(byte) fileNumber,
					idata.GetData(),
					(byte) 0);
			CheckStatus(ret, null);
		}

		/// <summary>
		/// Set or unset a file or directory lock.
		/// </summary>
		/// <param name="filename">file name</param>
		/// <param name="locktype">lock type.</param>
		/// <param name="keynumbers">0, 1 or 2 keynumber. his may be null</param>
		private void SetFileLock(string filename, int locktype, int[] keynumbers)
        {
			byte[] file = ToBString(filename);
			byte[] keys = {(byte) 0xFF, (byte) 0xFF };
			ByteBuffer idata = new ByteBuffer();
			if (null != keynumbers)
			{
				if (keynumbers.Length >= 1)
				{
					keys[0] = (byte) keynumbers[0];
				}
				if (keynumbers.Length >= 2)
				{
					keys[1] = (byte) keynumbers[1];
				}
			}
			idata.Append(keys);
			idata.Append(file);
			byte[] ret =
				service.CmdFileIO(
					(byte) 0xA7,
					(byte) locktype,
					idata.GetData(),
					(byte) 0);
			CheckStatus(ret, null);
		}

		/// <summary>
		/// Set or unset a read lock for a file or directory.
		/// </summary>
		/// <param name="filename">file name</param>
		/// <param name="lockVal">true to lock the file, false to unlock the file</param>
		/// <param name="keynumbers">keys to allow for locked access (may be null)</param>
		public void SetReadLock(string filename, bool lockVal, int[] keynumbers)
		{
			int locktype = 0x4; // read lock
			if (lockVal)
			{
				if ((null != keynumbers) && (0 != keynumbers.Length))
				{
					locktype |= 0x2; // 2 keys (always because of bug)
				}
			}
			else
			{
				locktype |= 0x10;
			}
			SetFileLock(filename, locktype, keynumbers);
		}

		/// <summary>
		/// Set or unset a write lock for a file or directory.
		/// </summary>
		/// <param name="filename">file name</param>
		/// <param name="lockVal">true to lock the file, false to unlock the file</param>
		/// <param name="keynumbers">keys to allow for locked access (may be null)</param>
		public void SetWriteLock(string filename, bool lockVal, int[] keynumbers)
        {
			int locktype = 0x8; // write lock
			if (lockVal)
			{
				if ((null != keynumbers) && (0 != keynumbers.Length))
				{
					locktype |= 0x2; // 2 keys (always because of bug)
				}
			}
			else
			{
				locktype |= 0x10;
			}
			SetFileLock(filename, locktype, keynumbers);
		}

		/// <summary>
		/// Set or unset a read/write lock for a file or directory.
		/// </summary>
		/// <param name="filename">file name</param>
		/// <param name="lockVal">true to lock the file, false to unlock the file</param>
		/// <param name="keynumbers">keys to allow for locked access (may be null)</param>
		public void SetReadWriteLock(
			string filename,
			bool lockVal,
			int[] keynumbers)
        {
			int locktype = 0xC; // read/write lock
			if (lockVal)
			{
				if ((null != keynumbers) && (0 != keynumbers.Length))
				{
					locktype |= 0x2; // 2 keys (always because of bug)
				}
			}
			else
			{
				locktype |= 0x10;
			}
			SetFileLock(filename, locktype, keynumbers);
		}

		/// <summary>
		/// Set a custom lock for a file or directory.
		/// </summary>
		/// <remarks>
		/// A custom lock cannot be removed from outside card anymore.
		/// it must be removed by code within BasicCard. 
		/// </remarks>
		/// <param name="filename">file name</param>
		public void SetCustomLock(string filename)
        {
			int locktype = 0x0; // custom lock
			SetFileLock(filename, locktype, null);
		}

		/// <summary>
		/// Get file/directory lock info bytes.
		/// </summary>
		/// <param name="filename">filename</param>
		/// <returns>lock info bytes</returns>
		private byte[] GetLockInfoBytes(String filename)
		{
			byte[] ret =
				service.CmdFileIO(
					(byte) 0xA8,
					(byte) 0,
					ToBString(filename),
					(byte) 0);
			CheckStatus(ret, null);
			ByteBuffer result = new ByteBuffer(ret);
			return result.GetBytes(1, result.Length - 1);
		}

		/// <summary>
		/// Get file/directory lock information.
		/// </summary>
		/// <param name="filename">file name</param>
		/// <returns>lock information</returns>
		public BasicCardFileLockInfo GetLockInfo(string filename)
        {
			return new BasicCardFileLockInfo(GetLockInfoBytes(filename));
		}

		/// <summary>
		/// Get file/directory attribute.
		/// </summary>
		/// <param name="filename">file name</param>
		/// <returns>attributes</returns>
		public BasicCardFileAttrib GetFileAttrib(string filename)
        {
			byte[] ret =
				service.CmdFileIO(
					(byte) 0xA9,
					(byte) 0,
					ToBString(filename),
					(byte) 0);
			CheckStatus(ret, null);
			ByteBuffer result = new ByteBuffer(ret);
			return (BasicCardFileAttrib) (result.GetShort(1) & 0xFFFF);
		}

		private void CheckStatus(byte[] result, string filename)
        {
			BasicCardFileError fe;
			if (null == result)
			{
				fe = BasicCardFileError.feUnexpectedError;
			}
			else if (0 == result.Length)
			{
				fe = BasicCardFileError.feUnexpectedError;
			}
			else
			{
				fe = (BasicCardFileError) (result[0] & 0xFF);
			}
			if (0 != fe)
			{
				if (null == filename)
				{
					throw new BasicCardFileIOException(fe);
				}
				else
				{
					throw new BasicCardFileIOException(filename, fe);
				}
			}
		}

		private byte[] ToBString(string str)
        {
			BasicCardString bstr = new BasicCardString(str, namespaceCharset);
			return bstr.Characters;
		}

		private string ToNString(byte[] str)
		{
			BasicCardString bstr = new BasicCardString(str);
			return bstr.ToString(namespaceCharset);
		}

		/// <summary>
		/// Charset used for file name conversion between BasicCard and .Net.
		/// </summary>
		public string NamespaceCharset
		{
			get
			{
				return namespaceCharset;
			}
			set
			{
				namespaceCharset = value;
			}
		}
	}
}
