/*****************************************************************************/
/* Copyright (c) 2005, ZeitControl cardsystems GmbH                          */
/* All rights reserved.                                                      */
/*                                                                           */
/* Redistribution and use in source and binary forms, with or without        */
/* modification, are permitted provided that the following conditions are    */
/* met:                                                                      */
/*    * Redistributions of source code must retain the above copyright       */
/*      notice, this list of conditions and the following disclaimer.        */
/*    * Redistributions in binary form must reproduce the above copyright    */
/*      notice, this list of conditions and the following disclaimer in the  */
/*      documentation and/or other materials provided with the distribution. */
/*    * Neither the name of the company ZeitControl nor the names of its     */
/*      contributors may be used to endorse or promote products derived from */
/*      this software without specific prior written permission.             */
/*                                                                           */
/* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS       */
/* "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED */
/* TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A           */
/* PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER  */
/* OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,  */
/* EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,       */
/* PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR        */
/* PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF    */
/* LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING      */
/* NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS        */
/* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.              */
/*****************************************************************************/
/* Created on 01.04.2005
 * Project: BasicCard .Net Library
 * File description: 
 *   Block encryption class.
 * --------------------------------------------------------
 * $Header: /opt/cvsroot/net/BasicCardLib.net/BasicCard/Crypto/Messaging/BlockCrypt.cs,v 1.1 2005/07/22 12:57:58 michael Exp $
 * $Revision: 1.1 $
 * $Date: 2005/07/22 12:57:58 $
 * --------------------------------------------------------
 * $Log: BlockCrypt.cs,v $
 * Revision 1.1  2005/07/22 12:57:58  michael
 * Initial checkin
 *
 *---------------------------------------------------------
 */
using System;
using System.Security.Cryptography;

namespace BasicCard.Crypto.Messaging
{
	/// <summary>
	/// Basic symmetric block encryption (ECB).
	/// </summary>
	/// <threadsafety static="true" instance="false"/>
	[CLSCompliant(true)]
	public class BlockCrypt
	{
		private SymmetricAlgorithm crypt;
		private ICryptoTransform encrypt;
		private ICryptoTransform decrypt;

		private int blocklen;

		private void initAlgo(BaseAlgorithm algo)
		{
			switch (algo)
			{
				case BaseAlgorithm.Aes:
					crypt=new RijndaelManaged();
					break;
				case BaseAlgorithm.Des:
					crypt=new DESCryptoServiceProvider();
					break;
				case BaseAlgorithm.TripleDes:
					crypt=new TripleDESCryptoServiceProvider();
					break;
				default:
					throw new ArgumentOutOfRangeException("Invalid crypto algorithm.");
			}
		}

		private void init(byte[] key)
		{
			crypt.Key=key;
			crypt.Mode=CipherMode.ECB;
			crypt.Padding=PaddingMode.None;
			blocklen = crypt.BlockSize/8;
			encrypt = crypt.CreateEncryptor();
			decrypt = crypt.CreateDecryptor();
		}

		/// <overloads>
		/// <summary>
		/// Constructs a BlockCrypt object. 
		/// </summary>
		/// <remarks>
		/// For AES and Triple-DES the key length decide 
		/// about the AES type (AES128, AES192 or AES256) or
		/// Triple-DES type (2 key 112bit or 3 key 168bit) used.
		/// </remarks>
		/// </overloads>
		/// <param name="algo">Cryptographic algorithm to use.</param>
		/// <param name="key">Secret key.</param>
		public BlockCrypt(BaseAlgorithm algo, byte[] key)
		{
			initAlgo(algo);
			init(key);
		}

		
		/// <param name="crypt">Symmetric encryption algorithm to use.</param>
		/// <param name="key">Secret key.</param>
		public BlockCrypt(SymmetricAlgorithm crypt, byte[] key)
		{
			this.crypt=crypt;
			init(key);
		}

		/// <summary>
		/// Block length in bytes.
		/// </summary>
		public int BlockLen
		{
			get
			{
				return blocklen;
			}
		}

		/// <summary>
		/// Encrypt block length bytes of data starting at startpos.
		/// </summary>
		/// <param name="data">Data to encrypt.</param>
		/// <param name="startpos">Start position within data.</param>
		public void Encrypt(byte[] data, int startpos)
		{
			byte[] temp;
			temp=encrypt.TransformFinalBlock(data, startpos, blocklen);
			Array.Copy(temp, 0, data, startpos, blocklen);
		}

		/// <summary>
		/// Decrypt block length bytes of data starting at startpos.
		/// </summary>
		/// <param name="data">Data to decrypt.</param>
		/// <param name="startpos">Start position within data.</param>
		public void Decrypt(byte[] data, int startpos)
		{
			byte[] temp;
			temp=decrypt.TransformFinalBlock(data, startpos, blocklen);
			Array.Copy(temp, 0, data, startpos, blocklen);
		}
	}
}
