/*****************************************************************************/
/* Copyright (c) 2005, ZeitControl cardsystems GmbH                          */
/* All rights reserved.                                                      */
/*                                                                           */
/* Redistribution and use in source and binary forms, with or without        */
/* modification, are permitted provided that the following conditions are    */
/* met:                                                                      */
/*    * Redistributions of source code must retain the above copyright       */
/*      notice, this list of conditions and the following disclaimer.        */
/*    * Redistributions in binary form must reproduce the above copyright    */
/*      notice, this list of conditions and the following disclaimer in the  */
/*      documentation and/or other materials provided with the distribution. */
/*    * Neither the name of the company ZeitControl nor the names of its     */
/*      contributors may be used to endorse or promote products derived from */
/*      this software without specific prior written permission.             */
/*                                                                           */
/* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS       */
/* "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED */
/* TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A           */
/* PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER  */
/* OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,  */
/* EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,       */
/* PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR        */
/* PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF    */
/* LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING      */
/* NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS        */
/* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.              */
/*****************************************************************************/
/* Created on 01.04.2005
 * Project: BasicCard .Net Library
 * File description: 
 *   EAX encryption class.
 * --------------------------------------------------------
 * $Header: /opt/cvsroot/net/BasicCardLib.net/BasicCard/Crypto/Messaging/Eax.cs,v 1.1 2005/07/22 12:57:58 michael Exp $
 * $Revision: 1.1 $
 * $Date: 2005/07/22 12:57:58 $
 * --------------------------------------------------------
 * $Log: Eax.cs,v $
 * Revision 1.1  2005/07/22 12:57:58  michael
 * Initial checkin
 *
 *---------------------------------------------------------
 */
using System;

namespace BasicCard.Crypto.Messaging
{
	/// <summary>
	/// Provides classes and interfaces specific to BasicCard transparent EAX command encryption.
	/// </summary>
	/// <threadsafety static="true" instance="false"/>
	[CLSCompliant(true)]
	public class Eax
	{

		/// <summary>
		/// Counter mode encryption.
		/// </summary>
		private class Ctr
		{
			private byte[] N;
			private byte[] S;
			private int pos;
			private int blocklen;
			private Eax eax;

			/// <summary>
			/// CTR constructor.
			/// </summary>
			/// <param name="eax">Eax object using this object.</param>
			/// <param name="N">Nonce.</param>
			public Ctr(Eax eax, byte[] N)
			{
				this.eax = eax;
				this.N = (byte[]) N.Clone();
				blocklen = eax.omacFactory.BlockLen;
				S = new byte[blocklen];
				pos = 0;
			}

			/// <summary>
			/// Encrypt or decrypt a block.
			/// </summary>
			/// <param name="block">Block to encrypt.</param>
			/// <param name="startPos">Start position within block.</param>
			/// <param name="len">Number of bytes to encrypt.</param>
			public void Encrypt(byte[] block, int startPos, int len)
            {
				for (int i = 0; i < len; i++)
				{
					if (pos == 0)
					{
						UpdateS();
						IncrementN();
					}
					block[startPos + i] ^= S[pos];
					pos++;
					if (pos >= blocklen)
						pos = 0;
				}
			}

			/// <summary>
			/// Update S by encrypt N.
			/// </summary>
			private void UpdateS()
			{
				System.Array.Copy(N, 0, S, 0, blocklen);
				eax.omacFactory.EncryptBlock(S);
			}

			/// <summary>
			/// Increment the counter N.
			/// </summary>
			private void IncrementN()
			{
				byte add;
				for (int i = N.Length - 1; i >= 0; i--)
				{
					switch (N[i])
					{
                    case 255 : // 11111111b -> overflow
                        add = 1;
                        N[i] = 0;
                        break;
                    default :
                        add = 0;
                        N[i]++;
                        break;
					}
					if (0 == add)
						break;
				}
			}

		}

		private OmacFactory omacFactory;
		private OmacFactory.Omac1 U;
		private OmacFactory.Omac1 V;
		private OmacFactory.Omac1 W;
		private byte[] N;
		private Ctr ctr;

		/// <summary>
		/// Construc EAX object.
		/// </summary>
		/// <param name="algo">Base algorithm.</param>
		/// <param name="key">Secret key to be used.</param>
		public Eax(BaseAlgorithm algo, byte[] key)
		{
			omacFactory = new OmacFactory(algo, key);
			U = omacFactory.CreateOmac1();
			V = omacFactory.CreateOmac1();
			W = omacFactory.CreateOmac1();
		}

		/// <summary>
		/// Provide Nonce for EAX encryption. This must be called before using
		/// for encryption or decryption.
		/// </summary>
		/// <param name="Nonce">Nonce.</param>
		public void ProvideNonce(byte[] Nonce)
		{
			byte[] Tweak = new byte[omacFactory.BlockLen];
			U.Start();
			U.Append(Tweak); // Tweak = 0 
			U.Append(Nonce);
			N = U.End();
			ctr = new Ctr(this, N);
			V.Start();
			Tweak[Tweak.Length - 1] = 1; // Tweak = 1
			V.Append(Tweak);
			W.Start();
			Tweak[Tweak.Length - 1] = 2; // Tweak = 2
			W.Append(Tweak);
		}

		/// <overloads>
		/// Provide header. This may be called multiple times.
		/// </overloads>
		/// <param name="Header">Header bytes.</param>
		/// <param name="startPos">Start position in Header.</param>
		/// <param name="len">Number of bytes to use.</param>
		public void ProvideHeader(byte[] Header, int startPos, int len)
        {
			V.Append(Header, startPos, len);
		}

		/// <param name="Header">Header bytes.</param>
		public void ProvideHeader(byte[] Header)
		{
			V.Append(Header);
		}

		/// <param name="Header">Header byte.</param>
		public void ProvideHeader(byte Header)
		{
			V.Append(Header);
		}

		/// <overloads>
		/// Encrypt plaintext to chiphertext and add this to final Tag.
		/// </overloads>
		/// <param name="data">Data to encrypt.</param>
		/// <param name="startPos">Start position within data.</param>
		/// <param name="len">Number of bytes to encrypt.</param>
		public void ComputeCiphertext(byte[] data, int startPos, int len)
        {
			ctr.Encrypt(data, startPos, len);
			W.Append(data, startPos, len);
		}

		/// <param name="data">Data to encrypt.</param>
		public void ComputeCiphertext(byte[] data)
		{
			ComputeCiphertext(data, 0, data.Length);
		}

		/// <overloads>
		/// Decrypt ciphertext to plaintext and add ciphertext to final Tag.
		/// </overloads>
		/// <param name="data">Data to decrypt.</param>
		/// <param name="startPos">Start position within data.</param>
		/// <param name="len">Number of bytes to decrypt.</param>
	    public void ComputePlaintext(byte[] data, int startPos, int len)
        {
			W.Append(data, startPos, len);
			ctr.Encrypt(data, startPos, len);
		}

		/// <param name="data">Data to decrypt.</param>
		public void ComputePlaintext(byte[] data)
		{
			ComputePlaintext(data, 0, data.Length);
		}

		/// <summary>
		/// Compute Tag. This should be the last call.
		/// </summary>
		/// <returns>Tag</returns>
		public byte[] ComputeTag()
		{
			byte[] Tag = W.End();
			byte[] H = V.End();
			for (int i = 0; i < Tag.Length; i++)
			{
				Tag[i] ^= (byte) (H[i] ^ N[i]);
			}
			return Tag;
		}

		/// <overloads>
		/// Compare two tag results.
		/// </overloads>
		/// <param name="tag">First tag.</param>
		/// <param name="verify">Second tag to compare with.</param>
		/// <param name="verifyStart">Position of first omac byte in verify.</param>
		/// <param name="verifyLen">Length of omac in verify.</param>
		/// <returns>True if equal, false if not equal.</returns>
		public static bool CompareTag(
			byte[] tag,
			byte[] verify,
			int verifyStart,
			int verifyLen)
		{
			return OmacFactory.CompareOmac(tag, verify, verifyStart, verifyLen);
		}

		
		/// <param name="tag">First tag.</param>
		/// <param name="verify">Second tag to compare with.</param>
		/// <returns>True if equal, false if not equal.</returns>
		public static bool CompareTag(byte[] tag, byte[] verify)
		{
			return OmacFactory.CompareOmac(tag, verify);
		}
	}
}
