/*****************************************************************************/
/* Copyright (c) 2005, ZeitControl cardsystems GmbH                          */
/* All rights reserved.                                                      */
/*                                                                           */
/* Redistribution and use in source and binary forms, with or without        */
/* modification, are permitted provided that the following conditions are    */
/* met:                                                                      */
/*    * Redistributions of source code must retain the above copyright       */
/*      notice, this list of conditions and the following disclaimer.        */
/*    * Redistributions in binary form must reproduce the above copyright    */
/*      notice, this list of conditions and the following disclaimer in the  */
/*      documentation and/or other materials provided with the distribution. */
/*    * Neither the name of the company ZeitControl nor the names of its     */
/*      contributors may be used to endorse or promote products derived from */
/*      this software without specific prior written permission.             */
/*                                                                           */
/* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS       */
/* "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED */
/* TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A           */
/* PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER  */
/* OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,  */
/* EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,       */
/* PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR        */
/* PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF    */
/* LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING      */
/* NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS        */
/* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.              */
/*****************************************************************************/
/* Created on 01.04.2005
 * Project: BasicCard .Net Library
 * File description: 
 *   Message encryption class.
 * --------------------------------------------------------
 * $Header: /opt/cvsroot/net/BasicCardLib.net/BasicCard/Crypto/Messaging/MessageCrypt.cs,v 1.3 2009/04/23 10:46:15 michael Exp $
 * $Revision: 1.3 $
 * $Date: 2009/04/23 10:46:15 $
 * --------------------------------------------------------
 * $Log: MessageCrypt.cs,v $
 * Revision 1.3  2009/04/23 10:46:15  michael
 * V1.3.8
 *
 * Revision 1.2  2009/03/18 12:50:41  michael
 * Release Candidate .Net 2.0
 *
 * Revision 1.1  2005/07/22 12:57:58  michael
 * Initial checkin
 *
 *---------------------------------------------------------
 */
using System;
using System.Security.Cryptography;

namespace BasicCard.Crypto.Messaging
{
	/// <summary>
	/// Class to encrypt BasicCard command data using CBC with BasicCard style CBS padding.
	/// </summary>
	/// <threadsafety static="true" instance="false"/>
	[CLSCompliant(true)]
	public class MessageCrypt
	{
		
		private BlockCrypt crypt;
		private byte[] iv;
		private byte[] nextiv;
		private int blocklen;
		

		/// <summary>
		/// Construct a MessageCrypt object.
		/// </summary>
		/// <param name="algo">Algorithm.</param>
		/// <param name="key">Secret key.</param>
		public MessageCrypt(BaseAlgorithm algo, byte[] key)
		{
			crypt = new BlockCrypt(algo, key);
			blocklen = crypt.BlockLen;
			iv = new byte[blocklen];
			nextiv = new byte[blocklen];
		}

		/// <summary>
		/// Encrypt len bytes of message starting at startpos.
		/// </summary>
		/// <param name="message">Message to encrypt.</param>
		/// <param name="startpos">Start position within message.</param>
		/// <param name="len">Number of bytes to encrypt.</param>
		public void Encrypt(byte[] message, int startpos, int len)
        {
			while (len >= blocklen)
			{
				for (int i = 0; i < blocklen; i++)
					iv[i] ^= message[i + startpos];
				crypt.Encrypt(iv, 0);
				System.Array.Copy(iv, 0, message, startpos, blocklen);
				len -= blocklen;
				startpos += blocklen;
			}

			if (len > 0)
			{
				for (int i = 0; i < len; i++)
					iv[i] ^= message[i + startpos];
				crypt.Encrypt(iv, 0);
				System.Array.Copy(
					iv,
					0,
					message,
					startpos + len - blocklen,
					blocklen);
			}
		}

		/// <summary>
		/// Decrypt len bytes of message starting at startpos.
		/// </summary>
		/// <param name="message">Message to decrypt.</param>
		/// <param name="startpos">Start position within message.</param>
		/// <param name="len">Number of bytes to decrypt.</param>
		public void Decrypt(byte[] message, int startpos, int len)
		{
			if (0==len) return;
			int pos = startpos + len - blocklen;
			System.Array.Copy(message, pos, nextiv, 0, blocklen);

			if ((len % blocklen) > 0) // Any fiddly stuff at the end?
			{
				crypt.Decrypt(message, pos);
				for (int shift = 0; shift < (len % blocklen); shift++)
				{
					byte temp = message[pos];
					for (int i = 0; i < (blocklen-1); i++)
						message[i + pos] = message[i + 1 + pos];
					message[pos + blocklen - 1] = temp;
				}
				pos -= len % blocklen;
				for (int i = 0; i < (len % blocklen); i++)
					message[pos + i + blocklen] ^= message[i + pos];
				len -= len % blocklen;
			}

			while (len > 0)
			{
				crypt.Decrypt(message, pos);
				byte[] p;
				int pPos;
				if (len > blocklen)
				{
					p = message;
					pPos = pos - blocklen;
				}
				else
				{
	                p = iv;
		            pPos = 0;
			    }

	            for (int i = 0; i < blocklen; i++)
		            message[i + pos] ^= p[pPos + i];
			    len -= blocklen;
	            pos -= blocklen;
			}

	        System.Array.Copy(nextiv, 0, iv, 0, blocklen);
	    }

		/// <summary>
		/// Set IV used for encryption.
		/// </summary>
		/// <param name="ivdata">IV</param>
		/// <param name="start">Start position of IV within ivdata</param>
		public void SetIV(byte[] ivdata, int start)
		{
			System.Array.Copy(ivdata, start, iv, 0, blocklen);
		}

		/// <summary>
		/// Block length of used encryption algorithm.
		/// </summary>
		public int BlockLen
		{
			get
			{
				return blocklen;
			}
		}
	}
}
