/*****************************************************************************/
/* Copyright (c) 2005, ZeitControl cardsystems GmbH                          */
/* All rights reserved.                                                      */
/*                                                                           */
/* Redistribution and use in source and binary forms, with or without        */
/* modification, are permitted provided that the following conditions are    */
/* met:                                                                      */
/*    * Redistributions of source code must retain the above copyright       */
/*      notice, this list of conditions and the following disclaimer.        */
/*    * Redistributions in binary form must reproduce the above copyright    */
/*      notice, this list of conditions and the following disclaimer in the  */
/*      documentation and/or other materials provided with the distribution. */
/*    * Neither the name of the company ZeitControl nor the names of its     */
/*      contributors may be used to endorse or promote products derived from */
/*      this software without specific prior written permission.             */
/*                                                                           */
/* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS       */
/* "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED */
/* TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A           */
/* PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER  */
/* OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,  */
/* EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,       */
/* PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR        */
/* PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF    */
/* LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING      */
/* NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS        */
/* SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.              */
/*****************************************************************************/
/* Created on 01.04.2005
 * Project: BasicCard .Net Library
 * File description: 
 *   One Key CBC MAC (OMAC).
 * --------------------------------------------------------
 * $Header: /opt/cvsroot/net/BasicCardLib.net/BasicCard/Crypto/Messaging/OmacFactory.cs,v 1.1 2005/07/22 12:57:58 michael Exp $
 * $Revision: 1.1 $
 * $Date: 2005/07/22 12:57:58 $
 * --------------------------------------------------------
 * $Log: OmacFactory.cs,v $
 * Revision 1.1  2005/07/22 12:57:58  michael
 * Initial checkin
 *
 *---------------------------------------------------------
 */
using System;
using BasicCard;

namespace BasicCard.Crypto.Messaging
{
	/// <summary>
	/// OMACFactory: One-Key CBC MAC.
	/// </summary>
	/// <threadsafety static="true" instance="false"/>
	[CLSCompliant(true)]
	public class OmacFactory
	{
		/// <summary>
		/// Generic OMAC class, to be inherited by OMAC1 and OMAC2
		/// </summary>
		/// <threadsafety static="true" instance="false"/>
		[CLSCompliant(true)]
		abstract public class Omac
		{
			internal byte[] block;
			internal int blockPtr;
			internal OmacFactory factory;

			/// <summary>
			/// OMAC constructor.
			/// </summary>
			/// <param name="factory">Factory.</param>
			protected internal Omac(OmacFactory factory)
			{
				this.factory=factory;
				block = new byte[factory.blocklen];
				Start();
			}

        	/// <summary>
			/// (Re)Start OMAC.
			/// </summary>
			public void Start()
			{
				blockPtr = 0;
				for (int i = 0; i < factory.blocklen; i++)
					block[i] = 0;
			}

       		/// <overloads>
			/// Append data to OMAC.
			/// </overloads>
			/// <summary>
			/// Append bytes to OMAC.
			/// </summary>
			/// <param name="data">Data to append.</param>
			public void Append(byte[] data)
			{
				Append(data, 0, data.Length);
			}

			/// <summary>
			/// Append len bytes of data to OMAC starting at start.
			/// </summary>
			/// <param name="data">Data to append.</param>
			/// <param name="start">Start position within data.</param>
			/// <param name="len">Number of bytes to append.</param>
			public void Append(byte[] data, int start, int len)
            {
				if (len == 0)
					return;
				for (;;)
				{
					int n = factory.blocklen - blockPtr;
					if (n > len)
						n = len;
					for (int i = 0; i < n; i++)
						block[blockPtr + i] ^= data[start + i];

					blockPtr += n;
					len -= n;
					if (len == 0)
						return;

					factory.EncryptBlock(block);
					blockPtr = 0;
					start += n;
				}
			}

        	/// <summary>
			/// Append one byte of data to OMAC.
			/// </summary>
			/// <param name="data">Data to append.</param>
			public void Append(byte data)
			{
				if (blockPtr >= factory.blocklen)
				{
					factory.EncryptBlock(block);
					blockPtr = 0;
				}
				block[blockPtr] ^= data;
				blockPtr++;
			}

        	/// <summary>
			/// End OMAC calculation.
			/// </summary>
			/// <returns>OMAC value.</returns>
			public abstract byte[] End();
		}


		/// <summary>
		/// OMAC1.
		/// </summary>
		/// <threadsafety static="true" instance="false"/>
		[CLSCompliant(true)]
		public class Omac1 : Omac
		{
			/// <summary>
			/// OMAC1 constructor.
			/// </summary>
			/// <param name="factory">Factory.</param>
			public Omac1(OmacFactory factory)
				: base(factory)
			{
			}

			/// <summary>
			/// End OMAC calculation.
			/// </summary>
			/// <returns>OMAC value.</returns>
			public override byte[] End()
			{
				byte[] MAC;
				if (blockPtr < factory.blocklen)
				{
					MAC = (byte[]) factory.GetLuu().Clone();
					MAC[blockPtr] ^= (byte) 0x80; // First padding bit
				}
				else
				{
					MAC = (byte[]) factory.GetLu().Clone();
				}
				for (int i = 0; i < factory.blocklen; i++)
					MAC[i] ^= block[i];
				factory.EncryptBlock(MAC);
				return MAC;
			}
		}

		/// <summary>
		/// OMAC2.
		/// </summary>
		/// <threadsafety static="true" instance="false"/>
		[CLSCompliant(true)]
		public class Omac2 : Omac
		{

			/// <summary>
			/// OMAC2 constructor.
			/// </summary>
			/// <param name="factory">Factory.</param>
			public Omac2(OmacFactory factory)
				: base(factory)
			{
			}

			/// <summary>
			/// End OMAC calculation.
			/// </summary>
			/// <returns>OMAC value.</returns>
			public override byte[] End()
			{
				byte[] MAC;
				if (blockPtr == factory.blocklen)
				{
					MAC = (byte[]) factory.GetLu().Clone();
				}
				else
				{
					MAC = (byte[]) factory.GetLdivu().Clone();
					MAC[blockPtr] ^= (byte) 0x80; // First padding bit
				}
				for (int i = 0; i < factory.blocklen; i++)
					MAC[i] ^= block[i];
				factory.EncryptBlock(MAC);
				return MAC;
			}
		}

		private byte[] L;
		private byte[] Lu; // L*u (use getLu() to use)
		private byte[] Luu; // L*u^2 (use getLuu() to use)
		private byte[] Ldivu; // L/u (use getLdivu() to use)
		private BlockCrypt crypt;
		private int blocklen;

		/// <summary>
		/// OMACFactory Constructor.
		/// </summary>
		/// <param name="algo">Base algorithm.</param>
		/// <param name="key">Secret key to be used.</param>
		public OmacFactory(BaseAlgorithm algo, byte[] key)
		{
			crypt = new BlockCrypt(algo, key);
			blocklen = crypt.BlockLen;
			if (16 != blocklen)
				throw new CommsException(StatusWord.swNotSupported);
			L = new byte[blocklen];
			EncryptBlock(L); // L = Ek(0n)
		}

		/// <summary>
		/// Create Omac1 object.
		/// </summary>
		/// <returns>OMAC1</returns>
		public Omac1 CreateOmac1()
		{
			return new Omac1(this);
		}

		/// <summary>
		/// Create Omac2 object.
		/// </summary>
		/// <returns>OMAC2</returns>
		public Omac2 CreateOmac2()
		{
			return new Omac2(this);
		}

		static private void GF128_MulBy_u(byte[] L)
		{
			bool carry = false;
			byte temp;
			for (int i = 15; i >= 0; i--)
			{
				temp = (byte) (carry ? 1 : 0);
				carry = (0 != (L[i] & (byte) 0x80));
				L[i] <<= 1;
				L[i] += temp;
			}
			if (carry)
			{
				L[15] ^= (byte) 0x87;
			}
		}

		static private void GF128_DivBy_u(byte[] L)
		{
			bool carry = false;
			byte temp;
			for (int i = 0; i < 16; i++)
			{
				temp = (byte) (carry ? 0x80 : 0);
				carry = (0 != (L[i] & (byte) 1));
				L[i] >>= 1;
				L[i] += temp;
			}
			if (carry)
			{
				L[0] ^= (byte) 0x80;
				L[1 ^ 5] ^= (byte) 0x43;
			}
		}

    	/// <summary>
		/// Get Lu which is L multiplied by u
		/// </summary>
		/// <returns>Lu</returns>
		private byte[] GetLu()
		{
			lock(this)
			{
				if (null == Lu)
				{
					Lu = (byte[]) L.Clone();
					GF128_MulBy_u(Lu);
				}
				return Lu;
			}
		}

    	/// <summary>
		/// Get Luu which is L multiplied by u^2
		/// </summary>
		/// <returns>Luu</returns>
		private byte[] GetLuu()
		{
			lock(this)
			{
				if (null == Luu)
				{
					Luu = (byte[]) GetLu().Clone();
					GF128_MulBy_u(Luu);
				}
				return Luu;
			}
		}

		/// <summary>
		/// Get Ldivu which is L divided by u
		/// </summary>
		/// <returns>Ldivu</returns>
		private byte[] GetLdivu()
		{
			lock(this)
			{
				if (null == Ldivu)
				{
					Ldivu = (byte[]) L.Clone();
					GF128_DivBy_u(Ldivu);
				}
				return Ldivu;
			}
		}

		internal void EncryptBlock(byte[] block)
		{
			this.crypt.Encrypt(block, 0);
		}

		/// <summary>
		/// Get block length of used encryption algorithm
		/// </summary>
		public int BlockLen
		{
			get
			{
				return blocklen;
			}
		}

		/// <overloads>
		/// Compare two omac results.
		/// </overloads>
		/// <param name="omac">First omac.</param>
		/// <param name="verify">Second omac to compare with.</param>
		/// <param name="verifyStart">Position of first omac byte in verify.</param>
		/// <param name="verifyLen">Length of omac in verify.</param>
		/// <returns>True if equal, false if not equal.</returns>
		public static bool CompareOmac(
			byte[] omac,
			byte[] verify,
			int verifyStart,
			int verifyLen)
		{
			if (omac.Length != verifyLen)
				return false;
			for (int i = 0; i < verifyLen; i++)
			{
				if (omac[i] != verify[i + verifyStart])
					return false;
			}
			return true;
		}

		/// <param name="omac">First omac.</param>
		/// <param name="verify">Second omac to compare with.</param>
		/// <returns>True if equal, false if not equal.</returns>
		public static bool CompareOmac(byte[] omac, byte[] verify)
		{
			return CompareOmac(omac, verify, 0, verify.Length);
		}
	}
}
