using System;
using System.Collections.Generic;
using System.Text;

namespace BasicCardExplorer
{
    enum BasicCardState
    {
        Unknown=-1,
        New,
        Load,
        Pers,
        Test,
        Run        
    }

    enum BasicCardCategory
    {
        NoBasicCard=-1,
        Compact,
        Enhanced,
        MultiApplication,
        Professional
    }

    class BasicCardInfo
    {
        private int major=-1;
        private int minor=-1;
        private string revision=null;
        private string versionText = null;
        private BasicCardState state=BasicCardState.Unknown;

        public bool IsBasicCard
        {
            get { return (major >= 0) && (minor >=0); }
        }

        public bool HasRevision
        {
            get { return revision != null; }
        }

        public int MajorVersion { get { return major; } }

        public int MinorVersion { get { return minor; } }

        public string Revision 
        { 
            get 
            {
                if (revision == null) return "";
                else return revision; 
            } 
        }

        public BasicCardState State
        {
            get
            {
                return state;
            }
        }

        public string StateString
        {
            get
            {
                switch (State)
                {
                    case BasicCardState.New: return "NEW";
                    case BasicCardState.Load: return "LOAD";
                    case BasicCardState.Pers: return "PERS";
                    case BasicCardState.Run: return "RUN";
                    case BasicCardState.Test: return "TEST";
                    case BasicCardState.Unknown:
                    default:
                        return "Unknown";
                }
            }
        }

        public BasicCardCategory Category
        {
            get
            {
                switch (major)
                {
                    case 1: 
                        return BasicCardCategory.Compact;
                    case 2: 
                    case 3:
                        return BasicCardCategory.Enhanced;
                    case 4:
                    case 5:
                    case 7:
                        return BasicCardCategory.Professional;
                    case 6:
                        return BasicCardCategory.MultiApplication;
                    default:
                        return BasicCardCategory.NoBasicCard;
                }
            }
        }

        public string CategoryString
        {
            get
            {
                switch (Category)
                {
                    case BasicCardCategory.Compact:
                        return "Compact BasicCard";
                    case BasicCardCategory.Enhanced:
                        return "Enhanced BasicCard";
                    case BasicCardCategory.Professional:
                        return "Professional BasicCard";
                    case BasicCardCategory.MultiApplication:
                        return "MultiApplication BasicCard";
                    case BasicCardCategory.NoBasicCard:
                    default:
                        return "No BasicCard";
                }
            }
        }

        public override string ToString()
        {
            return CategoryString + " " + versionText;
        }

        public override bool Equals(object obj)
        {
            if (null == obj) return false;
            if (typeof(BasicCardInfo).IsInstanceOfType(obj))
            {
                BasicCardInfo b=(BasicCardInfo) obj;
                return (major == b.major) &&
                    (minor == b.minor) &&
                    Revision.Equals(b.Revision);
            }
            else
            {
                return false;
            }
        }

        public override int GetHashCode()
        {
            return major * 127 + minor;
        }

        /// <summary>
        /// Constructor
        /// </summary>
        /// <param name="getStateData"></param>
        internal BasicCardInfo(byte[] getStateData)
        {
            try
            {
                if ((null != getStateData) && (getStateData.Length >= 3))
                {
                    if (getStateData.Length == 3)
                    {
                        // state
                        switch (getStateData[0])
                        {
                            case 0:
                                state = BasicCardState.New;
                                break;
                            case 1:
                                state = BasicCardState.Load;
                                break;
                            case 2:
                                state = BasicCardState.Test;
                                break;
                            case 3:
                                state = BasicCardState.Run;
                                break;
                            default:
                                state = BasicCardState.Unknown;
                                break;
                        }
                        if (state != BasicCardState.Unknown)
                        {
                            major = getStateData[1];
                            minor = getStateData[2];
                            revision = null;
                            versionText = string.Format("ZC{0}.{1}", major, minor);
                        }                        
                    }
                    else
                    {
                        versionText = Encoding.ASCII.GetString(getStateData, 1, getStateData.Length - 1);
                        if (versionText.StartsWith("ZC"))
                        {
                            int pointPos = versionText.IndexOf('.');
                            int spacePos = versionText.IndexOf(' ');
                            int revPos = versionText.ToUpper().IndexOf("REV");
                            if (pointPos != 0)
                            {
                                if (spacePos < 0) spacePos = versionText.Length;
                                string strMajor = versionText.Substring(2, pointPos - 2);
                                major = int.Parse(strMajor);
                                string strMinor = versionText.Substring(pointPos + 1, spacePos - (pointPos + 1));
                                minor = int.Parse(strMinor);
                                if (revPos >= 0)
                                {
                                    revision = versionText.Substring(revPos + 3).Trim();
                                }
                                int stateNum = getStateData[0];
                                if (major <= 3 && stateNum >= 2) stateNum++;
                                switch (stateNum)
                                {
                                    case 0:
                                        state = BasicCardState.New;
                                        break;
                                    case 1:
                                        state = BasicCardState.Load;
                                        break;
                                    case 2:
                                        state = BasicCardState.Pers;
                                        break;
                                    case 3:
                                        state = BasicCardState.Test;
                                        break;
                                    case 4:
                                        state = BasicCardState.Run;
                                        break;
                                    default:
                                        state = BasicCardState.Unknown;
                                        break;
                                }
                            }
                            
                        }
                    }
                }                
            }
            catch (Exception)
            {
                state = BasicCardState.Unknown;
            }
            if (state == BasicCardState.Unknown)
            {
                major = -1;
                minor = -1;
                versionText = "";
                revision = null;
            }
        }
    }
}
