using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.IO;
using BasicCard.Framework;
using BasicCard.Terminals;
using BasicCard.BasicCards.Loader;

namespace BasicCardLoader
{
    public partial class LoaderForm : Form
    {
        private ICardReader reader = null;
        private BasicCardImage _image = null;
        private bool _cardAvailable = false;
        private bool _cardRemoved = false;
        private bool autoRepeat=false;
        private CardReaderEventProvider provider = null;
        private EventHandler<CardServiceEventArgs> handler;
        private EventHandler<CardReaderEventArgs> cardHandler;
        private IRequest request = null;
        private BasicCardLoadService service = null;
        private object sync = new object();
        

        public LoaderForm()
        {
            InitializeComponent();
            ActionState = ActionStateEnum.Waiting;
            handler=new EventHandler<CardServiceEventArgs>(provider_ServiceEvent);
            cardHandler = new EventHandler<CardReaderEventArgs>(provider_CardEvent);
        }

        private void LoaderForm_Load(object sender, EventArgs e)
        {
            ICardReader[] readers = CardReaderManager.Readers;
            comboBoxReaders.Items.AddRange(readers);
            comboBoxState.Items.Add(BasicCardLoadService.CardState.LOAD);
            comboBoxState.Items.Add(BasicCardLoadService.CardState.TEST);
            comboBoxState.Items.Add(BasicCardLoadService.CardState.RUN);
            comboBoxState.SelectedItem = BasicCardLoadService.CardState.TEST;            
        }

        /*
         * State diagramm
         *  [ Waiting ] - (1) -> [ Idle ] - (2) -> [ StartRequested ] - (3) -> [ Running ] - (4) -> [ Ready ]
         *     ^  ^                                        v                         v                 v v
         *     |  |                                        |                         |                 | |
         *     |  |                                       (5)                       (6)                |(8)
         *     |  ---------(7)----------[ StopRequested ]<-|-------------------------|-------(9)-------- |
         *     ------------------------------------------------------------------------------------------|
         * 
         *  State change conditions
         *  (1) Card available, Image avaiable
         *  (2) Start button clicked OR autoRepeat
         *  (3) CardService available
         *  (4) Download finished or failed
         *  (5), (6), (9) Stop button pressed
         *  (7), (8) Card removed         
         * 
         */
        private enum ActionStateEnum
        {
            Invalid=0,
            Waiting,
            Idle,
            StartRequested,
            Running,
            Ready,
            StopRequested,            
        }

        private ActionStateEnum actionState;
        private ActionStateEnum ActionState
        {
            get { return actionState; }
            set
            {
                lock (sync)
                {
                    if (value == actionState) return;
                    actionState = value;
                    switch (value)
                    {
                        case ActionStateEnum.Waiting:
                            buttonStart.Text = "Start";
                            labelRequestCard.Text = "Please insert card";
                            buttonStart.Enabled = false;
                            comboBoxReaders.Enabled = true;
                            checkBoxMultipleCards.Enabled = true;
                            comboBoxState.Enabled = true;
                            buttonBrowseFile.Enabled = true;                            
                            CheckIdleChangeCondition();
                            break;
                        case ActionStateEnum.Idle:
                            buttonStart.Text = "Start";
                            labelRequestCard.Text = "Press Start";
                            buttonStart.Enabled = true;
                            comboBoxReaders.Enabled = true;
                            checkBoxMultipleCards.Enabled = true;
                            comboBoxState.Enabled = true;
                            buttonBrowseFile.Enabled = true;
                            CardRemoved = false;
                            if (autoRepeat) ActionState = ActionStateEnum.StartRequested;
                            break;
                        case ActionStateEnum.StartRequested:                            
                            buttonStart.Text = "Stop";
                            labelRequestCard.Text = "Starting...";
                            buttonStart.Enabled = true;
                            comboBoxReaders.Enabled = false;
                            checkBoxMultipleCards.Enabled = false;
                            comboBoxState.Enabled = false;
                            buttonBrowseFile.Enabled = false;
                            if (autoRepeat)
                            {
                                if (service != null) ActionState = ActionStateEnum.Running;
                            }
                            else
                            {
                                if (service != null) { service.Dispose(); service = null; }
                                autoRepeat = checkBoxMultipleCards.Checked;
                                request = provider.RequestService(typeof(BasicCardLoadService), !autoRepeat, null);
                            }
                            break;
                        case ActionStateEnum.Running:
                            buttonStart.Text = "Stop";
                            labelRequestCard.Text = "Processing card";
                            buttonStart.Enabled = true;
                            comboBoxReaders.Enabled = false;
                            checkBoxMultipleCards.Enabled = false;
                            comboBoxState.Enabled = false;
                            buttonBrowseFile.Enabled = false;
                            break;
                        case ActionStateEnum.StopRequested:
                            CancelServiceRequest();
                            autoRepeat = false;
                            if (service != null) { service.Dispose(); service = null; }
                            buttonStart.Text = "Stop";
                            labelRequestCard.Text = "Please remove card";
                            buttonStart.Enabled = false;
                            comboBoxReaders.Enabled = false;
                            checkBoxMultipleCards.Enabled = false;
                            comboBoxState.Enabled = false;
                            buttonBrowseFile.Enabled = false;
                            break;
                        case ActionStateEnum.Ready:
                            buttonStart.Text = "Stop";
                            labelRequestCard.Text = "Please remove card";
                            buttonStart.Enabled = true;
                            comboBoxReaders.Enabled = false;
                            checkBoxMultipleCards.Enabled = false;
                            comboBoxState.Enabled = false;
                            buttonBrowseFile.Enabled = false;
                            break;
                    }                    
                }
            }
        }

        // Condition 1
        private void CheckIdleChangeCondition()
        {
            if ((ActionState == ActionStateEnum.Waiting) &&
                 (Image != null) && (CardAvailable)) ActionState = ActionStateEnum.Idle;
        }

        private BasicCardImage Image
        {
            get { return _image; }
            set
            {
                lock (sync)
                {
                    if (value != _image)
                    {
                        _image = value;
                        CheckIdleChangeCondition();
                    }
                }
            }
        }

        private bool CardAvailable
        {
            get { return _cardAvailable; }
            set
            {
                lock (sync)
                {
                    if (_cardAvailable != value)
                    {
                        _cardAvailable = value;
                        CheckIdleChangeCondition();
                    }
                }
            }
        }

        private bool CardRemoved
        {
            get { return _cardRemoved; }
            set
            {
                lock (sync)
                {
                    if (_cardRemoved != value)
                    {
                        _cardRemoved = value;
                        if (value &&
                            ((ActionState == ActionStateEnum.Ready) ||
                             (ActionState == ActionStateEnum.StopRequested)))
                        {
                            ActionState = ActionStateEnum.Waiting;
                        }
                    }
                }
            }
        }

        private void CancelServiceRequest()
        {
            lock (sync)
            {
                if (request != null)
                {
                    request.Cancel();
                    request = null;
                }
                autoRepeat = false;
            }
        }

        private void comboBoxReaders_SelectedIndexChanged(object sender, EventArgs e)
        {
            reader = (ICardReader)comboBoxReaders.SelectedItem;
            lock (sync)
            {
                if (null != provider)
                {
                    CardReaderEventProvider temp = provider;
                    provider = null;
                    temp.Dispose();
                }
                if (null != reader)
                {
                    provider = new CardReaderEventProvider(this, reader);
                    provider.ServiceEvent += handler;
                    provider.CardEvent += cardHandler;
                    CardAvailable=reader.Status.CardAvailableExclusive;                    
                }
                else
                {
                    CardAvailable = false;
                }
            }            
        }

        private void buttonBrowseFile_Click(object sender, EventArgs e)
        {
            openImageFileDialog.ReadOnlyChecked = true;
            Stream stream = null;
            if (DialogResult.OK == openImageFileDialog.ShowDialog(this))
            {
                try
                {
                    stream = openImageFileDialog.OpenFile();
                    BasicCardImage temp = new BasicCardImage(stream, checkBoxIgnoreImageVersion.Checked);
                    textBoxFile.Text = openImageFileDialog.FileName;
                    Image = temp;                    
                }
                catch (Exception ex)
                {
                    MessageBox.Show(this, ex.Message, "BasicCard Loader Failed", MessageBoxButtons.OK, MessageBoxIcon.Error);
                }
                finally
                {
                    if (null != stream) stream.Close();
                }
            }
        }

        private void buttonStart_Click(object sender, EventArgs e)
        {
            lock (sync)
            {
                switch (ActionState)
                {
                    case ActionStateEnum.Idle:
                        ActionState = ActionStateEnum.StartRequested;
                        break;
                    case ActionStateEnum.Running:
                    case ActionStateEnum.StartRequested:
                    case ActionStateEnum.Ready:
                        ActionState = ActionStateEnum.StopRequested;
                        break;
                }                
            }
        }

        void provider_CardEvent(object sender, CardReaderEventArgs e)
        {
            lock (sync)
            {
                switch (e.CardEvent.EventType)
                {
                    case CardEventType.StatusChange:
                        if (null != e.CardReader)
                        {
                            CardAvailable = e.CardReader.Status.CardAvailableExclusive;                            
                        }
                        else
                        {
                            CardAvailable = false;
                        }    
                
                        break;
                    case CardEventType.Remove:
                        CardRemoved = true;
                        break;
                    default:
                        break;
                }
            }
        }

        private void Download()
        {
            BasicCardImage image = this.Image;
            BasicCardLoadService.CardState state = (BasicCardLoadService.CardState)comboBoxState.SelectedItem;
            try
            {
                service.DownloadImage(state, image);
            }
            catch (Exception ex)
            {
                ActionState = ActionStateEnum.Idle;
                MessageBox.Show(this, ex.Message, "BasicCard Loader Failed", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }            
        }

        private void provider_ServiceEvent(object sender, CardServiceEventArgs e)
        {            
            CardReaderEventProvider sendingProvider = (CardReaderEventProvider)sender;
            if (e.EventType == CardServiceEventType.ServiceUnavailable)
            {
                lock (sync)
                {
                    service = null;                    
                }

            }
            else if (e.EventType == CardServiceEventType.ServiceAvailable)
            {
                bool doDownload = false;
                lock (sync)
                {
                    if (null != service) service.Dispose();
                    service = (BasicCardLoadService)e.CardService;
                    service.Invoker = this;
                    service.DownloadEvent += new EventHandler<DownloadEventArgs>(service_DownloadEvent);
                    e.ServiceConsumed = true;
                    switch (ActionState)
                    {
                        case ActionStateEnum.StartRequested:
                            ActionState = ActionStateEnum.Running;
                            doDownload = true;
                            break;                        
                        case ActionStateEnum.StopRequested:
                            CancelServiceRequest();
                            service.Dispose();
                            service = null;
                            break;

                    }
                }
                if (doDownload) Download();
            }
        }

        void service_DownloadEvent(object sender, DownloadEventArgs e)
        {
            richTextBoxLog.AppendText(e.Info + Environment.NewLine);
            progressBarCard.Maximum = e.MaxProgress;
            progressBarCard.Value = e.CurrentProgress;
            switch (e.EventType)
            {
                case DownloadEventArgs.EventTypeEnum.Ready:
                case DownloadEventArgs.EventTypeEnum.Error:                    
                    lock (sync)
                    {
                        ActionState = ActionStateEnum.Ready;
                    }
                    break;
            }
            Application.DoEvents();
        }

        private void checkBoxIgnoreImageVersion_CheckedChanged(object sender, EventArgs e)
        {
            if (checkBoxIgnoreImageVersion.Checked)
            {
                string message = "Using this option will cause image version, " +
                "by means of compiler version generated this image file, to be ignored." +
                " Loading unsupported image version may cause damaged cards. Use on your own risk." +
                "\nActivate \"Ignore Image Version\"?";
                if (DialogResult.No == MessageBox.Show(this,
                    message, "Activate Ignore Image Version?",
                    MessageBoxButtons.YesNo, MessageBoxIcon.Question, MessageBoxDefaultButton.Button2))
                {
                    checkBoxIgnoreImageVersion.Checked = false;
                }
            }
        }

        private void checkBoxUseColdReset_CheckedChanged(object sender, EventArgs e)
        {
            BasicCardLoadService.UseColdReset = checkBoxUseColdReset.Checked;
        }
        
    }
}