/*
 * Created on 14.07.2005
 * Project: BasicCard .Net
 * --------------------------------
 * Copyright 2005 ZeitControl cardsystems GmbH
 * --------------------------------
 * $Archive: $
 * $Revision: 1.1 $
 * $Modtime: $
 * --------------------------------
 * $History: $
 */
using System;
using BasicCard;
using BasicCard.BasicCards;

namespace Calc
{
	public enum CalcStatusWord
	{
		swSyntaxError = 0x6481,
		swParenthesisMismatch = 0x6482,
		swInvalidNumber = 0x6483,
	}

	/// <summary>
	/// Service class for managing calc example card.
	/// </summary>
	public class CalcService : AbstractBasicCardService
	{
		public CalcService()
			// specify Application ID of card we like to use
			: base("BasicCard Mini-Calculator")
		{
			doesNeedExclusiveConnection = false; // we like to share the card
			// set the encryption key see Calckeys.bas
			byte[] key={0xD1, 0x47, 0x52, 0xF8,
					   0xE2, 0xD8, 0xFE, 0xDA };
			SetKey(1, key);
		}

		/// <summary>
		/// Activate encryption.
		/// </summary>
		public void StartEncryption()
		{
			// If we are using the card none exclusive,
			// others may access the same card at the same time.
			// This will not work if encryption is active. So
			// in this case lock the card first for exclusive use.
			if (!IsExclusive)
				Lock();
			// Activate encryption by use of autoEncryption function.
			// Depending on BasicCard type either key 1 (single application
			// BasicCard) or key component "\Key1" (MultiApplication BasicCard)
			// is used within card. To use autoEncryption the specified key
			// (1) must be known by the base class AbstractBasicCardService.
			// For this it must have been set by use of setKey function (see
			// CalcService constructor).
			AutoEncryption(1, "\\Key1");
		}

		/// <summary>
		/// Deactivate encryption.
		/// </summary>
		public void EndEncryption()
		{
			// End active encryption
			CmdEndEncryption();
			// Release exclusive use
			Unlock();
		}
		/// <summary>
		/// Call calculator command by calling one of following BasicCard commands:
		/// <code>
		/// Declare Command &H20 &H01 Calculator (Expression$)
		/// Declare Command &H20 &H01 HexCalculator (P1=1, Expression$)
		/// </code>
		/// </summary>
		/// <param name="hex">if true expression must be hexadecimal, otherwise decimal</param>
		/// <param name="expression">expression to evaluate</param>
		/// <returns>result</returns>
		protected string IntCalc(bool hex, string expression)
		{
			// Construct a new BasicCardCommand object with CLA=0x20 (&H20)
			// and INS=0x01 (&H01)
			BasicCardCommand cmd = new BasicCardCommand(0x20, 0x01);
			if (hex)
			{
				// we want to call HexCalculator, set P1 to 1
				cmd.P1=1;
			}
			// BasicCard strings and .Net strings vary regarding encoding.
			// A BasicCard string uses 8 bit per character, a .Net string uses
			// 16 bit per character (Unicode). Such we use our BasicCardString
			// class when sending strings to BasicCard. The BasicCardString class
			// provides translations between .Net and BasicCard strings. Please 
			// note: It depends on your application and expected string contents
			// which translation is a proper choice. Here we expect to use ACSII
			// characters only, such CharsetIdAscii is used.
			BasicCardString bsExpression =
				new BasicCardString(
					expression,
					BasicCardString.CharsetIdAscii);
			// Add expression string as first parameter to the BasicCardCommand object
			cmd.AppendBasicString(bsExpression);
			// Call the BasicCard command, now the command is transmitted to the BasicCard. 
			BasicCardResponse rsp = DoCommandAndResponse(cmd);
			// We have received the response from BasicCard. Check SW1SW2 for
			// error conditions. In case SW1SW2 is neither 0x9000 nor 0x61xx this
			// will cause a CommsExecption to be thrown.
			rsp.CheckSW1SW2();
			// Get the expression string returned from card. Again we get a BasicCardString
			// object, not a Java String object.
			bsExpression = rsp.FetchBasicString();
			// Translate the response expression to a valid Java string by use of
			// ASCII character set.
			return bsExpression.ToString(BasicCardString.CharsetIdAscii);
		}

		/// <summary>
		/// Call calculator command by calling the following BasicCard command:
		/// <code>
		/// Declare Command &H20 &H01 Calculator (Expression$)
		/// </code>
		/// </summary>
		/// <param name="expression">Expression in decimal format to evaluate, e.g. "10*9".</param>
		/// <returns>Resulting expression string, e.g. "90".</returns>
		public string Calculator(string expression)
		{
			return IntCalc(false, expression);
		}

		/// <summary>
		/// Call calculator command by calling the following BasicCard command:
		/// <code>
		/// Declare Command &H20 &H01 HexCalculator (P1=1, Expression$)
		/// </code>
		/// </summary>
		/// <param name="expression">Expression in hexadecimal format to evaluate, e.g. "1F+A".</param>
		/// <returns>Resulting expression string, e.g. "29".</returns>
		public string HexCalculator(String expression)
		{
			return IntCalc(true, expression);
		}
	}
}
