/*
 * Created on 14.07.2005
 * Project: BasicCard .Net
 * --------------------------------
 * Copyright 2005 ZeitControl cardsystems GmbH
 * --------------------------------
 * $Archive: $
 * $Revision: 1.1 $
 * $Modtime: $
 * --------------------------------
 * $History: $
 */
using System;
using BasicCard;
using BasicCard.BasicCards;

namespace Debit
{
	/// <summary>
	/// Card data
	/// </summary>
	public class CardData
	{
		private decimal amount;
		private int pinCount;
		private string name;

		/// <summary>
		/// Construct a CardData object.
		/// </summary>
		internal CardData()
		{
		}

		/// <summary>
		/// Amount left within card.
		/// </summary>
		public decimal Amount
		{
			get
			{
				return amount;
			}
			set
			{
				amount = value;
			}
		}

		/// <summary>
		/// Amount left within card as string.
		/// </summary>
		public string AmountString
		{
			get
			{
				return amount.ToString("0."+new string('0', DealerService.DecimalNumbers));
			}
		}

		/// <summary>
		/// Card owners name.
		/// </summary>
		public string Name
		{
			get
			{
				return name;
			}
			set
			{
				name = value;
			}
		}

		/// <summary>
		/// Number of PIN tries left.
		/// </summary>
		public int PinCount
		{
			get
			{
				return pinCount;
			}
			set
			{
				pinCount = value;
			}
		}
	}


	/// <summary>
	/// Service class for managing dealer functions for BasicCard debit example card.
	/// </summary>
	public class DealerService : AbstractBasicCardService
	{
		public enum DebitStatusWord
		{
			swNotPersonalised = 0x6B00,
			swInsufficientFunds = 0x6B01,
			swEncryptionRequired = 0x6B02,
			swIssuingKeyRequired = 0x6B03,
			swPINRequired = 0x6B04,
			swInvalidPIN = 0x6B05,
			swPINErrorsExceeded = 0x6B06,
		}

		// Change this to use different encodings when translating between
		// BasicCard String and .Net String
		protected const string StringEncoding = BasicCardString.CharsetIdAscii;

		// Decimal representation used internal. E.g. when
		// DecimalNumbers = 2 then an amount of 10.25 will be encoded to (int) 1025
		internal const int DecimalNumbers = 2;
		private const int DecimalFactor = 100; // DecimalNumbers^10

		protected static int FromDecimal(decimal val)
		{
			return (int) decimal.Round(val * DecimalFactor, 0);
		}

		protected static decimal ToDecimal(int val)
		{
			return ((decimal)val) / DecimalFactor;
		}


		/// <summary>
		/// Constructs a DealerService object.
		/// </summary>
		public DealerService()
			// Select card by application "ZeitControl DebitCard"
			: base("ZeitControl DebitCard")
		{
			// Dealer key
			byte[] key =
			{
				(byte) 0x86,
				(byte) 0x75,
				(byte) 0x90,
				(byte) 0x03,
				(byte) 0xED,
				(byte) 0x88,
				(byte) 0xF2,
				(byte) 0xB3 };
			SetKey(1, key);
		}

		/// <summary>
		/// Start encryption for BasicCard commands using dealer key.
		/// </summary>
		public virtual void StartEncryption()
		{
			AutoEncryption(1, "\\Key1");
		}

		/// <summary>
		/// Stop encryption of BasicCard commands.
		/// </summary>
		public void EndEncryption()
		{
			CmdEndEncryption();
		}

		/// <summary>
		/// Get card data.
		/// </summary>
		/// <returns>card data</returns>
		public CardData GetCardData()
		{
			CardData carddata = new CardData();
			BasicCardCommand cmd = new BasicCardCommand(0x80, 0x0A);
			cmd.LE=0;
			BasicCardResponse rsp = DoCommandAndResponse(cmd);
			rsp.CheckSW1SW2();
			carddata.Amount=ToDecimal(rsp.FetchBasicLong());
			carddata.PinCount=rsp.FetchBasicInteger();
			carddata.Name=rsp.FetchBasicString().ToString(StringEncoding);
			return carddata;
		}

		/// <summary>
		/// Call VerifyPIN command
		/// </summary>
		/// <param name="pin">PIN for verification</param>
		public void VerifyPin(string pin)
		{
			if (!IsExclusive)
			{
				throw new CommsException("Must not validate a PIN to a card not owned exclusivly.");
			}
			BasicCardCommand cmd = new BasicCardCommand(0x80, 0x02);
			// for PIN we always use ASCII encoding
			cmd.AppendBasicStringN(
				new BasicCardStringN(4, pin, BasicCardString.CharsetIdAscii));
			cmd.DisableLE(); // we do not need to receive data
			BasicCardResponse rsp = DoCommandAndResponse(cmd);
			rsp.CheckSW1SW2();
		}

		/// <summary>
		/// Call ChangePIN command.
		/// </summary>
		/// <param name="pin">PIN for verification</param>
		public void ChangePin(string pin)
		{
			BasicCardCommand cmd = new BasicCardCommand(0x80, 0x08);
			// for PIN we always use ASCII encoding
			cmd.AppendBasicStringN(
				new BasicCardStringN(4, pin, BasicCardString.CharsetIdAscii));
			cmd.DisableLE(); // we do not need to receive data
			BasicCardResponse rsp = DoCommandAndResponse(cmd);
			rsp.CheckSW1SW2();
		}

		/// <summary>
		/// Call IncreaseAmount command.
		/// </summary>
		/// <param name="diff"></param>
		public void IncreaseAmount(decimal diff)
		{
			BasicCardCommand cmd = new BasicCardCommand(0x80, 0x04);
			cmd.AppendBasicLong(FromDecimal(diff));
			cmd.DisableLE(); // we do not need response data
			BasicCardResponse rsp = DoCommandAndResponse(cmd);
			rsp.CheckSW1SW2();
		}

		/// <summary>
		/// Call DecreaseAmount command.
		/// </summary>
		/// <param name="diff">value to subtract (>0)</param>
		public void DecreaseAmount(decimal diff)
		{
			BasicCardCommand cmd = new BasicCardCommand(0x80, 0x06);
			cmd.AppendBasicLong(FromDecimal(diff));
			cmd.DisableLE(); // we do not need response data
			BasicCardResponse rsp = DoCommandAndResponse(cmd);
			rsp.CheckSW1SW2();
		}

		public void ShowMessage(CommsException e)
		{
			string message = "";
			switch ((DebitStatusWord)e.SW1SW2)
			{
				case DebitStatusWord.swNotPersonalised :
					message = "Card is not personalised!";
					break;
				case DebitStatusWord.swInsufficientFunds :
					message = "Insufficient funds!";
					break;
				case DebitStatusWord.swEncryptionRequired :
					message = "Encryption required!";
					break;
				case DebitStatusWord.swIssuingKeyRequired :
					message = "Issuing key required!";
					break;
				case DebitStatusWord.swPINRequired :
					message = "PIN required!";
					break;
				case DebitStatusWord.swInvalidPIN :
					message = "Invalid PIN!";
					break;
				case DebitStatusWord.swPINErrorsExceeded :
					message = "Too many PIN errors - card disabled";
					break;
				default :
					e.Dump();
					break;
			}
			System.Console.WriteLine(message);
		}
	}
}
