' Created on 14.07.2005
' Project: BasicCard .Net
' --------------------------------
' Copyright 2005 ZeitControl cardsystems GmbH
' --------------------------------
' $Archive: $
' $Revision: 1.1 $
' $Modtime: $
' --------------------------------
' $History: $

Option Explicit On 
Imports BasicCard
Imports BasicCard.BasicCards

' Card data
Public Class CardData
    Private amountVal As Decimal
    Private pinCountVal As Integer
    Private nameVal As String

    ' Construct a CardData object.
    Friend Sub New()
    End Sub

    ' Amount left within card.
    Public Property Amount() As Decimal
        Get
            Return amountVal
        End Get
        Set(ByVal Value As Decimal)
            amountVal = Value
        End Set
    End Property

    ' Amount left within card as string.
    Public ReadOnly Property AmountString() As String
        Get
            Return amountVal.ToString("0." + New String("0", DealerService.DecimalNumbers))
        End Get
    End Property

    ' Card owners name.
    Public Property Name() As String
        Get
            Return nameVal
        End Get
        Set(ByVal Value As String)
            nameVal = Value
        End Set
    End Property

    ' Number of PIN tries left.
    Public Property PinCount() As Integer
        Get
            Return pinCountVal
        End Get
        Set(ByVal Value As Integer)
            pinCountVal = Value
        End Set
    End Property
End Class

Public Enum DebitStatusWord
    swNotPersonalised = &H6B00
    swInsufficientFunds = &H6B01
    swEncryptionRequired = &H6B02
    swIssuingKeyRequired = &H6B03
    swPINRequired = &H6B04
    swInvalidPIN = &H6B05
    swPINErrorsExceeded = &H6B06
End Enum


' Service class for managing dealer functions for BasicCard debit example card.
Public Class DealerService
    Inherits AbstractBasicCardService

    ' Change this to use different encodings when translating between
    ' BasicCard String and Java String
    Protected Const StringEncoding As String = BasicCardString.CharsetIdAscii

    ' Decimal representation used internal. E.g. when
    ' DecimalNumbers = 2 then an amount of 10.25 will be encoded to (int) 1025
    Friend Const DecimalNumbers As Integer = 2
    Private Const DecimalFactor As Integer = 100 ' DecimalNumbers^10

    Protected Shared Function FromDecimal(ByVal value As Decimal) As Integer
        Return Decimal.Round(value * DecimalFactor, 0)
    End Function

    Protected Shared Function ToDecimal(ByVal value As Integer) As Decimal
        Dim val As Decimal = value
        Return val / DecimalFactor
    End Function

    ' Constructs a DealerService object.
    Public Sub New()
        ' Select card by application "ZeitControl DebitCard"
        MyBase.New("ZeitControl DebitCard")
        ' Dealer key
        Dim key() As Byte = {&H86, &H75, &H90, &H3, _
          &HED, &H88, &HF2, &HB3}
        SetKey(1, key)
    End Sub

    ' Start encryption for BasicCard commands using dealer key.
    Public Overridable Sub StartEncryption()
        AutoEncryption(1, "\Key1")
    End Sub

    ' Stop encryption of BasicCard commands.
    Public Sub EndEncryption()
        CmdEndEncryption()
    End Sub

    ' Get card data.
    ' returns card data
    Public Function GetCardData() As CardData
        Dim carddata As CardData = New CardData
        Dim cmd As BasicCardCommand = New BasicCardCommand(&H80, &HA)
        cmd.LE = 0
        Dim rsp As BasicCardResponse = DoCommandAndResponse(cmd)
        rsp.CheckSW1SW2()
        carddata.Amount = ToDecimal(rsp.FetchBasicLong())
        carddata.PinCount = rsp.FetchBasicInteger()
        carddata.Name = rsp.FetchBasicString().ToString(StringEncoding)
        Return carddata
    End Function
		
    ' Call VerifyPIN command
    ' pin: PIN for verification
    Public Sub VerifyPin(ByVal pin As String)
        If Not IsExclusive Then
            Throw New CommsException("Must not validate a PIN to a card not owned exclusivly.")
        End If
        Dim cmd As BasicCardCommand = New BasicCardCommand(&H80, &H2)
        ' for PIN we always use ASCII encoding
        cmd.AppendBasicStringN(New BasicCardStringN(4, pin, BasicCardString.CharsetIdAscii))
        cmd.DisableLE() ' we do not need to receive data
        Dim rsp As BasicCardResponse = DoCommandAndResponse(cmd)
        rsp.CheckSW1SW2()
    End Sub

    ' Call ChangePIN command.
    ' pin: PIN for verification
    Public Sub ChangePin(ByVal pin as String)
        Dim cmd As BasicCardCommand = New BasicCardCommand(&H80, &H8)
        ' for PIN we always use ASCII encoding
        cmd.AppendBasicStringN(New BasicCardStringN(4, pin, BasicCardString.CharsetIdAscii))
        cmd.DisableLE() ' we do not need to receive data
        Dim rsp As BasicCardResponse = DoCommandAndResponse(cmd)
        rsp.CheckSW1SW2()
    End Sub

    ' Call IncreaseAmount command.
    Public Sub IncreaseAmount(ByVal diff as Decimal)
        Dim cmd As BasicCardCommand = New BasicCardCommand(&H80, &H4)
        cmd.AppendBasicLong(FromDecimal(diff))
        cmd.DisableLE() ' we do not need response data
        Dim rsp As BasicCardResponse = DoCommandAndResponse(cmd)
        rsp.CheckSW1SW2()
    End Sub

    ' Call DecreaseAmount command.
    Public Sub DecreaseAmount(ByVal diff As Decimal)
        Dim cmd As BasicCardCommand = New BasicCardCommand(&H80, &H6)
        cmd.AppendBasicLong(FromDecimal(diff))
        cmd.DisableLE() ' we do not need response data
        Dim rsp As BasicCardResponse = DoCommandAndResponse(cmd)
        rsp.CheckSW1SW2()
    End Sub

    Public Sub ShowMessage(ByVal e As CommsException)
        Dim message As String = ""
        Select Case e.SW1SW2
            Case DebitStatusWord.swNotPersonalised
                message = "Card is not personalised!"

            Case DebitStatusWord.swInsufficientFunds
                message = "Insufficient funds!"

            Case DebitStatusWord.swEncryptionRequired
                message = "Encryption required!"

            Case DebitStatusWord.swIssuingKeyRequired
                message = "Issuing key required!"

            Case DebitStatusWord.swPINRequired
                message = "PIN required!"

            Case DebitStatusWord.swInvalidPIN
                message = "Invalid PIN!"

            Case DebitStatusWord.swPINErrorsExceeded
                message = "Too many PIN errors - card disabled"

            Case Else
                e.Dump()

        End Select
        System.Console.WriteLine(message)
    End Sub

End Class
